@@ -27,10 +27,10 @@
 #include "config.h"
 
 #define _POSIX_C_SOURCE 2
-#include <gnome.h>
 
 #include <stdio.h>
 #include <sys/stat.h>
+#include <libgnome/libgnome.h>
 
 #include "address-book.h"
 #include "address-book-extern.h"
@@ -140,11 +140,6 @@ libbalsa_address_book_externq_init(LibBalsaAddressBookExtern * ab)
     ab->save = NULL;
     ab->address_list = NULL;
     ab->mtime = 0;
-
-    ab->name_complete  = 
-	g_completion_new((GCompletionFunc)completion_data_extract);
-    ab->alias_complete = 
-	g_completion_new((GCompletionFunc)completion_data_extract);
 }
 
 static void
@@ -348,7 +343,7 @@ libbalsa_address_book_externq_alias_complete(LibBalsaAddressBook * ab,
 {
     LibBalsaAddressBookExtern *ex;
     GList *res = NULL;
-    *new_prefix = NULL;
+    if(new_prefix) *new_prefix = NULL;
 
     g_return_val_if_fail(LIBBALSA_IS_ADDRESS_BOOK_EXTERN(ab), NULL);
 
@@ -362,7 +357,7 @@ libbalsa_address_book_externq_alias_complete(LibBalsaAddressBook * ab,
 	
     g_list_reverse(res);
 
-    if(res != NULL)
+    if(res != NULL && new_prefix)
         *new_prefix = libbalsa_address_to_gchar((LibBalsaAddress *)
                                                 g_list_first(res)->data, 0);
 
