@@ -36,6 +36,7 @@
 #include <string.h>
 #include <sys/stat.h>
 #include <sys/types.h>
+#include <time.h>
 
 #include "address-book.h"
 #include "address-book-gpe.h"
@@ -233,11 +234,9 @@ gpe_read_attr(void *arg, int argc, char **argv, char **names)
             !a->nick_name) a->nick_name = g_strdup(argv[1]);
     else if(g_ascii_strcasecmp(argv[0], "WORK.ORGANIZATION") == 0 &&
             !a->organization) a->organization = g_strdup(argv[1]);
-    else if(g_ascii_strcasecmp(argv[0], "HOME.EMAIL") == 0 &&
-            !a->address_list)
+    else if(g_ascii_strcasecmp(argv[0], "HOME.EMAIL") == 0)
         a->address_list = g_list_prepend(a->address_list, g_strdup(argv[1]));
-    else if(g_ascii_strcasecmp(argv[0], "WORK.EMAIL") == 0 &&
-            !a->address_list)
+    else if(g_ascii_strcasecmp(argv[0], "WORK.EMAIL") == 0)
         a->address_list = g_list_prepend(a->address_list, g_strdup(argv[1]));
     return 0;
 }
@@ -330,7 +329,7 @@ libbalsa_address_book_gpe_load(LibBalsaAddressBook * ab,
              "select distinct urn from contacts where "
              "(upper(tag)='FAMILY_NAME' or upper(tag)='GIVEN_NAME' or "
              "upper(tag)='NAME' or "
-             "upper(tag)=WORK.EMAIL' or upper(tag)='HOME.EMAIL') "
+             "upper(tag)='WORK.EMAIL' or upper(tag)='HOME.EMAIL') "
              "and value LIKE '%q%%'",
              gpe_read_address, &gc, &err, filter);
     } else {
@@ -340,7 +339,7 @@ libbalsa_address_book_gpe_load(LibBalsaAddressBook * ab,
 
     if(r != SQLITE_OK) {
         printf("r=%d err=%s\n", r, err);
-        libbalsa_address_book_set_status(ab, err);
+        libbalsa_address_book_set_status(ab, g_strdup(err));
         free(err);
         ret = LBABERR_CANNOT_READ;
     } else {
@@ -376,58 +375,44 @@ libbalsa_address_book_gpe_add_address(LibBalsaAddressBook *ab,
     r = sqlite_exec(gpe_ab->db, "insert into contacts_urn values (NULL)",
                     NULL, NULL, &err);
     if (r != SQLITE_OK) {
-        libbalsa_address_book_set_status(ab, err);
+        libbalsa_address_book_set_status(ab, g_strdup(err));
         free(err);
         return LBABERR_CANNOT_WRITE;
     }
     /* FIXME: duplicate detection! */
 
     id = sqlite_last_insert_rowid(gpe_ab->db);
-    INSERT_ATTR(gpe_ab->db,id, "NAME",  address->full_name);
+    INSERT_ATTR(gpe_ab->db,id, "NAME",        address->full_name);
     INSERT_ATTR(gpe_ab->db,id, "GIVEN_NAME",  address->first_name);
     INSERT_ATTR(gpe_ab->db,id, "FAMILY_NAME", address->last_name);
     INSERT_ATTR(gpe_ab->db,id, "NICKNAME",    address->nick_name);
     INSERT_ATTR(gpe_ab->db,id, "WORK.ORGANIZATION", address->organization);
     INSERT_ATTR(gpe_ab->db,id, "WORK.EMAIL",
                 (char*)address->address_list->data);
-
+    sqlite_exec_printf(gpe_ab->db, "insert into contacts values "
+                       "('%d', 'MODIFIED', %d)", NULL, NULL, NULL,
+                       id, time(NULL));
     return LBABERR_OK;
 }
-static gboolean
+static gchar*
 db_delete_by_uid(sqlite *db, guint uid)
 {
-  int r;
-  gchar *err;
-  gboolean rollback = FALSE;
-
-  r = sqlite_exec (db, "begin transaction", NULL, NULL, &err);
-  if (r)
-    goto error;
-
-  rollback = TRUE;
-
-  r = sqlite_exec_printf (db, "delete from contacts where urn='%d'",
-                          NULL, NULL, &err, uid);
-  if (r)
-    goto error;
-
-  r = sqlite_exec_printf (db, "delete from contacts_urn where urn='%d'",
-                          NULL, NULL, &err, uid);
-  if (r)
-    goto error;
-
-  r = sqlite_exec (db, "commit transaction", NULL, NULL, &err);
-  if (r)
-    goto error;
-
-  return TRUE;
-
-error:
-  if (rollback)
-    sqlite_exec (db, "rollback transaction", NULL, NULL, NULL);
-  /* do something with err */
-  free (err);
-  return FALSE;
+    int r;
+    gchar *err;
+    
+    if( (r=sqlite_exec(db, "begin transaction",
+                       NULL, NULL, &err)) == SQLITE_OK)
+        return err;
+    
+    if( (r=sqlite_exec_printf(db, "delete from contacts where urn='%d'",
+                              NULL, NULL, &err, uid)) == SQLITE_OK &&
+        (r=sqlite_exec_printf(db, "delete from contacts_urn where urn='%d'",
+                              NULL, NULL, &err, uid)) == SQLITE_OK &&
+        (r=sqlite_exec(db, "commit transaction", NULL,
+                       NULL, &err)) == SQLITE_OK) return NULL;
+    
+    sqlite_exec(db, "rollback transaction", NULL, NULL, NULL);
+    return err;
 }
 
 static LibBalsaABErr
@@ -436,6 +421,7 @@ libbalsa_address_book_gpe_remove_address(LibBalsaAddressBook *ab,
 {
     LibBalsaAddressBookGpe *gpe_ab = LIBBALSA_ADDRESS_BOOK_GPE(ab);
     guint uid;
+    char *err;
     g_return_val_if_fail(address, LBABERR_CANNOT_WRITE);
     g_return_val_if_fail(address->address_list, LBABERR_CANNOT_WRITE);
 
@@ -447,31 +433,105 @@ libbalsa_address_book_gpe_remove_address(LibBalsaAddressBook *ab,
     if(!uid)/* safety check, perhaps unnecessary */
         return LBABERR_CANNOT_WRITE;
         
-    return db_delete_by_uid(gpe_ab->db, uid) 
-        ? LBABERR_OK : LBABERR_CANNOT_WRITE;
+    err = db_delete_by_uid(gpe_ab->db, uid);
+    if(err) {
+        libbalsa_address_book_set_status(ab, g_strdup(err));
+        free(err);
+        return LBABERR_CANNOT_WRITE;
+    } else return LBABERR_OK;
 }
 
 
 /** libbalsa_address_book_gpe_modify_address:
     modify given address.
 */
+#define INSERT_ATTR_R(db,id,attr,val) \
+ if( (val) && *(val)) {\
+ if( (r=sqlite_exec_printf \
+         ((db), "insert into contacts values ('%d', '%q', '%q')",\
+                    NULL, NULL, &err, (id), (attr), (val))) != SQLITE_OK)\
+    goto rollback;}
+
 static LibBalsaABErr
 libbalsa_address_book_gpe_modify_address(LibBalsaAddressBook *ab,
-                                          LibBalsaAddress *address,
-                                          LibBalsaAddress *newval)
+                                         LibBalsaAddress *address,
+                                         LibBalsaAddress *newval)
 {
+    LibBalsaAddressBookGpe *gpe_ab = LIBBALSA_ADDRESS_BOOK_GPE(ab);
+    guint uid;
+    int r;
+    char *err;
+
     g_return_val_if_fail(address, LBABERR_CANNOT_WRITE);
     g_return_val_if_fail(address->address_list, LBABERR_CANNOT_WRITE);
     g_return_val_if_fail(newval->address_list, LBABERR_CANNOT_WRITE);
-    /* alternatively follow up commit_person */
+
+    if (gpe_ab->db == NULL) {
+        if( !libbalsa_address_book_gpe_open_db(gpe_ab))
+	    return LBABERR_CANNOT_CONNECT;
+    }
+    uid = GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(address), "urn"));
+    /* safety check, perhaps unnecessary */
+    g_return_val_if_fail(uid, LBABERR_CANNOT_WRITE);
+
+    /* do the real work here */
+    if( (r=sqlite_exec(gpe_ab->db, "begin transaction",
+                       NULL, NULL, &err)) != SQLITE_OK) {
+        libbalsa_address_book_set_status(ab, g_strdup(err));
+        free(err); /* failed, so soon!? */
+        return LBABERR_CANNOT_WRITE;
+    }
+
+    if( (r=sqlite_exec_printf
+         (gpe_ab->db,
+          "delete from contacts where urn='%d' and "
+          "(upper(tag)='NAME' or upper(tag)='GIVEN_NAME' or "
+          "upper(tag)='NICKNAME' or upper(tag)='WORK.ORGANIZATION' or "
+          "upper(tag)='WORK.EMAIL' or upper(tag)='MODIFIED')",
+          NULL, NULL, &err, uid)) != SQLITE_OK)
+        goto rollback;
+    INSERT_ATTR_R(gpe_ab->db,uid, "NAME",        newval->full_name);
+    INSERT_ATTR_R(gpe_ab->db,uid, "GIVEN_NAME",  newval->first_name);
+    INSERT_ATTR_R(gpe_ab->db,uid, "FAMILY_NAME", newval->last_name);
+    INSERT_ATTR_R(gpe_ab->db,uid, "NICKNAME",    newval->nick_name);
+    INSERT_ATTR_R(gpe_ab->db,uid, "WORK.ORGANIZATION", newval->organization);
+    INSERT_ATTR_R(gpe_ab->db,uid, "WORK.EMAIL",
+                (char*)newval->address_list->data);
+    if( (r=sqlite_exec_printf(gpe_ab->db, "insert into contacts values "
+                              "('%d', 'MODIFIED', %d)", NULL, NULL, &err,
+                              uid, time(NULL))) != SQLITE_OK)
+        goto rollback;
+    
+    if( (r=sqlite_exec(gpe_ab->db, "commit transaction", NULL,
+                       NULL, &err)) == SQLITE_OK) return LBABERR_OK;
+
+ rollback:
+    libbalsa_address_book_set_status(ab, g_strdup(err));
+    free(err);
+    sqlite_exec(gpe_ab->db, "rollback transaction", NULL, NULL, NULL);
+
     return LBABERR_CANNOT_WRITE;
 }
 
 struct gpe_completion_closure {
     sqlite *db;
+    const gchar *prefix;
     gchar **new_prefix;
     GList *res;
 };
+static void
+strip_superflous_addresses(LibBalsaAddress *a)
+{
+    if(a->address_list) {
+        /* a person has several matching addresses of the same kind! */
+        GList *l, *tmp;
+        for(l=a->address_list->next; l; l = tmp){
+            tmp = l->next;
+            g_free(l->data);
+            a->address_list = g_list_delete_link(a->address_list, l);
+        }
+    }
+}
 
 static int
 gpe_read_completion(void *arg, int argc, char **argv, char **names)
@@ -479,19 +539,31 @@ gpe_read_completion(void *arg, int argc, char **argv, char **names)
     struct gpe_completion_closure *gc = arg;
     LibBalsaAddress * a= libbalsa_address_new();
     guint uid = atoi(argv[0]);
+    gchar *tag = argv[1];
 
     /* follow read_entry_data. FIXME: error reporting */
-    sqlite_exec_printf (gc->db,
-                        "select tag,value from contacts where urn=%d",
-                        gpe_read_attr, a, NULL, uid);
+    if(strstr(tag, "EMAIL") != NULL) { /* be extra cautious - return just one address */
+        sqlite_exec_printf
+            (gc->db,
+             "select tag,value from contacts where urn=%d"
+             " and (upper(tag)='FAMILY_NAME' or upper(tag)='GIVEN_NAME' or "
+             "upper(tag)='NAME' or "
+             "(upper(tag)='%q' and value LIKE '%q%%'))",
+             gpe_read_attr, a, NULL, uid, tag, gc->prefix);
+    } else {
+        sqlite_exec_printf(gc->db,
+                            "select tag,value from contacts where urn=%d",
+                            gpe_read_attr, a, NULL, uid);
+    }
     if(!a->address_list) { /* entry without address: ignore! */
         g_object_unref(a);
         return 0;
     }
+    strip_superflous_addresses(a);
     if(!a->full_name)
         a->full_name = create_name(a->first_name, a->last_name);
     g_object_set_data(G_OBJECT(a), "urn", GUINT_TO_POINTER(uid));
-    if(!*gc->new_prefix)
+    if(gc->new_prefix && !*gc->new_prefix)
         *gc->new_prefix = libbalsa_address_to_gchar(a, 0);
     gc->res = g_list_prepend(gc->res, a);
 
@@ -504,8 +576,8 @@ libbalsa_address_book_gpe_alias_complete(LibBalsaAddressBook * ab,
 					  gchar ** new_prefix)
 {
     static const char *query = 
-        "select distinct urn from contacts where "
-        "(upper(tag)='FAMILY_NAME' or upper(tag)='FIRST_NAME' or "
+        "select distinct urn,tag from contacts where "
+        "(upper(tag)='FAMILY_NAME' or upper(tag)='GIVEN_NAME' or "
         "upper(tag)='NAME' or "
         "upper(tag)='WORK.EMAIL' or upper(tag)='HOME.EMAIL') "
         "and upper(value) LIKE '%q%%'";
@@ -524,8 +596,9 @@ libbalsa_address_book_gpe_alias_complete(LibBalsaAddressBook * ab,
 	    return NULL;
     }
 
-    *new_prefix = NULL;
+    if(new_prefix) *new_prefix = NULL;
     gcc.db = gpe_ab->db;
+    gcc.prefix = prefix;
     gcc.new_prefix = new_prefix;
     gcc.res = NULL;
     if(prefix)
