@@ -22,9 +22,10 @@
 
 #include "config.h"
 
-#include <gnome.h>
 #include <gtk/gtkmarshal.h>
 
+#include <libgnome/libgnome.h>
+
 #include "address-book.h"
 #include "libbalsa-marshal.h"
 
