@@ -28,8 +28,8 @@
 #include <string.h>
 #include <glib.h>
 #include <fcntl.h>
-/* gnome-i18n.h needed for _() */
-#include <libgnome/gnome-i18n.h>
+/* libgnome.h needed for _() */
+#include <libgnome/libgnome.h>
 
 #include "libbalsa.h"
 #include "misc.h"
@@ -109,7 +109,7 @@ libbalsa_message_body_extract_embedded_headers(GMimeMessage* msg)
 
     subj = g_mime_message_get_subject(msg);
     if (subj) {
-	ehdr->subject = g_mime_utils_8bit_header_decode(subj);
+	ehdr->subject = g_mime_utils_header_decode_text(subj);
 	libbalsa_utf8_sanitize(&ehdr->subject, TRUE, NULL);
     } else 
 	ehdr->subject = g_strdup(_("(No subject)"));
@@ -180,18 +180,9 @@ libbalsa_message_body_set_message_part(LibBalsaMessageBody * body,
 	*next_part = libbalsa_message_body_new(body->message);
     libbalsa_message_body_set_mime_body(*next_part,
 					embedded_message->mime_part);
-    next_part = &(*next_part)->next;
-
-    if (GMIME_IS_PART(embedded_message->mime_part))
-	/* This part may not have a Content-Disposition header, but
-	 * we must treat it as inline. */
-	g_mime_part_set_content_disposition(GMIME_PART
-					    (embedded_message->mime_part),
-					    GMIME_DISPOSITION_INLINE);
-
     g_object_unref(embedded_message);
 
-    return next_part;
+    return &(*next_part)->next;
 }
 
 static LibBalsaMessageBody **
@@ -360,16 +351,66 @@ libbalsa_message_body_save(LibBalsaMessageBody * body,
     return libbalsa_message_body_save_fd(body, fd);
 }
 
-gboolean
-libbalsa_message_body_save_fd(LibBalsaMessageBody * body, int fd)
+static GMimeStream *
+libbalsa_message_body_assure_stream_filter(GMimeStream * stream)
 {
-    GMimeStream *stream, *filter_stream;
+    GMimeStream *filtered_stream;
+
+    if (GMIME_IS_STREAM_FILTER(stream))
+        return stream;
+
+    filtered_stream = g_mime_stream_filter_new_with_stream(stream);
+    g_object_unref(stream);
+
+    return filtered_stream;
+}
+
+GMimeStream *
+libbalsa_message_body_get_stream(LibBalsaMessageBody * body)
+{
+    GMimeStream *stream;
+    GMimeFilter *filter;
     gchar *mime_type = NULL;
-    const char *charset;
-    GMimeFilter *filter;
-    gboolean retval = TRUE;
+    const gchar *charset;
+
+    if (libbalsa_mailbox_get_message_part(body->message, body)) {
+        GMimeDataWrapper *wrapper;
+
+        wrapper =
+            g_mime_part_get_content_object(GMIME_PART(body->mime_part));
+        stream = g_mime_data_wrapper_get_stream(wrapper);
+
+        filter = NULL;
+        switch (g_mime_data_wrapper_get_encoding(wrapper)) {
+        case GMIME_PART_ENCODING_BASE64:
+            filter =
+                g_mime_filter_basic_new_type(GMIME_FILTER_BASIC_BASE64_DEC);
+            break;
+        case GMIME_PART_ENCODING_QUOTEDPRINTABLE:
+            filter =
+                g_mime_filter_basic_new_type(GMIME_FILTER_BASIC_QP_DEC);
+            break;
+        case GMIME_PART_ENCODING_UUENCODE:
+            filter =
+                g_mime_filter_basic_new_type(GMIME_FILTER_BASIC_UU_DEC);
+            break;
+        default:
+            break;
+        }
+        g_object_unref(wrapper);
+
+        if (filter) {
+	    stream = libbalsa_message_body_assure_stream_filter(stream);
+            g_mime_stream_filter_add(GMIME_STREAM_FILTER(stream), filter);
+            g_object_unref(filter);
+        }
+    } else if (body->mime_part) {
+        /* Not a GMimePart... */
+        stream = g_mime_stream_mem_new();
+        g_mime_object_write_to_stream(body->mime_part, stream);
+    } else
+        return NULL;
 
-    stream = g_mime_stream_fs_new(fd);
     /* convert text bodies but HTML - gtkhtml does conversion on its own. */
     if (libbalsa_message_body_type(body) == LIBBALSA_MESSAGE_BODY_TYPE_TEXT
 	&& strcmp(mime_type = libbalsa_message_body_get_mime_type(body),
@@ -377,30 +418,95 @@ libbalsa_message_body_save_fd(LibBalsaMessageBody * body, int fd)
 	&& (charset = libbalsa_message_body_charset(body))
 	&& g_ascii_strcasecmp(charset, "unknown-8bit") != 0
 	&& (filter = g_mime_filter_charset_new(charset, "UTF-8")) != NULL) {
-	filter_stream = g_mime_stream_filter_new_with_stream(stream);
-	g_mime_stream_filter_add(GMIME_STREAM_FILTER(filter_stream), filter);
+	stream = libbalsa_message_body_assure_stream_filter(stream);
+	g_mime_stream_filter_add(GMIME_STREAM_FILTER(stream), filter);
 	g_object_unref(filter);
-	g_object_unref(stream);
-	stream = filter_stream;
     }
     g_free(mime_type);
 
-    if (libbalsa_mailbox_get_message_part(body->message, body)) {
-	GMimeDataWrapper *wrapper;
+    g_mime_stream_reset(stream);
+    return stream;
+}
 
-	wrapper = g_mime_part_get_content_object(GMIME_PART(body->mime_part));
-	if (g_mime_data_wrapper_write_to_stream(wrapper, stream) == -1)
-            retval = FALSE;
-	g_object_unref(wrapper);
-    } else {
-	/* Not a GMimePart... */
-	if (body->mime_part != NULL
-	    && g_mime_object_write_to_stream(body->mime_part, stream) == -1)
-	    retval = FALSE;
+gssize
+libbalsa_message_body_get_content(LibBalsaMessageBody * body, gchar ** buf)
+{
+    GMimeStream *stream, *stream_mem;
+    GByteArray *array;
+    gssize len;
+
+    *buf = NULL;
+    stream = libbalsa_message_body_get_stream(body);
+    if (!stream)
+        return -1;
+
+    array = g_byte_array_new();
+    stream_mem = g_mime_stream_mem_new_with_byte_array(array);
+    g_mime_stream_mem_set_owner(GMIME_STREAM_MEM(stream_mem), FALSE);
+    len = g_mime_stream_write_to_stream(stream, stream_mem);
+    g_object_unref(stream_mem);
+    g_object_unref(stream);
+
+    if (len >= 0) {
+        len = array->len;
+	/* NULL-terminate, in case it is used as a string. */
+	g_byte_array_append(array, "", 1);
+        *buf = array->data;
+        g_byte_array_free(array, FALSE);
+    } else
+        g_byte_array_free(array, TRUE);
+
+    return len;
+}
+
+GdkPixbuf *
+libbalsa_message_body_get_pixbuf(LibBalsaMessageBody * body, GError ** err)
+{
+    GdkPixbufLoader *loader;
+    GMimeStream *stream;
+    gssize count;
+    gboolean ok = TRUE;
+    gchar buf[4096];
+    GdkPixbuf *pixbuf = NULL;
+
+    stream = libbalsa_message_body_get_stream(body);
+    if (!stream)
+        return pixbuf;
+
+    loader = gdk_pixbuf_loader_new();
+    while ((count = g_mime_stream_read(stream, buf, sizeof(buf))) > 0) {
+        if (!gdk_pixbuf_loader_write(loader, buf, count, err)) {
+            ok = FALSE;
+            break;
+        }
     }
-    if (retval && g_mime_stream_flush(stream) == -1)
+    g_object_unref(stream);
+
+    if (ok && !gdk_pixbuf_loader_close(loader, err))
+	ok = FALSE;
+
+    if (ok) {
+        pixbuf = gdk_pixbuf_loader_get_pixbuf(loader);
+	g_object_ref(pixbuf);
+    }
+    g_object_unref(loader);
+
+    return pixbuf;
+}
+
+gboolean
+libbalsa_message_body_save_fd(LibBalsaMessageBody * body, int fd)
+{
+    GMimeStream *stream, *stream_fs;
+    gboolean retval = TRUE;
+
+    stream = libbalsa_message_body_get_stream(body);
+    stream_fs = g_mime_stream_fs_new(fd);
+
+    if (g_mime_stream_write_to_stream(stream, stream_fs) < 0)
 	retval = FALSE;
     g_object_unref(stream);
+    g_object_unref(stream_fs);
 
     return retval;
 }
@@ -409,6 +515,12 @@ gchar *
 libbalsa_message_body_get_mime_type(LibBalsaMessageBody * body)
 {
     gchar *res, *tmp;
+
+    g_return_val_if_fail(body != NULL, NULL);
+
+    if (!body->content_type)
+	return g_strdup("text/plain");
+
     tmp = strchr(body->content_type, ';');
     res = g_ascii_strdown(body->content_type,
                           tmp ? tmp-body->content_type : -1);
@@ -444,6 +556,8 @@ libbalsa_message_body_is_inline(LibBalsaMessageBody * body)
 	 * message/rfc822; in either case, we want to in-line the part.
 	 */
         return (body->content_type == NULL
+                || g_ascii_strcasecmp(body->content_type,
+                                      "message/rfc822") == 0
                 || g_ascii_strcasecmp(body->content_type,
                                       "text/plain") == 0);
 
@@ -510,6 +624,9 @@ libbalsa_message_body_get_by_id(LibBalsaMessageBody * body,
 
 	if (bodyid && strcmp(id, bodyid) == 0)
 	    return body;
+    } else {
+        if(body->content_id && strcmp(id, body->content_id) == 0)
+            return body;
     }
 
     if ((res = libbalsa_message_body_get_by_id(body->parts, id)) != NULL)
