@@ -33,6 +33,7 @@
 #define _XOPEN_SOURCE
 /* extensions  needed additonally on Solaris for strptime */
 #define __EXTENSIONS__
+#include <stdlib.h>
 #include <stdio.h>
 /* yellow dog has crappy libc and needs pthread.h to be included here */
 #ifdef BALSA_USE_THREADS
@@ -40,13 +41,14 @@
 #endif
 #include <string.h>
 #include <time.h>
+
+#include <libgnome/libgnome.h>
+
 #include <sys/types.h>
 #include "filter-file.h"
 #include "filter-private.h"
 #include "filter-funcs.h"
 #include "mailbox-filter.h"
-#include <gnome.h>
-
 
 /* Load the header of a filter filter (you have to separately load the
  * associated conditions) Filter is marked as invalid Position
