@@ -32,7 +32,7 @@
 #include <ctype.h>
 #include <string.h>
 
-#include <gnome.h>
+#include <libgnome/libgnome.h>
 
 #include "libbalsa.h"
 #include "libbalsa_private.h"
@@ -282,7 +282,7 @@ libbalsa_filter_mailbox_messages(LibBalsaFilter * filt,
 {
     gboolean result=FALSE;
     LibBalsaMailbox *mbox;
-    LibBalsaMailboxSearchIter *iter_view;
+    GError *err = NULL;
 
     if (msgnos->len == 0)
 	return FALSE;
@@ -302,18 +302,20 @@ libbalsa_filter_mailbox_messages(LibBalsaFilter * filt,
 	    libbalsa_information(LIBBALSA_INFORMATION_ERROR,
 				 _("Bad mailbox name for filter: %s"),
 				 filt->name);
-	else if (!libbalsa_mailbox_messages_copy(mailbox, msgnos, mbox))
+	else if (!libbalsa_mailbox_messages_copy(mailbox, msgnos, mbox, &err))
 	    libbalsa_information(LIBBALSA_INFORMATION_ERROR,
-				 _("Error when copying messages"));
+				 _("Error when copying messages: %s"),
+                                 err ? err->message : "?");
 	else if (mbox == filters_trash_mbox)
 	    result = TRUE;
 	break;
     case FILTER_TRASH:
 	if (!filters_trash_mbox ||
 	    !libbalsa_mailbox_messages_move(mailbox, msgnos,
-					    filters_trash_mbox))
+					    filters_trash_mbox, &err))
 	    libbalsa_information(LIBBALSA_INFORMATION_ERROR,
-				 _("Error when trashing messages"));
+				 _("Error when trashing messages: %s"),
+                                 err ? err->message : "?");
 	else
 	    result = TRUE;
 	break;
@@ -323,9 +325,10 @@ libbalsa_filter_mailbox_messages(LibBalsaFilter * filt,
 	    libbalsa_information(LIBBALSA_INFORMATION_ERROR,
 				 _("Bad mailbox name for filter: %s"),
 				 filt->name);
-	else if (!libbalsa_mailbox_messages_move(mailbox, msgnos, mbox))
+	else if (!libbalsa_mailbox_messages_move(mailbox, msgnos, mbox, &err))
 	    libbalsa_information(LIBBALSA_INFORMATION_ERROR,
-				 _("Error when moving messages"));
+				 _("Error when moving messages: %s"),
+                                 err ? err->message : "?");
 	else if (mbox == filters_trash_mbox)
 	    result = TRUE;
 	break;
@@ -339,18 +342,7 @@ libbalsa_filter_mailbox_messages(LibBalsaFilter * filt,
 	/* Nothing to do */
 	break;
     }
-
-    iter_view = libbalsa_mailbox_search_iter_view(mailbox);
-    if (iter_view) {
-	guint i;
-
-	for (i = 0; i < msgnos->len; i++)
-	    libbalsa_mailbox_msgno_filt_check(mailbox,
-					      g_array_index(msgnos, guint, i),
-					      iter_view);
-	libbalsa_mailbox_search_iter_free(iter_view);
-    }
-
+    g_clear_error(&err);
     libbalsa_unlock_mailbox(mailbox);
 
     return result;
@@ -404,3 +396,56 @@ libbalsa_condition_can_match(LibBalsaCondition * cond,
 	return TRUE;
     }
 }
+
+/* Check whether a condition looks only at flags; if it does, test
+ * whether the given message's flags match it, and return the result in
+ * *match; used by mailbox backends to decide when the full
+ * LibBalsaMessage is needed. */
+gboolean
+libbalsa_condition_is_flag_only(LibBalsaCondition * cond,
+                                LibBalsaMailbox * mailbox,
+                                guint msgno,
+				gboolean * match)
+{
+    gboolean retval;
+    gboolean left_match, right_match;
+
+    switch (cond->type) {
+    case CONDITION_FLAG:
+        if (match)
+            *match =
+                libbalsa_mailbox_msgno_has_flags(mailbox, msgno,
+                                                 cond->match.flags, 0);
+        retval = TRUE;
+        break;
+    case CONDITION_AND:
+        retval =
+            libbalsa_condition_is_flag_only(cond->match.andor.left,
+                                            mailbox, msgno,
+					    &left_match)
+            && libbalsa_condition_is_flag_only(cond->match.andor.right,
+                                               mailbox, msgno,
+                                               &right_match);
+        if (retval && match)
+            *match = left_match && right_match;
+        break;
+    case CONDITION_OR:
+        retval =
+            libbalsa_condition_is_flag_only(cond->match.andor.left,
+                                            mailbox, msgno,
+					    &left_match)
+            && libbalsa_condition_is_flag_only(cond->match.andor.right,
+                                               mailbox, msgno,
+                                               &right_match);
+        if (retval && match)
+            *match = left_match || right_match;
+        break;
+    default:
+        return FALSE;
+    }
+
+    if (retval && match && cond->negate)
+        *match = !*match;
+
+    return retval;
+}
