@@ -13,8 +13,7 @@
 #include "imap-server.h"
 #include "imap-commands.h"
 
-#include <libgnome/gnome-config.h>
-#include <libgnome/gnome-i18n.h> 
+#include <libgnome/libgnome.h> 
 
 #ifdef USE_TLS
 #define REQ_SSL(s) (LIBBALSA_SERVER(s)->use_ssl)
@@ -31,15 +30,16 @@ struct LibBalsaImapServer_ {
     gchar *key;
     guint max_connections;
     gboolean offline_mode;
-    gboolean persistent_cache; /* if TRUE, messages will be cached in
-                                  $HOME and preserved between
-                                  sessions. If FALSE, messages will be
-                                  kept in /tmp and cleaned on exit. */
     
     GMutex *lock; /* protects the following members */
     guint used_connections;
     GList *used_handles;
     GList *free_handles;
+    unsigned persistent_cache:1; /* if TRUE, messages will be cached in
+                                    $HOME and preserved between
+                                    sessions. If FALSE, messages will be
+                                    kept in /tmp and cleaned on exit. */
+    unsigned has_fetch_bug:1;
 };
 
 typedef struct LibBalsaImapServerClass_ {
@@ -124,13 +124,15 @@ static void libbalsa_imap_server_set_username(LibBalsaServer * server,
 static void libbalsa_imap_server_set_host(LibBalsaServer * server,
                       const gchar * host, gboolean use_ssl)
 {
-    LibBalsaImapServer *imap_server = LIBBALSA_IMAP_SERVER(server);
-    G_LOCK(imap_servers);
-    g_hash_table_steal(imap_servers, imap_server->key);
-    g_free(imap_server->key);
-    imap_server->key = g_strdup_printf("%s@%s", server->user, host);
-    g_hash_table_insert(imap_servers, imap_server->key, imap_server);
-    G_UNLOCK(imap_servers);
+    if(server->user && host) { /* we have been initialized... */
+        LibBalsaImapServer *imap_server = LIBBALSA_IMAP_SERVER(server);
+        G_LOCK(imap_servers);
+        g_hash_table_steal(imap_servers, imap_server->key);
+        g_free(imap_server->key);
+        imap_server->key = g_strdup_printf("%s@%s", server->user, host);
+        g_hash_table_insert(imap_servers, imap_server->key, imap_server);
+        G_UNLOCK(imap_servers);
+    }
     (parent_class)->set_host(server, host, use_ssl);
 }
 static void
@@ -283,6 +285,7 @@ lb_imap_server_info_new(LibBalsaServer *server)
      * libbalsa_server_user_cb(). */
 
     handle = imap_mbox_handle_new();
+    imap_handle_set_timeout(handle, 30000); /* wait 30 seconds for packets */
     info = g_new0(struct handle_info, 1);
     info->handle = handle;
     imap_handle_set_monitorcb(handle, monitor_cb, info);
@@ -457,7 +460,9 @@ libbalsa_imap_server_new_from_config(void)
     conn_limit = gnome_config_get_int_with_default("ConnectionLimit", &d);
     if(!d) imap_server->max_connections = conn_limit;
     d1 = gnome_config_get_bool_with_default("PersistentCache", &d);
-    if(!d) imap_server->persistent_cache = d1;
+    if(!d) imap_server->persistent_cache = !!d1;
+    d1 = gnome_config_get_bool_with_default("HasFetchBug", &d);
+    if(!d) imap_server->has_fetch_bug = !!d1;
     if (!server->passwd) {
         server->remember_passwd = gnome_config_get_bool("RememberPasswd=false");
         if(server->remember_passwd)
@@ -482,6 +487,29 @@ libbalsa_imap_server_save_config(LibBalsaImapServer *server)
     libbalsa_server_save_config(LIBBALSA_SERVER(server));
     gnome_config_set_int("ConnectionLimit", server->max_connections);
     gnome_config_set_bool("PersistentCache", server->persistent_cache);
+    gnome_config_set_bool("HasFetchBug", server->has_fetch_bug);
+}
+
+/* handle_connection_error() releases handle_info data, clears password
+   and sets err apriopriately */
+static void
+handle_connection_error(int rc, struct handle_info *info,
+                        LibBalsaServer *server, GError **err)
+{
+    gchar *msg = imap_mbox_handle_get_last_msg(info->handle);
+    if(rc == IMAP_AUTH_FAILURE)
+        libbalsa_server_set_password(server, NULL);
+    if(rc != IMAP_CONNECT_FAILED) {
+        g_set_error(err, LIBBALSA_MAILBOX_ERROR,
+                    LIBBALSA_MAILBOX_AUTH_ERROR,
+                    _("Cannot connect to the server: %s"), msg);
+    } else
+        g_set_error(err, LIBBALSA_MAILBOX_ERROR,
+                    LIBBALSA_MAILBOX_NETWORK_ERROR,
+                    _("Cannot connect to the server"));
+    
+    g_free(msg);
+    lb_imap_server_info_free(info);
 }
 
 /**
@@ -540,16 +568,7 @@ libbalsa_imap_server_get_handle(LibBalsaImapServer *imap_server, GError **err)
             rc=imap_mbox_handle_connect(info->handle, server->host,
                                         REQ_SSL(server));
             if(rc != IMAP_SUCCESS) {
-                gchar *msg = imap_mbox_handle_get_last_msg(info->handle);
-                if(rc == IMAP_AUTH_FAILURE)
-                    libbalsa_server_set_password(server, NULL);
-                g_set_error(err, LIBBALSA_MAILBOX_ERROR,
-                            rc != IMAP_CONNECT_FAILED
-                            ? LIBBALSA_MAILBOX_AUTH_ERROR :
-                            LIBBALSA_MAILBOX_NETWORK_ERROR,
-                            _("Cannot connect to the server: %s"), msg);
-                g_free(msg);
-                lb_imap_server_info_free(info);
+                handle_connection_error(rc, info, server, err);
                 g_mutex_unlock(imap_server->lock);
                 return NULL;
             }
@@ -645,14 +664,7 @@ libbalsa_imap_server_get_handle_with_user(LibBalsaImapServer *imap_server,
         rc=imap_mbox_handle_connect(info->handle, server->host,
                                     REQ_SSL(server));
         if(rc != IMAP_SUCCESS) {
-            gchar *msg = imap_mbox_handle_get_last_msg(info->handle);
-            if(rc == IMAP_AUTH_FAILURE)
-                libbalsa_server_set_password(server, NULL);
-            g_set_error(err, LIBBALSA_MAILBOX_ERROR,
-                        LIBBALSA_MAILBOX_OPEN_ERROR,
-                    _("Cannot connect to the server: %s"), msg);
-            g_free(msg);
-            lb_imap_server_info_free(info);
+            handle_connection_error(rc, info, server, err);
             g_mutex_unlock(imap_server->lock);
             return NULL;
         }
@@ -728,7 +740,7 @@ void
 libbalsa_imap_server_enable_persistent_cache(LibBalsaImapServer *server,
                                              gboolean enable)
 {
-    server->persistent_cache = enable;
+    server->persistent_cache = !!enable;
 }
 gboolean
 libbalsa_imap_server_has_persistent_cache(LibBalsaImapServer *srv)
@@ -827,3 +839,17 @@ void libbalsa_imap_server_set_offline_mode(LibBalsaImapServer *server,
     if (offline)
         libbalsa_imap_server_force_disconnect(server);
 }
+
+void
+libbalsa_imap_server_set_bug(LibBalsaImapServer *server,
+                             LibBalsaImapServerBug bug, gboolean hasp)
+{
+    server->has_fetch_bug = !! hasp;
+}
+
+gboolean
+libbalsa_imap_server_has_bug(LibBalsaImapServer *server,
+                             LibBalsaImapServerBug bug)
+{
+    return server->has_fetch_bug;
+}
