@@ -96,6 +96,9 @@ static ImapResult imap_mbox_connect(ImapMboxHandle* handle);
 
 static ImapResponse ir_handle_response(ImapMboxHandle *h);
 
+static ImapAddress* imap_address_from_string(const gchar *string);
+static gchar*       imap_address_to_string(const ImapAddress *addr);
+
 static GType
 imap_mbox_handle_get_type()
 {
@@ -125,12 +128,14 @@ imap_mbox_handle_init(ImapMboxHandle *handle)
 {
   handle->host   = NULL;
   handle->mbox   = NULL;
+  handle->timeout = -1;
   handle->state  = IMHS_DISCONNECTED;
   handle->has_capabilities = FALSE;
   handle->exists = 0;
   handle->recent = 0;
   handle->last_msg = NULL;
   handle->msg_cache = NULL;
+  handle->flag_cache=  g_array_new(FALSE, TRUE, sizeof(ImapFlagCache));
   handle->doing_logout = FALSE;
 #ifdef USE_TLS
   handle->using_tls = 0;
@@ -230,6 +235,14 @@ imap_handle_set_flagscb(ImapMboxHandle* h, ImapFlagsCb cb, void* arg)
   h->flags_arg = arg;
 }
 
+void
+imap_handle_set_timeout(ImapMboxHandle *h, int milliseconds)
+{
+  h->timeout = milliseconds;
+  if(h->sio)
+    sio_set_timeout(h->sio, milliseconds);
+}
+
 int imap_mbox_is_disconnected (ImapMboxHandle *h)
 { return IMAP_MBOX_IS_DISCONNECTED(h); }
 int imap_mbox_is_connected    (ImapMboxHandle *h)
@@ -273,6 +286,7 @@ imap_mbox_resize_cache(ImapMboxHandle *h, unsigned new_size)
     }
   }
   h->msg_cache = g_realloc(h->msg_cache, new_size*sizeof(ImapMessage*));
+  g_array_set_size(h->flag_cache, new_size);
   for(i=h->exists; i<new_size; i++) 
     h->msg_cache[i] = NULL;
   h->exists = new_size;
@@ -384,6 +398,7 @@ imap_mbox_connect(ImapMboxHandle* handle)
 
   /* reset some handle status */
   handle->has_capabilities = FALSE;
+  handle->can_fetch_body = TRUE;
   if(handle->sio) {
     sio_detach(handle->sio);
     handle->sio = NULL;
@@ -403,6 +418,8 @@ imap_mbox_connect(ImapMboxHandle* handle)
     close(handle->sd);
     return IMAP_NOMEM;
   }
+  if(handle->timeout>0)
+    sio_set_timeout(handle->sio, handle->timeout);
 #ifdef USE_TLS
   if(handle->over_ssl) {
     SSL *ssl = imap_create_ssl();
@@ -428,7 +445,8 @@ imap_mbox_connect(ImapMboxHandle* handle)
     handle->state = IMHS_DISCONNECTED;
     return IMAP_PROTOCOL_ERROR;
   }
-
+  handle->can_fetch_body = 
+    (strncmp(handle->last_msg, "Microsoft Exchange", 18) != 0);
 #if defined(USE_TLS)
   if(handle->over_ssl)
     resp = IMR_OK; /* secured already with SSL */
@@ -536,6 +554,11 @@ imap_mbox_handle_get_validity(ImapMboxHandle* handle)
 {
   return handle->uidval;
 }
+unsigned
+imap_mbox_handle_get_uidnext(ImapMboxHandle* handle)
+{
+  return handle->uidnext;
+}
 
 static void
 get_delim(ImapMboxHandle* handle, int delim, ImapMboxFlags flags,
@@ -595,6 +618,8 @@ imap_mbox_handle_finalize(GObject* gobject)
 
   mbox_view_dispose(&handle->mbox_view);
   imap_mbox_resize_cache(handle, 0);
+  g_free(handle->msg_cache); handle->msg_cache = NULL;
+  g_array_free(handle->flag_cache, TRUE);
 
   G_OBJECT_CLASS(parent_class)->finalize(gobject);  
 }
@@ -835,6 +860,7 @@ imap_body_free(ImapBody* body)
   g_free(body->media_basic_name);
   g_free(body->media_subtype);
   g_free(body->desc);
+  g_free(body->content_id);
   g_free(body->content_dsp_other);
   g_free(body->content_uri);
   g_hash_table_destroy(body->params);
@@ -995,7 +1021,7 @@ void
 imap_body_set_id(ImapBody *body, char *id)
 {
   if(id) printf("part ID='%s'\n", id);
-  g_free(id);
+  body->content_id = id;
 }
 /* ================ END OF BODY STRUCTURE FUNCTIONS ==================== */
 
@@ -1022,7 +1048,110 @@ imap_message_free(ImapMessage *msg)
   g_free(msg);
 }
 
+void
+imap_mbox_handle_msg_deserialize(ImapMboxHandle *h, unsigned msgno,
+                                 void *data)
+{
+  if(msgno>=1 && msgno <=h->exists && !h->msg_cache[msgno-1])
+    h->msg_cache[msgno-1] = imap_message_deserialize(data);
+}
+/* serialize message itself and the envelope - but not the body */
+struct ImapMsgEnvSerialized {
+  ssize_t total_size; /* for checksumming */
+  /* Message */
+  ImapUID      uid;
+  ImapMsgFlags flags;
+  ImapDate     internal_date; /* delivery date */
+  ImapDate     date;          /* sending date */
+  int rfc822size;
+  ImapFetchType available_headers;
+  gchar fetched_headers_first_char;
+};
 
+void*
+imap_message_serialize(ImapMessage *imsg)
+{
+#define SER_STR_CNT 9
+  gchar *strings[SER_STR_CNT];
+  int    lengths[SER_STR_CNT], i;
+  ssize_t tot_size;
+  gchar *ptr;
+  struct ImapMsgEnvSerialized *imes;
+  if(!imsg->envelope) /* envelope is required */
+    return NULL; 
+  strings[0] = imsg->fetched_header_fields;
+  strings[1] = imsg->envelope->subject;
+  strings[2] = imap_address_to_string(imsg->envelope->from);
+  strings[3] = imap_address_to_string(imsg->envelope->sender);
+  strings[4] = imap_address_to_string(imsg->envelope->replyto);
+  strings[5] = imap_address_to_string(imsg->envelope->to);
+  strings[6] = imap_address_to_string(imsg->envelope->cc);
+  strings[7] = imap_address_to_string(imsg->envelope->bcc);
+  strings[8] = imsg->envelope->in_reply_to;
+  tot_size = sizeof(struct ImapMsgEnvSerialized)-1;
+  for(i=0; i<SER_STR_CNT; i++) {
+    lengths[i] = strings[i] ? strlen(strings[i]) : 0;
+    tot_size += 1 + lengths[i];
+  }
+  imes = g_malloc(tot_size);
+  imes->total_size = tot_size;
+  /* Message */
+  imes->uid           = imsg->uid;
+  imes->flags         = imsg->flags;
+  imes->internal_date = imsg->internal_date; /* delivery date */
+  imes->rfc822size    = imsg->rfc822size;
+  imes->available_headers = imsg->available_headers;
+  /* Envelope */
+  imes->date = imsg->envelope->date;
+  ptr = &imes->fetched_headers_first_char;
+  for(i=0; i<SER_STR_CNT; i++) {
+    if(strings[i]) strcpy(ptr, strings[i]);
+    ptr += lengths[i];
+    *ptr++ = '\0';
+  }
+  for(i=2; i<=7; i++)
+    g_free(strings[i]);
+  /* printf("Serialization offset: %d (tot size %d may include alignment)\n",
+     ptr-(gchar*)imes, tot_size); */
+  return imes;
+}
+/* obs - no data checking is done as it should be if the data were read
+   from a file or other external source. */
+ImapMessage*
+imap_message_deserialize(void *data)
+{
+  struct ImapMsgEnvSerialized *imes = (struct ImapMsgEnvSerialized*)data;
+  ImapMessage* imsg = imap_message_new();
+  gchar *ptr;
+
+  imsg->envelope = imap_envelope_new();
+  imsg->uid = imes->uid;
+  imsg->flags = imes->flags;
+  imsg->internal_date = imes->internal_date; /* delivery date */
+  imsg->rfc822size = imes->rfc822size;
+  imsg->available_headers = imes->available_headers;
+  /* Envelope */
+  imsg->envelope->date = imes->date;
+  ptr = &imes->fetched_headers_first_char;
+  imsg->fetched_header_fields = *ptr ? g_strdup(ptr) : NULL;
+  ptr += strlen(ptr) + 1;
+  imsg->envelope->subject = *ptr ? g_strdup(ptr) : NULL;
+  ptr += strlen(ptr) + 1;
+  imsg->envelope->from    = *ptr ? imap_address_from_string(ptr) : NULL;
+  ptr += strlen(ptr) + 1;
+  imsg->envelope->sender  = *ptr ? imap_address_from_string(ptr) : NULL;
+  ptr += strlen(ptr) + 1;
+  imsg->envelope->replyto = *ptr ? imap_address_from_string(ptr) : NULL;
+  ptr += strlen(ptr) + 1;
+  imsg->envelope->to      = *ptr ? imap_address_from_string(ptr) : NULL;
+  ptr += strlen(ptr) + 1;
+  imsg->envelope->cc      = *ptr ? imap_address_from_string(ptr) : NULL;
+  ptr += strlen(ptr) + 1;
+  imsg->envelope->bcc     = *ptr ? imap_address_from_string(ptr) : NULL;
+  ptr += strlen(ptr) + 1;
+  imsg->envelope->in_reply_to = *ptr ? g_strdup(ptr) : NULL;
+  return imsg;
+}
 
 /* =================================================================== */
 /*                Imap command processing routines                     */
@@ -1529,6 +1658,7 @@ ir_list_lsub(ImapMboxHandle *h, ImapHandleSignal signal)
     if(delim            != 'N' ||
        sio_getc(h->sio) != 'I' ||
        sio_getc(h->sio) != 'L') return IMR_PROTOCOL;
+    delim = '\0'; /* NIL */
   }
   if(sio_getc(h->sio) != ' ') return IMR_PROTOCOL;
   /* mailbox */
@@ -1629,6 +1759,7 @@ ir_expunge(ImapMboxHandle *h, unsigned seqno)
   g_signal_emit(G_OBJECT(h), imap_mbox_handle_signals[EXPUNGE_NOTIFY],
 		0, seqno);
   
+  g_array_remove_index(h->flag_cache, seqno-1);
   if(h->msg_cache[seqno-1] != NULL)
     imap_message_free(h->msg_cache[seqno-1]);
   while(seqno<h->exists) {
@@ -1653,6 +1784,7 @@ ir_msg_att_flags(ImapMboxHandle *h, int c, unsigned seqno)
 {
   unsigned i;
   ImapMessage *msg = h->msg_cache[seqno-1];
+  ImapFlagCache *flags;
 
   if(sio_getc(h->sio) != '(') return IMR_PROTOCOL;
   msg->flags = 0;
@@ -1667,6 +1799,10 @@ ir_msg_att_flags(ImapMboxHandle *h, int c, unsigned seqno)
       }
   } while(c!=-1 && c != ')');
 
+  flags = &g_array_index(h->flag_cache, ImapFlagCache, seqno-1);
+  flags->flag_values = msg->flags;
+  flags->known_flags = ~0; /* all of them are known */
+
   if(h->flags_cb)
     imap_handle_add_task(h, flags_tasklet, GUINT_TO_POINTER(seqno));
   return IMR_OK;
@@ -1694,6 +1830,75 @@ imap_address_free(ImapAddress* addr)
   }
 }
 
+static gchar*
+get_quoted_string(const gchar *source, gchar const **endpos)
+{
+  GString *s;
+  if(*source == '\0') {
+    *endpos = source;
+    return NULL;
+  }
+  if(*source != '"') {
+    *endpos = source+1;
+    return NULL;
+  }
+  source++;
+  s = g_string_new("");
+  for(;*source && *source != '"'; source++) {
+    if(source[0] == '\\' && source[1])
+      source++;
+    g_string_append_c(s, *source);
+  }
+  *endpos = *source ? source+1 : source;
+  return g_string_free(s, FALSE);
+}
+static ImapAddress*
+imap_address_from_string(const gchar *string)
+{
+  const gchar *t, *t1;
+  gchar *comment, *mailbox;
+  ImapAddress *res = NULL;
+
+  comment = get_quoted_string(string, &t);
+  if(*t == ' ') {
+    mailbox = get_quoted_string(t+1, &t1);
+    if(comment || mailbox) {
+      res = imap_address_new(comment, mailbox);
+      if(*t1 == ' ')
+        res->next = imap_address_from_string(t1+1);
+    }
+  } else g_free(comment);
+  return res;
+}
+
+static gchar*
+imap_address_to_string(const ImapAddress *addr)
+{
+  GString *res = g_string_sized_new(4);
+  gchar *p;
+  for(; addr; addr = addr->next) {
+    if(addr->name) {
+      g_string_append_c(res, '"');
+      for(p=addr->name; *p; p++) {
+        if(*p == '\\' || *p == '"') g_string_append_c(res, '\\');
+        g_string_append_c(res, *p);
+      }
+      g_string_append_c(res, '"');
+    } else g_string_append_c(res, 'N');
+    g_string_append_c(res, ' ');
+    if(addr->addr_spec) {
+      g_string_append_c(res, '"');
+      for(p=addr->addr_spec; *p; p++) {
+        if(*p == '\\' || *p == '"') g_string_append_c(res, '\\');
+        g_string_append_c(res, *p);
+      }
+      g_string_append_c(res, '"');
+    } else g_string_append_c(res, 'N');
+    g_string_append_c(res, ' ');
+  }
+  g_string_append(res, "N N");
+  return g_string_free(res, FALSE);
+}
 /* imap_get_address: returns null if no beginning of address is found
    (eg., when end of list is found instead).
 */
@@ -2473,7 +2678,7 @@ ir_body (struct siobuf *sio, int c, ImapBody * body,
 static ImapResponse
 ir_body_section(struct siobuf *sio, ImapFetchBodyCb body_cb, void *arg)
 {
-  char buf[12], *str;
+  char buf[80], *str;
   int c = imap_get_atom(sio, buf, sizeof(buf));
   if(c != ']') { puts("] expected"); return IMR_PROTOCOL; }
   if(sio_getc(sio) != ' ') { puts("space expected"); return IMR_PROTOCOL;}
@@ -2522,12 +2727,18 @@ ir_msg_att_body(ImapMboxHandle *h, int c, unsigned seqno)
       break;
     }
     c = imap_get_atom(h->sio, buf, sizeof buf);
-    if (c == ' ' && 
-        (g_ascii_strcasecmp(buf, "HEADER.FIELDS") == 0 ||
-         g_ascii_strcasecmp(buf, "HEADER.FIELDS.NOT") == 0))
-      rc = ir_body_header_fields(h->sio, msg);
-    else
-      rc = IMR_PROTOCOL;
+    if (c == ']' &&
+        g_ascii_strcasecmp(buf, "HEADER") == 0) {
+      sio_ungetc (h->sio); /* put the ']' back */
+      rc = ir_body_section(h->sio, h->body_cb, h->body_arg);
+    } else {
+      if (c == ' ' && 
+          (g_ascii_strcasecmp(buf, "HEADER.FIELDS") == 0 ||
+           g_ascii_strcasecmp(buf, "HEADER.FIELDS.NOT") == 0))
+        rc = ir_body_header_fields(h->sio, msg);
+      else
+        rc = IMR_PROTOCOL;
+    }
     break;
   case ' ':
     rc = ir_body(h->sio, sio_getc(h->sio),
