@@ -74,6 +74,9 @@ typedef enum
   IMCAP_CHILDREN,               /* RFC 3348 */
   IMCAP_LITERAL,                /* RFC 2088 */
   IMCAP_SASLIR,                 /* imap-sasl-initial-response: SASL-IR */
+  IMCAP_FETCHBODY,              /* basic imap implemented correctly by
+                                 * most imap servers but not all. We
+                                 * have to detect that. */
   IMCAP_MAX
 } ImapCapability;
 
@@ -106,6 +109,7 @@ void imap_handle_set_monitorcb(ImapMboxHandle* h, ImapMonitorCb cb, void*);
 void imap_handle_set_infocb(ImapMboxHandle* h, ImapInfoCb cb, void*);
 void imap_handle_set_usercb(ImapMboxHandle* h, ImapUserCb cb, void*);
 void imap_handle_set_flagscb(ImapMboxHandle* h, ImapFlagsCb cb, void*);
+void imap_handle_set_timeout(ImapMboxHandle *, int milliseconds);
 
 ImapResult imap_mbox_handle_connect(ImapMboxHandle* r, const char *hst, 
                                     int over_ssl);
@@ -117,6 +121,7 @@ ImapTlsMode imap_handle_set_tls_mode(ImapMboxHandle *h, ImapTlsMode option);
 int      imap_mbox_handle_can_do(ImapMboxHandle* handle, ImapCapability cap);
 unsigned imap_mbox_handle_get_exists(ImapMboxHandle* handle);
 unsigned imap_mbox_handle_get_validity(ImapMboxHandle* handle);
+unsigned imap_mbox_handle_get_uidnext(ImapMboxHandle* handle);
 int      imap_mbox_handle_get_delim(ImapMboxHandle* handle,
                                     const char *namespace);
 char* imap_mbox_handle_get_last_msg(ImapMboxHandle *handle);
@@ -142,8 +147,6 @@ unsigned imap_mbox_find_next(ImapMboxHandle* handle, unsigned start,
                              const char *search_str);
 ImapResponse imap_mbox_find_all(ImapMboxHandle *h, const char *search_str,
                                 unsigned *msgcnt, unsigned**msgs);
-ImapResponse imap_mbox_find_unseen(ImapMboxHandle * h, unsigned *msgcnt,
-				   unsigned **msgs);
 
 unsigned imap_mbox_set_view(ImapMboxHandle *h, ImapMsgFlag f, gboolean state);
 unsigned imap_mbox_set_sort(ImapMboxHandle *h, ImapSortKey isr, 
