@@ -25,6 +25,7 @@
 
 #include <gdk/gdk.h>
 #include <gmime/gmime.h>
+#include "config.h"
 
 #ifdef BALSA_USE_THREADS
 #include <pthread.h>
@@ -132,6 +133,19 @@ typedef enum {
     LB_MAILBOX_STATE_CLOSING
 } LibBalsaMailboxState;
 
+#ifdef HAVE_GPGME
+typedef enum {
+    LB_MAILBOX_CHK_CRYPT_NEVER,     /* never auto decrypt/signature check */
+    LB_MAILBOX_CHK_CRYPT_MAYBE,     /* auto decrypt/signature check if possible */
+    LB_MAILBOX_CHK_CRYPT_ALWAYS     /* always auto decrypt/signature check */
+} LibBalsaChkCryptoMode;
+#endif
+
+enum LibBalsaMailboxCapability {
+    LIBBALSA_MAILBOX_CAN_SORT,
+    LIBBALSA_MAILBOX_CAN_THREAD
+};
+
 /*
  * structures
  */
@@ -156,6 +170,10 @@ struct _LibBalsaMailboxView {
     unsigned open:1;
     unsigned in_sync:1;		/* view is in sync with config */
     unsigned frozen:1;		/* don't update view if set    */
+
+#ifdef HAVE_GPGME
+    LibBalsaChkCryptoMode gpg_chk_mode;
+#endif
 };
 
 struct _LibBalsaMailbox {
@@ -222,9 +240,9 @@ struct _LibBalsaMailboxClass {
     void (*close_mailbox) (LibBalsaMailbox * mailbox, gboolean expunge);
     LibBalsaMessage *(*get_message) (LibBalsaMailbox * mailbox, guint msgno);
     void (*prepare_threading)(LibBalsaMailbox *mailbox, guint lo, guint hi);
-    void (*fetch_message_structure)(LibBalsaMailbox *mailbox,
-                                    LibBalsaMessage * message,
-                                    LibBalsaFetchFlag flags);
+    gboolean (*fetch_message_structure)(LibBalsaMailbox *mailbox,
+                                        LibBalsaMessage * message,
+                                        LibBalsaFetchFlag flags);
     void (*fetch_headers)(LibBalsaMailbox *mailbox,
                           LibBalsaMessage * message);
     void (*release_message) (LibBalsaMailbox * mailbox,
@@ -245,16 +263,21 @@ struct _LibBalsaMailboxClass {
     void (*save_config) (LibBalsaMailbox * mailbox, const gchar * prefix);
     void (*load_config) (LibBalsaMailbox * mailbox, const gchar * prefix);
     gboolean (*sync) (LibBalsaMailbox * mailbox, gboolean expunge);
-    int (*add_message) (LibBalsaMailbox * mailbox, LibBalsaMessage * message );
-    void (*change_message_flags) (LibBalsaMailbox * mailbox, guint msgno,
-                                  LibBalsaMessageFlag set,
-                                  LibBalsaMessageFlag clear);
+    int (*add_message) (LibBalsaMailbox * mailbox, LibBalsaMessage * message,
+                        GError **err);
     gboolean (*messages_change_flags) (LibBalsaMailbox * mailbox,
 				       GArray *msgnos,
 				       LibBalsaMessageFlag set,
 				       LibBalsaMessageFlag clear);
     gboolean (*messages_copy) (LibBalsaMailbox * mailbox, GArray *msgnos,
-			       LibBalsaMailbox * dest);
+			       LibBalsaMailbox * dest, GError **err);
+    /* Test message flags */
+    gboolean(*msgno_has_flags) (LibBalsaMailbox * mailbox, guint msgno,
+                                LibBalsaMessageFlag set,
+                                LibBalsaMessageFlag unset);
+
+    gboolean (*can_do) (LibBalsaMailbox *mailbox,
+                        enum LibBalsaMailboxCapability cap);
     void (*set_threading) (LibBalsaMailbox * mailbox,
 			   LibBalsaMailboxThreadingType thread_type);
     void (*update_view_filter) (LibBalsaMailbox * mailbox,
@@ -307,9 +330,11 @@ void libbalsa_mailbox_prepare_threading(LibBalsaMailbox *mailbox,
     message structure for given message. It can also fetch all RFC822
     headers of the message.
 */
-void libbalsa_mailbox_fetch_message_structure(LibBalsaMailbox *mailbox,
-					      LibBalsaMessage *message,
-					      LibBalsaFetchFlag flags);
+gboolean libbalsa_mailbox_fetch_message_structure(LibBalsaMailbox *
+                                                  mailbox,
+                                                  LibBalsaMessage *
+                                                  message,
+                                                  LibBalsaFetchFlag flags);
 
 /** libbalsa_mailbox_release_message() is called when the message
     content and structure are no longer needed. It's passed to the
@@ -380,7 +405,7 @@ void libbalsa_mailbox_load_config(LibBalsaMailbox * mailbox,
 				  const gchar * prefix);
 
 int libbalsa_mailbox_copy_message(LibBalsaMessage *message,
-				  LibBalsaMailbox *dest);
+				  LibBalsaMailbox *dest, GError **err);
 gboolean libbalsa_mailbox_close_backend(LibBalsaMailbox * mailbox);
 
 /* Message number-list methods */
@@ -390,10 +415,10 @@ gboolean libbalsa_mailbox_messages_change_flags(LibBalsaMailbox * mailbox,
 						LibBalsaMessageFlag clear);
 gboolean libbalsa_mailbox_messages_copy(LibBalsaMailbox * mailbox,
 					GArray * msgnos,
-					LibBalsaMailbox * dest);
+					LibBalsaMailbox * dest, GError **err);
 gboolean libbalsa_mailbox_messages_move(LibBalsaMailbox * mailbox,
 					GArray * msgnos,
-					LibBalsaMailbox * dest);
+					LibBalsaMailbox * dest, GError **err);
 
                                
 /*
@@ -401,9 +426,6 @@ gboolean libbalsa_mailbox_messages_move(LibBalsaMailbox * mailbox,
  */
 GType libbalsa_mailbox_type_from_path(const gchar * filename);
 
-void libbalsa_mailbox_messages_status_changed(LibBalsaMailbox * mbox,
-					      GList * messages,
-					      gint flag);
 guint libbalsa_mailbox_total_messages(LibBalsaMailbox * mailbox);
 
 /*
@@ -421,6 +443,9 @@ void libbalsa_mailbox_set_view_filter(LibBalsaMailbox   *mailbox,
                                       LibBalsaCondition *filter_condition,
                                       gboolean update_immediately);
 
+gboolean libbalsa_mailbox_can_do(LibBalsaMailbox *mailbox,
+                                 enum LibBalsaMailboxCapability cap);
+
 /** libbalsa_mailbox_set_threading() uses backend-optimized threading mode
     to produce a tree of messages. The tree is put in msg_tree and used
     later by GtkTreeModel interface.
@@ -456,6 +481,10 @@ void libbalsa_mailbox_set_exposed(LibBalsaMailbox * mailbox,
 void libbalsa_mailbox_set_open(LibBalsaMailbox * mailbox, gboolean open);
 void libbalsa_mailbox_set_filter(LibBalsaMailbox * mailbox, gint filter);
 void libbalsa_mailbox_set_frozen(LibBalsaMailbox * mailbox, gboolean frozen);
+#ifdef HAVE_GPGME
+gboolean libbalsa_mailbox_set_crypto_mode(LibBalsaMailbox * mailbox,
+					  LibBalsaChkCryptoMode gpg_chk_mode);
+#endif
 
 LibBalsaAddress *libbalsa_mailbox_get_mailing_list_address(LibBalsaMailbox
 							   * mailbox);
@@ -471,6 +500,9 @@ gboolean libbalsa_mailbox_get_exposed(LibBalsaMailbox * mailbox);
 gboolean libbalsa_mailbox_get_open(LibBalsaMailbox * mailbox);
 gint libbalsa_mailbox_get_filter(LibBalsaMailbox * mailbox);
 gboolean libbalsa_mailbox_get_frozen(LibBalsaMailbox * mailbox);
+#ifdef HAVE_GPGME
+LibBalsaChkCryptoMode libbalsa_mailbox_get_crypto_mode(LibBalsaMailbox * mailbox);
+#endif
 
 /** force update of given msgno */
 void libbalsa_mailbox_msgno_changed(LibBalsaMailbox  *mailbox, guint seqno);
@@ -488,6 +520,11 @@ gboolean libbalsa_mailbox_msgno_find(LibBalsaMailbox * mailbox,
 				     guint seqno,
 				     GtkTreePath ** path,
 				     GtkTreeIter * iter);
+/* Test message flags */
+gboolean libbalsa_mailbox_msgno_has_flags(LibBalsaMailbox * mailbox,
+                                          guint seqno,
+                                          LibBalsaMessageFlag set,
+                                          LibBalsaMessageFlag unset);
 
 /* set icons */
 void libbalsa_mailbox_set_unread_icon(GdkPixbuf * pixbuf);
@@ -538,4 +575,6 @@ typedef enum {
     LB_MBOX_N_COLS
 } LibBalsaMailboxColumn;
 
+extern gchar *libbalsa_mailbox_date_format;
+
 #endif				/* __LIBBALSA_MAILBOX_H__ */
