@@ -33,8 +33,7 @@
 #include "mailbox-filter.h"
 #include "misc.h"
 
-#include <libgnome/gnome-config.h> 
-#include <libgnome/gnome-i18n.h> 
+#include <libgnome/libgnome.h> 
 
 enum {
     REMOVE_FILES,
@@ -69,15 +68,14 @@ static void lbm_local_update_view_filter(LibBalsaMailbox * mailbox,
 static void libbalsa_mailbox_local_prepare_threading(LibBalsaMailbox *mailbox, 
                                                      guint lo, guint hi);
 
-static void libbalsa_mailbox_local_fetch_structure(LibBalsaMailbox *mailbox,
-                                                   LibBalsaMessage *message,
-                                                   LibBalsaFetchFlag flags);
+static gboolean libbalsa_mailbox_local_fetch_structure(LibBalsaMailbox *
+                                                       mailbox,
+                                                       LibBalsaMessage *
+                                                       message,
+                                                       LibBalsaFetchFlag
+                                                       flags);
 static void libbalsa_mailbox_local_fetch_headers(LibBalsaMailbox *mailbox,
                                                  LibBalsaMessage *message);
-static void libbalsa_mailbox_local_release_message(LibBalsaMailbox *
-						   mailbox,
-						   LibBalsaMessage *
-						   message);
 
 static gboolean libbalsa_mailbox_local_get_msg_part(LibBalsaMessage *msg,
 						    LibBalsaMessageBody *);
@@ -150,8 +148,6 @@ libbalsa_mailbox_local_class_init(LibBalsaMailboxLocalClass * klass)
         libbalsa_mailbox_local_fetch_structure;
     libbalsa_mailbox_class->fetch_headers = 
         libbalsa_mailbox_local_fetch_headers;
-    libbalsa_mailbox_class->release_message =
-	        libbalsa_mailbox_local_release_message;
     libbalsa_mailbox_class->get_message_part = 
         libbalsa_mailbox_local_get_msg_part;
     klass->load_message = NULL;
@@ -547,14 +543,15 @@ libbalsa_mailbox_local_prepare_threading(LibBalsaMailbox *mailbox,
  * to this one.
  */
 
-static void
+static gboolean
 libbalsa_mailbox_local_fetch_structure(LibBalsaMailbox *mailbox,
-                                      LibBalsaMessage *message,
-                                      LibBalsaFetchFlag flags)
+                                       LibBalsaMessage *message,
+                                       LibBalsaFetchFlag flags)
 {
     GMimeMessage *mime_message = message->mime_msg;
-    g_assert(mime_message != NULL);
-    g_assert(mime_message->mime_part != NULL);
+
+    if (!mime_message || !mime_message->mime_part)
+	return FALSE;
 
     if(flags & LB_FETCH_STRUCTURE) {
         LibBalsaMessageBody *body = libbalsa_message_body_new(message);
@@ -568,16 +565,8 @@ libbalsa_mailbox_local_fetch_structure(LibBalsaMailbox *mailbox,
             libbalsa_message_user_hdrs_from_gmime(mime_message);
         message->has_all_headers = 1;
     }
-}
 
-static void
-libbalsa_mailbox_local_release_message(LibBalsaMailbox * mailbox,
-				       LibBalsaMessage * message)
-{
-    if (message->mime_msg) {
-	g_object_unref(message->mime_msg);
-	message->mime_msg = NULL;
-    }
+    return TRUE;
 }
 
 static void
@@ -592,7 +581,7 @@ libbalsa_mailbox_local_fetch_headers(LibBalsaMailbox * mailbox,
     else {
 	libbalsa_mailbox_fetch_message_structure(mailbox, message,
 						 LB_FETCH_RFC822_HEADERS);
-	libbalsa_mailbox_local_release_message(mailbox, message);
+	libbalsa_mailbox_release_message(mailbox, message);
     }
 }
 
@@ -1215,7 +1204,8 @@ lbml_thread_message(GNode * node, ThreadingInfo * ti)
 				     g_list_last(refs)->data);
     if (!parent)
 	parent = ti->msg_tree;
-    if (parent != node->parent)
+    if (parent != node->parent && parent != node
+	&& !g_node_is_ancestor(node, parent))
 	libbalsa_mailbox_unlink_and_prepend(ti->mailbox, node, parent);
 
     return FALSE;
@@ -1226,7 +1216,7 @@ lbml_thread_message(GNode * node, ThreadingInfo * ti)
 
 /* Helper for maildir and mh. */
 GMimeMessage *
-_libbalsa_mailbox_local_get_mime_message(LibBalsaMailbox * mailbox,
+libbalsa_mailbox_local_get_mime_message(LibBalsaMailbox * mailbox,
 					 const gchar * name1,
 					 const gchar * name2)
 {
@@ -1235,19 +1225,22 @@ _libbalsa_mailbox_local_get_mime_message(LibBalsaMailbox * mailbox,
     GMimeMessage *mime_message;
 
     mime_stream =
-	_libbalsa_mailbox_local_get_message_stream(mailbox, name1, name2);
+	libbalsa_mailbox_local_get_message_stream(mailbox, name1, name2);
+    if (!mime_stream)
+	return NULL;
+
     mime_parser = g_mime_parser_new_with_stream(mime_stream);
     g_mime_parser_set_scan_from(mime_parser, FALSE);
     mime_message = g_mime_parser_construct_message(mime_parser);
 
     g_object_unref(mime_parser);
-    g_mime_stream_unref(mime_stream);
+    g_object_unref(mime_stream);
 
     return mime_message;
 }
 
 GMimeStream *
-_libbalsa_mailbox_local_get_message_stream(LibBalsaMailbox * mailbox,
+libbalsa_mailbox_local_get_message_stream(LibBalsaMailbox * mailbox,
 					   const gchar * name1,
 					   const gchar * name2)
 {
