@@ -24,9 +24,10 @@
 
 #define _XOPEN_SOURCE 500
 
-#include <gnome.h>
+#include <libgnome/libgnome.h>
 #include <gmime/gmime-stream-fs.h>
 
+#include <stdlib.h>
 #include <unistd.h>
 #include <fcntl.h>
 #include <errno.h>
@@ -46,6 +47,7 @@ struct message_info {
     off_t start;
     off_t status;		/* Offset of the "Status:" header. */
     off_t x_status;		/* Offset of the "X-Status:" header. */
+    off_t mime_version;		/* Offset of the "MIME-Version:" header. */
     off_t end;
     char *from;
     LibBalsaMessage *message; /* registers only referenced messages
@@ -76,19 +78,25 @@ static gboolean libbalsa_mailbox_mbox_sync(LibBalsaMailbox * mailbox,
 static LibBalsaMessage *libbalsa_mailbox_mbox_get_message(
 						  LibBalsaMailbox * mailbox,
 						  guint msgno);
-static void libbalsa_mailbox_mbox_fetch_message_structure(LibBalsaMailbox *
-							  mailbox,
-							  LibBalsaMessage *
-							  message,
-							  LibBalsaFetchFlag
-							  flags);
+static gboolean
+libbalsa_mailbox_mbox_fetch_message_structure(LibBalsaMailbox * mailbox,
+                                              LibBalsaMessage * message,
+                                              LibBalsaFetchFlag flags);
 static int libbalsa_mailbox_mbox_add_message(LibBalsaMailbox * mailbox,
-                                             LibBalsaMessage *message );
-static void libbalsa_mailbox_mbox_change_message_flags(LibBalsaMailbox * mailbox, guint msgno,
-					   LibBalsaMessageFlag set,
-					   LibBalsaMessageFlag clear);
-static guint libbalsa_mailbox_mbox_total_messages(LibBalsaMailbox *
-						  mailbox);
+                                             LibBalsaMessage *message,
+                                             GError **err);
+static gboolean
+libbalsa_mailbox_mbox_messages_change_flags(LibBalsaMailbox * mailbox,
+                                            GArray * msgnos,
+                                            LibBalsaMessageFlag set,
+                                            LibBalsaMessageFlag clear);
+static gboolean
+libbalsa_mailbox_mbox_msgno_has_flags(LibBalsaMailbox * mailbox,
+                                      guint msgno,
+                                      LibBalsaMessageFlag set,
+                                      LibBalsaMessageFlag unset);
+static guint
+libbalsa_mailbox_mbox_total_messages(LibBalsaMailbox * mailbox);
 
 
 GType libbalsa_mailbox_mbox_get_type(void)
@@ -143,8 +151,10 @@ libbalsa_mailbox_mbox_class_init(LibBalsaMailboxMboxClass * klass)
     libbalsa_mailbox_class->fetch_message_structure =
 	libbalsa_mailbox_mbox_fetch_message_structure;
     libbalsa_mailbox_class->add_message = libbalsa_mailbox_mbox_add_message;
-    libbalsa_mailbox_class->change_message_flags =
-	libbalsa_mailbox_mbox_change_message_flags;
+    libbalsa_mailbox_class->messages_change_flags =
+	libbalsa_mailbox_mbox_messages_change_flags;
+    libbalsa_mailbox_class->msgno_has_flags =
+	libbalsa_mailbox_mbox_msgno_has_flags;
     libbalsa_mailbox_class->total_messages =
 	libbalsa_mailbox_mbox_total_messages;
 
@@ -214,6 +224,29 @@ libbalsa_mailbox_mbox_new(const gchar * path, gboolean create)
     return G_OBJECT(mailbox);
 }
 
+/* Helper: seek to offset, and return TRUE if the seek succeeds and a
+ * message begins there. */
+static gboolean
+lbm_mbox_stream_seek_to_message(GMimeStream * stream, off_t offset)
+{
+    char buffer[5];
+    gboolean retval;
+
+    retval = g_mime_stream_seek(stream, offset, GMIME_STREAM_SEEK_SET) >= 0
+        && g_mime_stream_read(stream, buffer, sizeof(buffer)) >= 0
+        && strncmp("From ", buffer, 5) == 0;
+
+    g_mime_stream_seek(stream, offset, GMIME_STREAM_SEEK_SET);
+
+    return retval;
+}
+
+static gboolean
+lbm_mbox_seek_to_message(LibBalsaMailboxMbox * mbox, off_t offset)
+{
+    return lbm_mbox_stream_seek_to_message(mbox->gmime_stream, offset);
+}
+
 static GMimeStream *
 libbalsa_mailbox_mbox_get_message_stream(LibBalsaMailbox *mailbox,
                                          LibBalsaMessage *message)
@@ -229,6 +262,8 @@ libbalsa_mailbox_mbox_get_message_stream(LibBalsaMailbox *mailbox,
 
 	msg_info = &g_array_index(mbox->messages_info,
 				  struct message_info, message->msgno - 1);
+	if (!lbm_mbox_seek_to_message(mbox, msg_info->start))
+	    return NULL;
 	stream = g_mime_stream_substream(mbox->gmime_stream,
 					 msg_info->start
 					 + strlen(msg_info->from) + 1,
@@ -272,25 +307,31 @@ static void mbox_unlock(LibBalsaMailbox * mailbox, GMimeStream *stream)
 }
 
 /* GMimeParserHeaderRegexFunc callback; save the header's offset if it's
- * either "Status" or "X-Status".
+ * "Status", "X-Status", or "MIME-Version".  Save only the first one, to
+ * avoid headers in encapsulated messages.
  * 
- * We'll save only the first of each that we see, to try to avoid
- * headers on embedded messages.  This isn't guaranteed, as the real
- * message may be missing one of the headers, if it was delivered by
- * someone else--we make sure every message has both, with blank flags
- * if necessary, when delivering our own mail.
+ * If a message has no status headers but an encapsulated message does,
+ * we may save the offset of the encapsulated header; we check for that
+ * later.
+ * 
+ * We use the offset of the "MIME-Version" header as the location to
+ * insert status headers, if necessary.
  */
 static void
 lbm_mbox_header_cb(GMimeParser * parser, const char *header,
-		   const char *value, off_t offset,
-		   struct message_info ** msg_info_p)
+                   const char *value, off_t offset,
+                   struct message_info **msg_info_p)
 {
     struct message_info *msg_info = *msg_info_p;
-    if (g_ascii_strcasecmp(header, "status") == 0 && msg_info->status < 0)
-	msg_info->status = offset;
-    else if (g_ascii_strcasecmp(header, "x-status") == 0
-	     && msg_info->x_status < 0)
-	msg_info->x_status = offset;
+
+    if (g_ascii_strcasecmp(header, "Status") == 0 && msg_info->status < 0)
+        msg_info->status = offset;
+    else if (g_ascii_strcasecmp(header, "X-Status") == 0
+             && msg_info->x_status < 0)
+        msg_info->x_status = offset;
+    else if (g_ascii_strcasecmp(header, "MIME-Version") == 0
+             && msg_info->mime_version < 0)
+        msg_info->mime_version = offset;
 }
 
 static LibBalsaMessage *lbm_mbox_message_new(GMimeMessage * mime_message,
@@ -309,16 +350,16 @@ parse_mailbox(LibBalsaMailboxMbox * mbox)
     g_mime_parser_set_scan_from(gmime_parser, TRUE);
     g_mime_parser_set_respect_content_length(gmime_parser, TRUE);
     g_mime_parser_set_header_regex(gmime_parser,
-				   "^(X-)?Status",
+                                   "^Status|^X-Status|^MIME-Version",
 				   (GMimeParserHeaderRegexFunc)
 				   lbm_mbox_header_cb, &msg_info_p);
 
     msg_info.message = NULL;
-    while (!g_mime_stream_eos(mbox->gmime_stream)) {
+    while (!g_mime_parser_eos(gmime_parser)) {
 	GMimeMessage *mime_message;
         LibBalsaMessage *msg;
 
-        msg_info.status = msg_info.x_status = -1;
+        msg_info.status = msg_info.x_status = msg_info.mime_version = -1;
         mime_message   = g_mime_parser_construct_message(gmime_parser);
         msg_info.start = g_mime_parser_get_from_offset(gmime_parser);
         msg_info.end   = g_mime_parser_tell(gmime_parser);
@@ -328,6 +369,17 @@ parse_mailbox(LibBalsaMailboxMbox * mbox)
             continue;
 	}
 
+	/* Make sure we don't have offsets for any encapsulated headers. */
+	if (!g_mime_object_get_header(GMIME_OBJECT(mime_message),
+                                      "Status"))
+	    msg_info.status = -1;
+	if (!g_mime_object_get_header(GMIME_OBJECT(mime_message),
+                                      "X-Status"))
+	    msg_info.x_status = -1;
+	if (!g_mime_object_get_header(GMIME_OBJECT(mime_message),
+                                      "MIME-Version"))
+	    msg_info.mime_version = -1;
+
         msg = lbm_mbox_message_new(mime_message, &msg_info);
         g_object_unref(mime_message);
         if (!msg) {
@@ -415,8 +467,17 @@ libbalsa_mailbox_mbox_open(LibBalsaMailbox * mailbox, GError **err)
     }
     gmime_stream = g_mime_stream_fs_new(fd);
 
+    if (st.st_size > 0
+        && !lbm_mbox_stream_seek_to_message(gmime_stream, 0)) {
+        g_object_unref(gmime_stream);
+        g_set_error(err, LIBBALSA_MAILBOX_ERROR,
+                    LIBBALSA_MAILBOX_OPEN_ERROR,
+                    _("Mailbox is not in mbox format."));
+        return FALSE;
+    }
+
     if (mbox_lock(mailbox, gmime_stream)) {
-	g_mime_stream_unref(gmime_stream);
+	g_object_unref(gmime_stream);
 	g_set_error(err, LIBBALSA_MAILBOX_ERROR, LIBBALSA_MAILBOX_OPEN_ERROR,
 		    _("Cannot lock mailbox."));
 	return FALSE;
@@ -533,16 +594,17 @@ static void
 libbalsa_mailbox_mbox_check(LibBalsaMailbox * mailbox)
 {
     struct stat st;
-    gchar buffer[10];
     const gchar *path;
     LibBalsaMailboxMbox *mbox;
-    guint last_msgno;
+    guint msgno;
 
     g_assert(LIBBALSA_IS_MAILBOX_MBOX(mailbox));
 
     mbox = LIBBALSA_MAILBOX_MBOX(mailbox);
     path = libbalsa_mailbox_local_get_path(mailbox);
-    if (stat(path, &st) != 0) {
+    if (mbox->gmime_stream ?
+        fstat(GMIME_STREAM_FS(mbox->gmime_stream)->fd, &st) :
+        stat(path, &st)) {
 	perror(path);
 	return;
     }
@@ -574,38 +636,36 @@ libbalsa_mailbox_mbox_check(LibBalsaMailbox * mailbox)
 	 */
 	return;
 
-    /*
-     * Check to make sure that the only change to the mailbox is that 
-     * message(s) were appended to this file.  The heuristic here is
-     * that we should see the message separator at *exactly* what used
-     * to be the end of the folder.
+    /* Find a good place to start parsing the mailbox.  If the only
+     * change is that message(s) were appended to this file, we should
+     * see the message separator at *exactly* what used to be the end of
+     * the folder.  If a message was expunged by another MUA, we back up
+     * over messages until we find one that is still where we expected
+     * to find it, and start parsing at its end.
      */
-    if (g_mime_stream_seek(mbox->gmime_stream, mbox->size,
-			   GMIME_STREAM_SEEK_SET) == -1 ||
-	g_mime_stream_read(mbox->gmime_stream, buffer,
-			   sizeof(buffer)) == -1 ||
-	strncmp("From ", buffer, 5) != 0 ||
-	g_mime_stream_seek(mbox->gmime_stream, mbox->size,
-			   GMIME_STREAM_SEEK_SET) == -1) {
-	/* Check failed--clear the mailbox. */
-	guint msgno;
+    while ((msgno = mbox->messages_info->len) > 0) {
+        struct message_info *msg_info =
+            &g_array_index(mbox->messages_info, struct message_info,
+                           msgno - 1);
+        if (lbm_mbox_seek_to_message(mbox, msg_info->end))
+	    /* A message begins at the end of this one, so it must(?) be
+	     * in its original position--start parsing here. */
+            break;
 
-	for (msgno = mbox->messages_info->len; msgno > 0; --msgno) {
-	    struct message_info *msg_info =
-		&g_array_index(mbox->messages_info, struct message_info,
-			       msgno - 1);
-	    libbalsa_mailbox_local_msgno_removed(mailbox, msgno);
-	    free_message_info(msg_info);
-	    g_array_remove_index(mbox->messages_info, msgno - 1);
-	}
-	g_mime_stream_reset(mbox->gmime_stream);
-	mailbox->unread_messages = 0;
+	/* Back up over this message and try again. */
+        if ((msg_info->flags & LIBBALSA_MESSAGE_FLAG_NEW)
+            && !(msg_info->flags & LIBBALSA_MESSAGE_FLAG_DELETED))
+            --mailbox->unread_messages;
+        libbalsa_mailbox_local_msgno_removed(mailbox, msgno);
+        free_message_info(msg_info);
+        g_array_remove_index(mbox->messages_info, msgno - 1);
     }
-    last_msgno = mbox->messages_info->len;
+    if(msgno == 0)
+        g_mime_stream_seek(mbox->gmime_stream, 0, GMIME_STREAM_SEEK_SET);
     parse_mailbox(mbox);
     mbox->size = g_mime_stream_tell(mbox->gmime_stream);
     mbox_unlock(mailbox, NULL);
-    libbalsa_mailbox_local_load_messages(mailbox, last_msgno);
+    libbalsa_mailbox_local_load_messages(mailbox, msgno);
 }
 
 static void
@@ -625,7 +685,7 @@ libbalsa_mailbox_mbox_close_mailbox(LibBalsaMailbox * mailbox,
 	mbox->messages_info = NULL;
     }
     if (mbox->gmime_stream) {
-	g_mime_stream_unref(mbox->gmime_stream);
+	g_object_unref(mbox->gmime_stream);
 	mbox->gmime_stream = NULL;	/* chbm: is this correct? */
     }
     if (LIBBALSA_MAILBOX_CLASS(parent_class)->close_mailbox)
@@ -642,8 +702,10 @@ lbm_mbox_get_mime_message(LibBalsaMailbox * mailbox,
     GMimeMessage *mime_message;
 
     stream = libbalsa_mailbox_mbox_get_message_stream(mailbox, message);
+    if (!stream)
+	return NULL;
     parser = g_mime_parser_new_with_stream(stream);
-    g_mime_stream_unref(stream);
+    g_object_unref(stream);
 
     mime_message = g_mime_parser_construct_message(parser);
     g_object_unref(parser);
@@ -805,7 +867,7 @@ lbm_mbox_rewrite_in_place(struct message_info *msg_info,
 	g_mime_stream_seek(stream, msg_info->x_status, GMIME_STREAM_SEEK_SET);
 	g_mime_stream_write(stream, x_status_buf, sizeof(x_status_buf));
     }
-    msg_info->orig_flags = msg_info->flags;
+    msg_info->orig_flags = msg_info->flags & LIBBALSA_MESSAGE_FLAGS_REAL;
     return TRUE;
 }
 
@@ -900,7 +962,6 @@ libbalsa_mailbox_mbox_sync(LibBalsaMailbox * mailbox, gboolean expunge)
     GMimeStream *mbox_stream;
     gchar *tempfile;
     GError *error = NULL;
-    gchar buffer[5];
     gboolean save_failed;
     GMimeParser *gmime_parser;
     LibBalsaMailboxMbox *mbox;
@@ -913,6 +974,8 @@ libbalsa_mailbox_mbox_sync(LibBalsaMailbox * mailbox, gboolean expunge)
      */
     libbalsa_mailbox_mbox_check(mailbox);
     mbox = LIBBALSA_MAILBOX_MBOX(mailbox);
+    if (mbox->messages_info->len == 0)
+	return TRUE;
     mbox_stream = mbox->gmime_stream;
 
     path = libbalsa_mailbox_local_get_path(mailbox);
@@ -922,8 +985,8 @@ libbalsa_mailbox_mbox_sync(LibBalsaMailbox * mailbox, gboolean expunge)
 	return FALSE;
 
     /* Check to make sure that the file hasn't changed on disk */
-    if (stat(path, &st) != 0 || st.st_size != mbox->size)
-    {
+    if (fstat(GMIME_STREAM_FS(mbox_stream)->fd, &st) != 0
+        || st.st_size != mbox->size) {
 	mbox_unlock(mailbox, mbox_stream);
 	return FALSE;
     }
@@ -947,7 +1010,8 @@ libbalsa_mailbox_mbox_sync(LibBalsaMailbox * mailbox, gboolean expunge)
 	    break;
 	if (first < 0 && (msg_info->status < 0 || msg_info->x_status < 0))
 	    first = i;
-	if (msg_info->orig_flags != msg_info->flags) {
+        if ((msg_info->orig_flags ^ msg_info->flags)
+            & LIBBALSA_MESSAGE_FLAGS_REAL) {
 	    if (lbm_mbox_rewrite_in_place(msg_info, mbox_stream))
 		++j;
 	    else
@@ -962,6 +1026,12 @@ libbalsa_mailbox_mbox_sync(LibBalsaMailbox * mailbox, gboolean expunge)
 	    utimebuf.modtime = st.st_mtime;
 	    utime(path, &utimebuf);
 	}
+	if (g_mime_stream_flush(mbox_stream) < 0)
+	    g_warning("can't flush mailbox stream\n");
+	if (fstat(GMIME_STREAM_FS(mbox_stream)->fd, &st))
+	    g_warning("can't stat \"%s\"", path);
+	else
+	    mbox->mtime = st.st_mtime;
 	mbox_unlock(mailbox, mbox_stream);
 	return TRUE;
     }
@@ -997,8 +1067,10 @@ libbalsa_mailbox_mbox_sync(LibBalsaMailbox * mailbox, gboolean expunge)
 	    if (status_len < 0)
 		break;
 	} else {
-	    msg_info->status =
-		msg_info->start + strlen(msg_info->from) + 1;
+            msg_info->status = msg_info->mime_version >= 0 ?
+		msg_info->mime_version :
+                (off_t) (msg_info->start + strlen(msg_info->from) + 1);
+	    
 	    status_len = 0;
 	}
 
@@ -1045,7 +1117,7 @@ libbalsa_mailbox_mbox_sync(LibBalsaMailbox * mailbox, gboolean expunge)
     if (i < messages) {
 	/* We broke on an error. */
 	g_warning("error making temporary copy\n");
-	g_mime_stream_unref(temp_stream);
+	g_object_unref(temp_stream);
 	unlink(tempfile);
 	g_free(tempfile);
 	mbox_unlock(mailbox, mbox_stream);
@@ -1056,7 +1128,7 @@ libbalsa_mailbox_mbox_sync(LibBalsaMailbox * mailbox, gboolean expunge)
     if (g_mime_stream_flush(temp_stream) == -1)
     {
 	g_warning("can't flush temporary copy\n");
-	g_mime_stream_unref(temp_stream);
+	g_object_unref(temp_stream);
 	unlink(tempfile);
 	g_free(tempfile);
 	mbox_unlock(mailbox, mbox_stream);
@@ -1065,28 +1137,15 @@ libbalsa_mailbox_mbox_sync(LibBalsaMailbox * mailbox, gboolean expunge)
 
     save_failed = TRUE;
     if (g_mime_stream_reset(temp_stream) == -1) {
-	    g_warning("mbox_sync: can't rewind temporary copy.\n");
-    } else
-	if (g_mime_stream_seek(mbox_stream, offset,
-			       GMIME_STREAM_SEEK_SET) == -1
-	    || (g_mime_stream_length(mbox_stream)>0 &&
-                (g_mime_stream_read(mbox_stream, buffer, 5) == -1 ||
-                 strncmp ("From ", buffer, 5) != 0)) )
-	    {
-		g_warning("mbox_sync: message not in expected position.\n");
-	    } else {
-		if (g_mime_stream_seek(mbox_stream, offset,
-				       GMIME_STREAM_SEEK_SET) != -1
-		    && g_mime_stream_write_to_stream(temp_stream,
-						     mbox_stream) != -1)
-		    {
-			save_failed = FALSE;
-			mbox->size = g_mime_stream_tell(mbox_stream);
-			ftruncate(GMIME_STREAM_FS(mbox_stream)->fd, 
-				  mbox->size);
-		    }
-	    }
-    g_mime_stream_unref(temp_stream);
+        g_warning("mbox_sync: can't rewind temporary copy.\n");
+    } else if (!lbm_mbox_seek_to_message(mbox, offset))
+        g_warning("mbox_sync: message not in expected position.\n");
+    else if (g_mime_stream_write_to_stream(temp_stream, mbox_stream) != -1) {
+        save_failed = FALSE;
+        mbox->size = g_mime_stream_tell(mbox_stream);
+        ftruncate(GMIME_STREAM_FS(mbox_stream)->fd, mbox->size);
+    }
+    g_object_unref(temp_stream);
     mbox_unlock(mailbox, mbox_stream);
     if (g_mime_stream_flush(mbox_stream) == -1 || save_failed)
     {
@@ -1147,7 +1206,7 @@ libbalsa_mailbox_mbox_sync(LibBalsaMailbox * mailbox, gboolean expunge)
     g_mime_parser_set_scan_from(gmime_parser, TRUE);
     g_mime_parser_set_respect_content_length(gmime_parser, TRUE);
     g_mime_parser_set_header_regex(gmime_parser,
-				   "^(X-)?Status",
+                                   "^Status|^X-Status|^MIME-Version",
 				   (GMimeParserHeaderRegexFunc)
 				   lbm_mbox_header_cb, &msg_info);
     for (j = first; j < mbox->messages_info->len; ) {
@@ -1166,12 +1225,21 @@ libbalsa_mailbox_mbox_sync(LibBalsaMailbox * mailbox, gboolean expunge)
 	j++;
 
 	msg_info->start = g_mime_parser_tell(gmime_parser);
-	msg_info->status = msg_info->x_status = -1;
+	msg_info->status = msg_info->x_status = msg_info->mime_version = -1;
 	mime_msg = g_mime_parser_construct_message(gmime_parser);
+
+	/* Make sure we don't have offsets for any encapsulated headers. */
+	if (!g_mime_object_get_header(GMIME_OBJECT(mime_msg), "Status"))
+	    msg_info->status = -1;
+	if (!g_mime_object_get_header(GMIME_OBJECT(mime_msg), "X-Status"))
+	    msg_info->x_status = -1;
+	if (!g_mime_object_get_header(GMIME_OBJECT(mime_msg), "MIME-Version"))
+	    msg_info->mime_version = -1;
+
 	g_free(msg_info->from);
 	msg_info->from = g_mime_parser_get_from(gmime_parser);
 	msg_info->end = g_mime_parser_tell(gmime_parser);
-	msg_info->orig_flags = msg_info->flags;
+	msg_info->orig_flags = msg_info->flags & LIBBALSA_MESSAGE_FLAGS_REAL;
 	g_assert(mime_msg != NULL);
 	g_assert(mime_msg->mime_part != NULL);
 	if (!msg_info->message || !msg_info->message->mime_msg)
@@ -1212,7 +1280,8 @@ libbalsa_mailbox_mbox_get_message(LibBalsaMailbox * mailbox, guint msgno)
 	msg_info->message = lbm_mbox_message_new(mime_message, msg_info);
 	msg_info->message->mailbox = mailbox;
 	msg_info->message->msgno   = msgno;
-	msg_info->message->flags   = msg_info->flags;
+	msg_info->message->flags   =
+	    msg_info->flags & LIBBALSA_MESSAGE_FLAGS_REAL;
 	g_object_unref(mime_message);
 	g_object_unref(gmime_parser);
 	g_object_weak_ref(G_OBJECT(msg_info->message), mbox_msg_unref,
@@ -1221,7 +1290,7 @@ libbalsa_mailbox_mbox_get_message(LibBalsaMailbox * mailbox, guint msgno)
     return msg_info->message;
 }
 
-static void
+static gboolean
 libbalsa_mailbox_mbox_fetch_message_structure(LibBalsaMailbox * mailbox,
 					      LibBalsaMessage * message,
 					      LibBalsaFetchFlag flags)
@@ -1229,9 +1298,8 @@ libbalsa_mailbox_mbox_fetch_message_structure(LibBalsaMailbox * mailbox,
     if (!message->mime_msg)
 	message->mime_msg = lbm_mbox_get_mime_message(mailbox, message);
 
-    LIBBALSA_MAILBOX_CLASS(parent_class)->fetch_message_structure(mailbox,
-								  message,
-								  flags);
+    return LIBBALSA_MAILBOX_CLASS(parent_class)->
+        fetch_message_structure(mailbox, message, flags);
 }
 
 /* Create the LibBalsaMessage and call libbalsa_message_init_from_gmime
@@ -1315,7 +1383,7 @@ lbm_mbox_crlf_filter(GMimeStream * stream)
     GMimeFilter *filter;
 
     fstream = g_mime_stream_filter_new_with_stream(stream);
-    g_mime_stream_unref(stream);
+    g_object_unref(stream);
 
     filter = g_mime_filter_crlf_new(GMIME_FILTER_CRLF_DECODE,
 				    GMIME_FILTER_CRLF_MODE_CRLF_ONLY);
@@ -1365,7 +1433,7 @@ lbm_mbox_armor_part(GMimeObject ** part)
 	GMimeMessage *message =
 	    g_mime_message_part_get_message(GMIME_MESSAGE_PART(*part));
 	part = &message->mime_part;
-	g_mime_object_unref(GMIME_OBJECT(message));
+	g_object_unref(message);
 	g_mime_object_remove_header(GMIME_OBJECT(*part), "Content-Length");
     }
 
@@ -1411,19 +1479,19 @@ lbm_mbox_armor_part(GMimeObject ** part)
     g_mime_part_set_encoding(mime_part, GMIME_PART_ENCODING_QUOTEDPRINTABLE);
 
     mem = g_mime_stream_mem_new();
-    g_mime_part_write_to_stream(mime_part, mem);
-    g_mime_object_unref(GMIME_OBJECT(mime_part));
+    g_mime_object_write_to_stream(GMIME_OBJECT(mime_part), mem);
+    g_object_unref(mime_part);
 
     g_mime_stream_reset(mem);
     fstream = g_mime_stream_filter_new_with_stream(mem);
-    g_mime_stream_unref(mem);
+    g_object_unref(mem);
 
     filter = g_mime_filter_from_new(GMIME_FILTER_FROM_MODE_ARMOR);
     g_mime_stream_filter_add(GMIME_STREAM_FILTER(fstream), filter);
     g_object_unref(filter);
 
     parser = g_mime_parser_new_with_stream(fstream);
-    g_mime_stream_unref(fstream);
+    g_object_unref(fstream);
 
     *part = g_mime_parser_construct_part(parser);
     g_object_unref(parser);
@@ -1447,7 +1515,7 @@ lbm_mbox_armor_stream(GMimeStream * stream, LibBalsaMessageFlag flags)
     /* Make sure we have "Status" and "X-Status" headers, so we can
      * update them in place later, if necessary. */
     update_message_status_headers(message, flags);
-    g_mime_message_write_to_stream(message, mem);
+    g_mime_object_write_to_stream(GMIME_OBJECT(message), mem);
     g_object_unref(message);
 
     g_mime_stream_reset(mem);
@@ -1455,8 +1523,9 @@ lbm_mbox_armor_stream(GMimeStream * stream, LibBalsaMessageFlag flags)
 }
 
 /* Called with mailbox locked. */
-static int libbalsa_mailbox_mbox_add_message(LibBalsaMailbox * mailbox,
-                                             LibBalsaMessage *message )
+static int
+libbalsa_mailbox_mbox_add_message(LibBalsaMailbox * mailbox,
+                                  LibBalsaMessage *message, GError **err)
 {
     gchar date_string[27];
     gchar *sender;
@@ -1468,6 +1537,7 @@ static int libbalsa_mailbox_mbox_add_message(LibBalsaMailbox * mailbox,
     GMimeStream *orig;
     GMimeStream *dest;
     gint retval = 1;
+    off_t orig_length;
 
     ctime_r(&(message->headers->date), date_string);
 
@@ -1491,17 +1561,29 @@ static int libbalsa_mailbox_mbox_add_message(LibBalsaMailbox * mailbox,
     /* open in read-write mode */
     fd = open(path, O_RDWR);
     if (fd < 0) {
-	libbalsa_information(LIBBALSA_INFORMATION_WARNING, 
-			     _("%s: could not open %s."), "MBOX", path);
+        g_set_error(err, LIBBALSA_MAILBOX_ERROR,
+                    LIBBALSA_MAILBOX_APPEND_ERROR,
+                    _("%s: could not open %s."), "MBOX", path);
         return -1;
     }
     
-    lseek (fd, 0, SEEK_END);
+    orig_length = lseek (fd, 0, SEEK_END);
+    lseek (fd, 0, SEEK_SET);
     dest = g_mime_stream_fs_new (fd);
     if (!dest) {
-	libbalsa_information(LIBBALSA_INFORMATION_WARNING, 
-			     _("%s: could not get new mime stream."),
-			     "MBOX");
+        g_set_error(err, LIBBALSA_MAILBOX_ERROR,
+                    LIBBALSA_MAILBOX_APPEND_ERROR,
+                    _("%s: could not get new mime stream."),
+                    "MBOX");
+	g_free(from);
+	return -1;
+    }
+    if (orig_length > 0 && !lbm_mbox_stream_seek_to_message(dest, 0)) {
+        g_set_error(err, LIBBALSA_MAILBOX_ERROR,
+                    LIBBALSA_MAILBOX_APPEND_ERROR,
+                    _("%s: %s is not in mbox format."),
+                    "MBOX", path);
+	g_object_unref(dest);
 	g_free(from);
 	return -1;
     }
@@ -1509,11 +1591,12 @@ static int libbalsa_mailbox_mbox_add_message(LibBalsaMailbox * mailbox,
 
     orig = libbalsa_mailbox_get_message_stream(message->mailbox, message);
     if (!orig) {
-	libbalsa_information(LIBBALSA_INFORMATION_WARNING, 
-			     _("%s: could not get message stream."),
-			     "MBOX");
+        g_set_error(err, LIBBALSA_MAILBOX_ERROR,
+                    LIBBALSA_MAILBOX_APPEND_ERROR,
+                    _("%s: could not get message stream."),
+                    "MBOX");
 	mbox_unlock(mailbox, dest);
-	g_mime_stream_unref(dest);
+	g_object_unref(dest);
 	g_free(from);
 	return -1;
     }
@@ -1522,45 +1605,87 @@ static int libbalsa_mailbox_mbox_add_message(LibBalsaMailbox * mailbox,
     orig = lbm_mbox_armor_stream(orig, (message->flags |
 					LIBBALSA_MESSAGE_FLAG_RECENT));
 
+    g_mime_stream_seek(dest, 0, GMIME_STREAM_SEEK_END);
     if (g_mime_stream_write_string(dest, from) < (gint) strlen(from)
 	|| g_mime_stream_write_to_stream (orig, dest) < 0) {
-        libbalsa_information ( LIBBALSA_INFORMATION_ERROR,
-			       _("Error copying message to mailbox %s!\n"
-				 "Mailbox might have been corrupted!"),
-			       mailbox->name );
+        g_set_error(err, LIBBALSA_MAILBOX_ERROR,
+                    LIBBALSA_MAILBOX_APPEND_ERROR, _("Data copy error"));
 	retval = -1;
     }
 
     if (retval > 0)
 	retval = lbm_mbox_newline(dest);
  
-    g_mime_stream_unref(orig);
+    g_object_unref(orig);
+    if(retval<0)
+        truncate(path, orig_length);
     mbox_unlock (mailbox, dest);
-    g_mime_stream_unref(dest);
+    g_object_unref(dest);
     g_free(from);
 
     return retval;
 }
 
-
-static void
-libbalsa_mailbox_mbox_change_message_flags(LibBalsaMailbox * mailbox,
-					   guint msgno,
-					   LibBalsaMessageFlag set,
-					   LibBalsaMessageFlag clear)
+static gboolean
+libbalsa_mailbox_mbox_messages_change_flags(LibBalsaMailbox * mailbox,
+                                            GArray * msgnos,
+                                            LibBalsaMessageFlag set,
+                                            LibBalsaMessageFlag clear)
 {
-    struct message_info *msg_info;
+    guint i;
+    guint changed = 0;
 
-    g_return_if_fail (LIBBALSA_IS_MAILBOX_MBOX(mailbox));
-    g_return_if_fail (msgno > 0);
+    for (i = 0; i < msgnos->len; i++) {
+        guint msgno = g_array_index(msgnos, guint, i);
+        struct message_info *msg_info =
+            &g_array_index(LIBBALSA_MAILBOX_MBOX(mailbox)->messages_info,
+                           struct message_info, msgno - 1);
+        LibBalsaMessageFlag old_flags = msg_info->flags;
+        gboolean was_unread_undeleted, is_unread_undeleted;
 
-    msg_info = &g_array_index(LIBBALSA_MAILBOX_MBOX(mailbox)->messages_info,
-			      struct message_info, msgno - 1);
+        msg_info->flags |= set;
+        msg_info->flags &= ~clear;
+        if (!((old_flags ^ msg_info->flags) & LIBBALSA_MESSAGE_FLAGS_REAL))
+	    /* No real flags changed. */
+            continue;
+        ++changed;
 
-    msg_info->flags |= set;
-    msg_info->flags &= ~clear;
-    libbalsa_mailbox_index_set_flags(mailbox, msgno, msg_info->flags);
-    libbalsa_mailbox_local_queue_sync(LIBBALSA_MAILBOX_LOCAL(mailbox));
+        if (msg_info->message)
+            msg_info->message->flags =
+		msg_info->flags & LIBBALSA_MESSAGE_FLAGS_REAL;
+
+        libbalsa_mailbox_index_set_flags(mailbox, msgno, msg_info->flags);
+        libbalsa_mailbox_msgno_changed(mailbox, msgno);
+
+        was_unread_undeleted = (old_flags & LIBBALSA_MESSAGE_FLAG_NEW)
+            && !(old_flags & LIBBALSA_MESSAGE_FLAG_DELETED);
+        is_unread_undeleted = (msg_info->flags & LIBBALSA_MESSAGE_FLAG_NEW)
+            && !(msg_info->flags & LIBBALSA_MESSAGE_FLAG_DELETED);
+        mailbox->unread_messages +=
+            is_unread_undeleted - was_unread_undeleted;
+    }
+
+    if (changed > 0) {
+        libbalsa_mailbox_set_unread_messages_flag(mailbox,
+                                                  mailbox->unread_messages
+                                                  > 0);
+        libbalsa_mailbox_local_queue_sync(LIBBALSA_MAILBOX_LOCAL(mailbox));
+    }
+
+    return TRUE;
+}
+
+static gboolean
+libbalsa_mailbox_mbox_msgno_has_flags(LibBalsaMailbox * mailbox,
+                                      guint msgno, LibBalsaMessageFlag set,
+                                      LibBalsaMessageFlag unset)
+{
+    struct message_info *msg_info =
+        &g_array_index(LIBBALSA_MAILBOX_MBOX(mailbox)->messages_info,
+                       struct message_info, msgno - 1);
+
+    return (msg_info->flags & set) == set
+        && (msg_info->flags & unset) == 0;
 }
 
 static guint
