@@ -25,7 +25,7 @@
 #define _XOPEN_SOURCE          500
 #define _XOPEN_SOURCE_EXTENDED 1
 #define _POSIX_SOURCE          1
-#include <libgnome/gnome-i18n.h>
+#include <libgnome/libgnome.h>
 
 #include <unistd.h>
 #include <fcntl.h>
@@ -68,18 +68,25 @@ static struct message_info *lbm_mh_message_info_from_msgno(
 						  guint msgno);
 static LibBalsaMessage *libbalsa_mailbox_mh_get_message(LibBalsaMailbox * mailbox,
 						     guint msgno);
-static void libbalsa_mailbox_mh_fetch_message_structure(LibBalsaMailbox *
-							mailbox,
-							LibBalsaMessage *
-							message,
-							LibBalsaFetchFlag
-							flags);
+static gboolean libbalsa_mailbox_mh_fetch_message_structure(LibBalsaMailbox
+                                                            * mailbox,
+                                                            LibBalsaMessage
+                                                            * message,
+                                                            LibBalsaFetchFlag
+                                                            flags);
 static int libbalsa_mailbox_mh_add_message(LibBalsaMailbox * mailbox,
-					   LibBalsaMessage * message );
-static void libbalsa_mailbox_mh_change_message_flags(LibBalsaMailbox * mailbox,
-						     guint msgno,
-						     LibBalsaMessageFlag set,
-						     LibBalsaMessageFlag clear);
+					   LibBalsaMessage * message,
+                                           GError **err);
+static gboolean
+libbalsa_mailbox_mh_messages_change_flags(LibBalsaMailbox * mailbox,
+                                          GArray * msgnos,
+                                          LibBalsaMessageFlag set,
+                                          LibBalsaMessageFlag clear);
+static gboolean
+libbalsa_mailbox_mh_msgno_has_flags(LibBalsaMailbox * mailbox,
+                                    guint msgno,
+                                    LibBalsaMessageFlag set,
+                                    LibBalsaMessageFlag unset);
 static guint libbalsa_mailbox_mh_total_messages(LibBalsaMailbox * mailbox);
 
 
@@ -138,8 +145,10 @@ libbalsa_mailbox_mh_class_init(LibBalsaMailboxMhClass * klass)
     libbalsa_mailbox_class->fetch_message_structure =
 	libbalsa_mailbox_mh_fetch_message_structure;
     libbalsa_mailbox_class->add_message = libbalsa_mailbox_mh_add_message;
-    libbalsa_mailbox_class->change_message_flags =
-	libbalsa_mailbox_mh_change_message_flags;
+    libbalsa_mailbox_class->messages_change_flags =
+	libbalsa_mailbox_mh_messages_change_flags;
+    libbalsa_mailbox_class->msgno_has_flags =
+	libbalsa_mailbox_mh_msgno_has_flags;
     libbalsa_mailbox_class->total_messages =
 	libbalsa_mailbox_mh_total_messages;
 
@@ -247,7 +256,7 @@ libbalsa_mailbox_mh_get_message_stream(LibBalsaMailbox * mailbox,
     msg_info = lbm_mh_message_info_from_msgno(LIBBALSA_MAILBOX_MH(mailbox),
 					      message->msgno);
     tmp = MH_BASENAME(msg_info);
-    stream = _libbalsa_mailbox_local_get_message_stream(mailbox, tmp, NULL);
+    stream = libbalsa_mailbox_local_get_message_stream(mailbox, tmp, NULL);
     g_free(tmp);
 
     return stream;
@@ -259,7 +268,6 @@ libbalsa_mailbox_mh_remove_files(LibBalsaMailboxLocal *mailbox)
     const gchar* path;
     g_return_if_fail(LIBBALSA_IS_MAILBOX_MH(mailbox));
     path = libbalsa_mailbox_local_get_path(mailbox);
-    g_print("DELETE MH\n");
 
     if (!libbalsa_delete_directory_contents(path)) {
 	libbalsa_information(LIBBALSA_INFORMATION_ERROR,
@@ -434,7 +442,7 @@ lbm_mh_parse_sequences(LibBalsaMailboxMh * mailbox)
     gmime_stream = g_mime_stream_fs_new(fd);
     gmime_stream_buffer = g_mime_stream_buffer_new(gmime_stream,
 					GMIME_STREAM_BUFFER_BLOCK_READ);
-    g_mime_stream_unref(gmime_stream);
+    g_object_unref(gmime_stream);
     line = g_byte_array_new();
     do {
 	g_mime_stream_buffer_readln(gmime_stream_buffer, line);
@@ -443,7 +451,7 @@ lbm_mh_parse_sequences(LibBalsaMailboxMh * mailbox)
 	lbm_mh_handle_seq_line(mailbox, line->data);
 	line->len = 0;
     } while (!g_mime_stream_eos(gmime_stream_buffer));
-    g_mime_stream_unref(gmime_stream_buffer);
+    g_object_unref(gmime_stream_buffer);
     g_byte_array_free(line, TRUE);
 }
 
@@ -520,7 +528,7 @@ lbm_mh_check(LibBalsaMailboxMh * mh, const gchar * path)
     gmime_stream = g_mime_stream_fs_new(fd);
     gmime_stream_buffer = g_mime_stream_buffer_new(gmime_stream,
 					GMIME_STREAM_BUFFER_BLOCK_READ);
-    g_mime_stream_unref(gmime_stream);
+    g_object_unref(gmime_stream);
 
     line = (GByteArray *) g_array_new(TRUE, FALSE, 1);
     do {
@@ -559,7 +567,7 @@ lbm_mh_check(LibBalsaMailboxMh * mh, const gchar * path)
 	}
     } while (!g_mime_stream_eos(gmime_stream_buffer));
 
-    g_mime_stream_unref(gmime_stream_buffer);
+    g_object_unref(gmime_stream_buffer);
     g_byte_array_free(line, TRUE);
 
     return retval;
@@ -572,16 +580,12 @@ static void
 libbalsa_mailbox_mh_check(LibBalsaMailbox * mailbox)
 {
     struct stat st, st_sequences;
-    LibBalsaMailboxMh *mh;
-    const gchar *path;
+    LibBalsaMailboxMh *mh = LIBBALSA_MAILBOX_MH(mailbox);
+    const gchar *path = libbalsa_mailbox_local_get_path(mailbox);
     int modified = 0;
-    guint last_msgno;
+    guint renumber, msgno;
+    struct message_info *msg_info;
 
-    g_assert(LIBBALSA_IS_MAILBOX_MH(mailbox));
-
-    mh = LIBBALSA_MAILBOX_MH(mailbox);
-
-    path = libbalsa_mailbox_local_get_path(mailbox);
     if (stat(path, &st) == -1)
 	return;
 
@@ -627,10 +631,37 @@ libbalsa_mailbox_mh_check(LibBalsaMailbox * mailbox)
 	return;
     }
 
-    last_msgno = mh->msgno_2_msg_info->len;
-    lbm_mh_parse_both(mh);
+    /* Was any message removed? */
+    renumber = mh->msgno_2_msg_info->len + 1;
+    for (msgno = 1; msgno <= mh->msgno_2_msg_info->len; ) {
+	gchar *tmp, *filename;
 
-    libbalsa_mailbox_local_load_messages(mailbox, last_msgno);
+	msg_info = lbm_mh_message_info_from_msgno(mh, msgno);
+	tmp = MH_BASENAME(msg_info);
+	filename = g_build_filename(path, tmp, NULL);
+	g_free(tmp);
+	if (access(filename, F_OK) == 0)
+	    msgno++;
+	else {
+	    g_ptr_array_remove(mh->msgno_2_msg_info, msg_info);
+	    g_hash_table_remove(mh->messages_info, 
+		    		GINT_TO_POINTER(msg_info->fileno));
+	    libbalsa_mailbox_local_msgno_removed(mailbox, msgno);
+	    if (renumber > msgno)
+		/* First message that needs renumbering. */
+		renumber = msgno;
+	}
+	g_free(filename);
+    }
+    for (msgno = renumber; msgno <= mh->msgno_2_msg_info->len; msgno++) {
+	msg_info = lbm_mh_message_info_from_msgno(mh, msgno);
+	if (msg_info->message)
+	    msg_info->message->msgno = msgno;
+    }
+
+    msgno = mh->msgno_2_msg_info->len;
+    lbm_mh_parse_both(mh);
+    libbalsa_mailbox_local_load_messages(mailbox, msgno);
 }
 
 static void
@@ -813,7 +844,8 @@ libbalsa_mailbox_mh_sync(LibBalsaMailbox * mailbox, gboolean expunge)
 		old_file = g_build_filename(path, tmp, NULL);
 		g_free(tmp);
 
-		msg_info->orig_flags = msg_info->flags;
+		msg_info->orig_flags =
+		    msg_info->flags & LIBBALSA_MESSAGE_FLAGS_REAL;
 
 		tmp = MH_BASENAME(msg_info);
 		new_file = g_build_filename(path, tmp, NULL);
@@ -826,7 +858,8 @@ libbalsa_mailbox_mh_sync(LibBalsaMailbox * mailbox, gboolean expunge)
 		g_free(old_file);
 		g_free(new_file);
 	    } else
-		msg_info->orig_flags = msg_info->flags;
+		msg_info->orig_flags =
+		    msg_info->flags & LIBBALSA_MESSAGE_FLAGS_REAL;
 	    msgno++;
 	}
     }
@@ -847,10 +880,10 @@ libbalsa_mailbox_mh_sync(LibBalsaMailbox * mailbox, gboolean expunge)
     fd = libbalsa_mailbox_mh_open_temp(path, &tmp);
     if (fd == -1)
     {
-	g_mime_stream_unref(unseen.line);
-	g_mime_stream_unref(flagged.line);
-	g_mime_stream_unref(replied.line);
-	g_mime_stream_unref(recent.line);
+	g_object_unref(unseen.line);
+	g_object_unref(flagged.line);
+	g_object_unref(replied.line);
+	g_object_unref(recent.line);
 	libbalsa_unlock_file(sequences_filename, sequences_fd, 1);
 	return FALSE;
     }
@@ -860,7 +893,7 @@ libbalsa_mailbox_mh_sync(LibBalsaMailbox * mailbox, gboolean expunge)
     gmime_stream = g_mime_stream_fs_new(sequences_fd);
     gmime_stream_buffer = g_mime_stream_buffer_new(gmime_stream,
 					GMIME_STREAM_BUFFER_BLOCK_READ);
-    g_mime_stream_unref(gmime_stream);
+    g_object_unref(gmime_stream);
     line = g_byte_array_new();
     do {
 	line->len = 0;
@@ -875,7 +908,7 @@ libbalsa_mailbox_mh_sync(LibBalsaMailbox * mailbox, gboolean expunge)
 	    g_mime_stream_write(temp_stream, line->data, line->len);
 	}
     } while (!g_mime_stream_eos(gmime_stream_buffer));
-    g_mime_stream_unref(gmime_stream_buffer);
+    g_object_unref(gmime_stream_buffer);
     g_byte_array_free(line, TRUE);
 
     /* write sequences */
@@ -885,17 +918,17 @@ libbalsa_mailbox_mh_sync(LibBalsaMailbox * mailbox, gboolean expunge)
 	!lbm_mh_finish_line(&recent, temp_stream, LibBalsaMailboxMhRecent)) {
 	unlink(tmp);
 	g_free(tmp);
-	g_mime_stream_unref(temp_stream);
-	g_mime_stream_unref(unseen.line);
-	g_mime_stream_unref(flagged.line);
-	g_mime_stream_unref(replied.line);
-	g_mime_stream_unref(recent.line);
+	g_object_unref(temp_stream);
+	g_object_unref(unseen.line);
+	g_object_unref(flagged.line);
+	g_object_unref(replied.line);
+	g_object_unref(recent.line);
 	libbalsa_unlock_file(sequences_filename, sequences_fd, 1);
 	return FALSE;
     }
 
     /* close tempfile */
-    g_mime_stream_unref(temp_stream);
+    g_object_unref(temp_stream);
 
     /* unlink '.mh_sequences' file */
     unlink(sequences_filename);
@@ -906,10 +939,10 @@ libbalsa_mailbox_mh_sync(LibBalsaMailbox * mailbox, gboolean expunge)
 	unlink (tmp);
 
     g_free(tmp);
-    g_mime_stream_unref(unseen.line);
-    g_mime_stream_unref(flagged.line);
-    g_mime_stream_unref(replied.line);
-    g_mime_stream_unref(recent.line);
+    g_object_unref(unseen.line);
+    g_object_unref(flagged.line);
+    g_object_unref(replied.line);
+    g_object_unref(recent.line);
 
     /* Record the mtimes; we'll just use the current time--someone else
      * might have changed something since we did, despite the file
@@ -943,37 +976,24 @@ libbalsa_mailbox_mh_get_message(LibBalsaMailbox * mailbox, guint msgno)
     if (msg_info->message)
 	g_object_ref(msg_info->message);
     else {
-	const gchar *path;
-	gchar *filename;
-	gchar *tmp;
 	LibBalsaMessage *message;
 
 	msg_info->message = message = libbalsa_message_new();
 	g_object_add_weak_pointer(G_OBJECT(message),
 				  (gpointer) & msg_info->message);
 
-	path = libbalsa_mailbox_local_get_path(mailbox);
-	tmp = MH_BASENAME(msg_info);
-	filename = g_build_filename(path, tmp, NULL);
-	g_free(tmp);
-	if (!libbalsa_message_load_envelope_from_file(message, filename)) {
-	    g_free(filename);
-	    printf(" no envelope--returning NULL\n");
-	    return NULL;
-	}
-	g_free(filename);
-
 	if (msg_info->flags == INVALID_FLAG)
 	    msg_info->flags = msg_info->orig_flags;
-	message->flags = msg_info->flags;
+	message->flags = msg_info->flags & LIBBALSA_MESSAGE_FLAGS_REAL;
 	message->mailbox = mailbox;
 	message->msgno = msgno;
+	libbalsa_message_load_envelope(message);
     }
 
     return msg_info->message;
 }
 
-static void
+static gboolean
 libbalsa_mailbox_mh_fetch_message_structure(LibBalsaMailbox * mailbox,
 					    LibBalsaMessage * message,
 					    LibBalsaFetchFlag flags)
@@ -988,18 +1008,19 @@ libbalsa_mailbox_mh_fetch_message_structure(LibBalsaMailbox * mailbox,
 
 	tmp = MH_BASENAME(msg_info);
 	message->mime_msg =
-	    _libbalsa_mailbox_local_get_mime_message(mailbox, tmp, NULL);
+	    libbalsa_mailbox_local_get_mime_message(mailbox, tmp, NULL);
 	g_free(tmp);
 
-	g_mime_object_remove_header(GMIME_OBJECT(message->mime_msg),
-				    "Status");
-	g_mime_object_remove_header(GMIME_OBJECT(message->mime_msg),
-				    "X-Status");
+	if (message->mime_msg) {
+	    g_mime_object_remove_header(GMIME_OBJECT(message->mime_msg),
+				        "Status");
+	    g_mime_object_remove_header(GMIME_OBJECT(message->mime_msg),
+				        "X-Status");
+	}
     }
 
-    LIBBALSA_MAILBOX_CLASS(parent_class)->fetch_message_structure(mailbox,
-								  message,
-								  flags);
+    return LIBBALSA_MAILBOX_CLASS(parent_class)->
+        fetch_message_structure(mailbox, message, flags);
 }
 
 /* Update .mh_sequences when a new message is added to the mailbox;
@@ -1030,7 +1051,7 @@ lbm_mh_update_sequences(LibBalsaMailboxMh * mh, gint fileno,
 /* Called with mailbox locked. */
 static int
 libbalsa_mailbox_mh_add_message(LibBalsaMailbox * mailbox,
-					   LibBalsaMessage * message )
+                                LibBalsaMessage * message, GError **err)
 {
     LibBalsaMailboxMh *mh;
     const char *path;
@@ -1047,8 +1068,12 @@ libbalsa_mailbox_mh_add_message(LibBalsaMailbox * mailbox,
     /* open tempfile */
     path = libbalsa_mailbox_local_get_path(mailbox);
     fd = libbalsa_mailbox_mh_open_temp(path, &tmp);
-    if (fd == -1)
+    if (fd == -1) {
+        g_set_error(err, LIBBALSA_MAILBOX_ERROR,
+                    LIBBALSA_MAILBOX_APPEND_ERROR,
+                    _("Cannot create message"));
 	return -1;
+    }
     out_stream = g_mime_stream_fs_new(fd);
     {
 	GMimeStream *tmp = libbalsa_mailbox_get_message_stream( message->mailbox, message );
@@ -1058,17 +1083,20 @@ libbalsa_mailbox_mh_add_message(LibBalsaMailbox * mailbox,
 	in_stream = g_mime_stream_filter_new_with_stream (tmp);
 	g_mime_stream_filter_add ( GMIME_STREAM_FILTER (in_stream), crlffilter );
 	g_object_unref (crlffilter);
-	g_mime_stream_unref (tmp);
+	g_object_unref (tmp);
     }    
     if (g_mime_stream_write_to_stream( in_stream, out_stream) == -1)
     {
-	g_mime_stream_unref(out_stream);
+	g_object_unref(out_stream);
 	unlink (tmp);
 	g_free(tmp);
+        g_set_error(err, LIBBALSA_MAILBOX_ERROR,
+                    LIBBALSA_MAILBOX_APPEND_ERROR,
+                    _("Data copy error"));
 	return -1;
     }
-    g_mime_stream_unref(out_stream);
-    g_mime_stream_unref(in_stream);
+    g_object_unref(out_stream);
+    g_object_unref(in_stream);
 
     fileno = mh->last_fileno; 
     retries = 10;
@@ -1087,16 +1115,20 @@ libbalsa_mailbox_mh_add_message(LibBalsaMailbox * mailbox,
 	{
 	    unlink (tmp);
 	    g_free(tmp);
-	    /* FIXME: report error ... */
+            g_set_error(err, LIBBALSA_MAILBOX_ERROR,
+                        LIBBALSA_MAILBOX_APPEND_ERROR,
+                        _("Message rename error"));
 	    return -1;
 	}
     } while (--retries > 0);
     g_free(tmp);
 
-    if (retries == 0)
-	/* FIXME: report error ... */
+    if (retries == 0) {
+        g_set_error(err, LIBBALSA_MAILBOX_ERROR,
+                    LIBBALSA_MAILBOX_APPEND_ERROR,
+                    "Too high activity?");
 	return -1;
-
+    }
     mh->last_fileno = fileno;
 
     lbm_mh_update_sequences(mh, fileno,
@@ -1105,24 +1137,73 @@ libbalsa_mailbox_mh_add_message(LibBalsaMailbox * mailbox,
     return 1;
 }
 
-static void
-libbalsa_mailbox_mh_change_message_flags(LibBalsaMailbox * mailbox, guint msgno,
-					   LibBalsaMessageFlag set,
-					   LibBalsaMessageFlag clear)
+static gboolean
+libbalsa_mailbox_mh_messages_change_flags(LibBalsaMailbox * mailbox,
+                                          GArray * msgnos,
+                                          LibBalsaMessageFlag set,
+                                          LibBalsaMessageFlag clear)
 {
-    struct message_info *msg_info;
+    guint i;
+    guint changed = 0;
 
-    msg_info =
-	lbm_mh_message_info_from_msgno(LIBBALSA_MAILBOX_MH(mailbox), msgno);
+    for (i = 0; i < msgnos->len; i++) {
+        guint msgno = g_array_index(msgnos, guint, i);
+        struct message_info *msg_info =
+            lbm_mh_message_info_from_msgno(LIBBALSA_MAILBOX_MH(mailbox),
+                                           msgno);
+        LibBalsaMessageFlag old_flags;
+        gboolean was_unread_undeleted, is_unread_undeleted;
+
+	if (msg_info->flags == INVALID_FLAG)
+	    msg_info->flags = msg_info->orig_flags;
+        old_flags = msg_info->flags;
+
+        msg_info->flags |= set;
+        msg_info->flags &= ~clear;
+        if (!((old_flags ^ msg_info->flags) & LIBBALSA_MESSAGE_FLAGS_REAL))
+	    /* No real flags changed. */
+            continue;
+        ++changed;
+
+        if (msg_info->message)
+            msg_info->message->flags =
+		msg_info->flags & LIBBALSA_MESSAGE_FLAGS_REAL;
+
+        libbalsa_mailbox_index_set_flags(mailbox, msgno, msg_info->flags);
+        libbalsa_mailbox_msgno_changed(mailbox, msgno);
+
+        was_unread_undeleted = (old_flags & LIBBALSA_MESSAGE_FLAG_NEW)
+            && !(old_flags & LIBBALSA_MESSAGE_FLAG_DELETED);
+        is_unread_undeleted = (msg_info->flags & LIBBALSA_MESSAGE_FLAG_NEW)
+            && !(msg_info->flags & LIBBALSA_MESSAGE_FLAG_DELETED);
+        mailbox->unread_messages +=
+            is_unread_undeleted - was_unread_undeleted;
+    }
+
+    if (changed > 0) {
+        libbalsa_mailbox_set_unread_messages_flag(mailbox,
+                                                  mailbox->unread_messages
+                                                  > 0);
+        libbalsa_mailbox_local_queue_sync(LIBBALSA_MAILBOX_LOCAL(mailbox));
+    }
+
+    return TRUE;
+}
+
+static gboolean
+libbalsa_mailbox_mh_msgno_has_flags(LibBalsaMailbox * mailbox, guint msgno,
+                                    LibBalsaMessageFlag set,
+                                    LibBalsaMessageFlag unset)
+{
+    struct message_info *msg_info =
+        lbm_mh_message_info_from_msgno(LIBBALSA_MAILBOX_MH(mailbox),
+                                       msgno);
 
-    g_return_if_fail (msg_info != NULL);
     if (msg_info->flags == INVALID_FLAG)
 	msg_info->flags = msg_info->orig_flags;
-    msg_info->flags |= set;
-    msg_info->flags &= ~clear;
 
-    libbalsa_mailbox_index_set_flags(mailbox, msgno, msg_info->flags);
-    libbalsa_mailbox_local_queue_sync(LIBBALSA_MAILBOX_LOCAL(mailbox));
+    return (msg_info->flags & set) == set
+        && (msg_info->flags & unset) == 0;
 }
 
 static guint
