@@ -38,7 +38,7 @@
 #include <fcntl.h>
 
 #include <glib.h>
-#include <libgnome/gnome-i18n.h> 
+#include <libgnome/libgnome.h> 
 
 #include "libbalsa.h"
 #include "libbalsa_private.h"
@@ -175,20 +175,24 @@ libbalsa_message_finalize(GObject * object)
     message->body_list = NULL;
 
     if (message->mime_msg) {
-	g_mime_object_unref(GMIME_OBJECT(message->mime_msg));
+	g_object_unref(message->mime_msg);
 	message->mime_msg = NULL;
     }
     G_OBJECT_CLASS(parent_class)->finalize(object);
     /* printf("%p message finalized.\n", message); */  message_cnt--;
 }
 
-static void libbalsa_message_find_charset(GMimeObject *mime_part, gpointer data)
+static void
+libbalsa_message_find_charset(GMimeObject * mime_part,
+                              const gchar ** charset)
 {
-	const GMimeContentType *type;
-	if (*(gchar **)data)
-		return;
-	type=g_mime_object_get_content_type(mime_part);
-	*(const gchar **)data=g_mime_content_type_get_parameter(type, "charset");
+    const GMimeContentType *type;
+
+    if (*charset)
+        return;
+
+    type = g_mime_object_get_content_type(mime_part);
+    *charset = g_mime_content_type_get_parameter(type, "charset");
 }
 
 static void
@@ -252,15 +256,16 @@ libbalsa_message_body_charset(LibBalsaMessageBody * body)
 {
     const gchar *charset = NULL;
 
+    if (body->charset) /* This overrides all! Important for non
+                        * us-ascii messages over IMAP. */
+        return body->charset;
     if (body->mime_part) {
-	gchar *tmp = NULL;
-
 	if (GMIME_IS_MULTIPART(body->mime_part))
 	    g_mime_multipart_foreach(GMIME_MULTIPART(body->mime_part),
-				     libbalsa_message_find_charset, &tmp);
+				     (GMimePartFunc)
+				     libbalsa_message_find_charset, &charset);
 	else
-	    libbalsa_message_find_charset(body->mime_part, &tmp);
-	charset = tmp;
+	    libbalsa_message_find_charset(body->mime_part, &charset);
     } else {
 	do {
 	    if (body->charset) {
@@ -295,8 +300,7 @@ libbalsa_message_charset(LibBalsaMessage * message)
         if (!charset)
             return NULL;
     }
-    /* FIXME: use function g_mime_charset_canon_name from gmime-2.2 */
-    tmp = g_mime_charset_name(charset);
+    tmp = g_mime_charset_canon_name(charset);
     return g_strdup(tmp);
 }
 
@@ -439,7 +443,7 @@ libbalsa_message_user_hdrs_from_gmime(GMimeMessage * message)
     value = g_mime_message_get_header(message, "In-Reply-To");
     if (value) {
 	res = g_list_prepend(res, libbalsa_create_hdr_pair("In-Reply-To",
-				g_mime_utils_8bit_header_decode(value)));
+				g_mime_utils_header_decode_text(value)));
     }
 
     g_mime_header_foreach(GMIME_OBJECT(message)->headers,
@@ -452,61 +456,13 @@ libbalsa_message_user_hdrs_from_gmime(GMimeMessage * message)
    return a message part identified by Content-ID=id
    message must be referenced. (FIXME?)
 */
-FILE*
+GMimeStream *
 libbalsa_message_get_part_by_id(LibBalsaMessage* msg, const gchar* id)
 {
     LibBalsaMessageBody* body = 
 	libbalsa_message_body_get_by_id(msg->body_list,	id);
     if(!body) return NULL;
-    if(!libbalsa_message_body_save_temporary(body)) return NULL;
-    return fopen(body->temp_filename, "r");
-}
-
-
-gboolean
-libbalsa_messages_move (GList* messages, LibBalsaMailbox* dest)
-{
-    gboolean r = TRUE;
-    LibBalsaMessage *message;
-    LibBalsaMailbox *mailbox = NULL;
-    GList *p;
-    GArray *msgnos;
-
-    g_return_val_if_fail(messages, FALSE);
-    g_return_val_if_fail(dest != NULL, FALSE);
-
-    if (LIBBALSA_MESSAGE(messages->data)->mailbox->readonly) {
-	libbalsa_information(
-	    LIBBALSA_INFORMATION_ERROR,
-	    _("Source mailbox (%s) is readonly. Cannot move messages"),
-	    LIBBALSA_MESSAGE(messages->data)->mailbox->name);
-	return FALSE;
-    }
-
-    for (p = messages; p; p = p->next) {
-	message = p->data;
-	if (message->mailbox) {
-	    mailbox = message->mailbox;
-	    break;
-	}
-    }
-    g_return_val_if_fail(mailbox != NULL, FALSE);
-    libbalsa_lock_mailbox(mailbox);
-    
-    msgnos = g_array_new(FALSE, FALSE, sizeof(guint));
-    for(p=messages; p; 	p=g_list_next(p)) {
-	message=LIBBALSA_MESSAGE(p->data);
-	if(message->mailbox==NULL) continue;
-	g_array_append_val(msgnos, message->msgno);
-    }
-    libbalsa_mailbox_register_msgnos(mailbox, msgnos);
-    r = libbalsa_mailbox_messages_move(mailbox, msgnos, dest);
-    libbalsa_mailbox_unregister_msgnos(mailbox, msgnos);
-    g_array_free(msgnos, TRUE);
-
-    libbalsa_unlock_mailbox(mailbox);
-
-    return r;
+    return libbalsa_message_body_get_stream(body);
 }
 
 /* libbalsa_message_save:
@@ -531,46 +487,12 @@ libbalsa_message_save(LibBalsaMessage * message, const gchar *filename)
     out_stream = g_mime_stream_file_new(outfile);
     res = g_mime_stream_write_to_stream(msg_stream, out_stream);
 
-    g_mime_stream_unref(msg_stream);
-    g_mime_stream_unref(out_stream);
+    g_object_unref(msg_stream);
+    g_object_unref(out_stream);
 
     return res >= 0;
 }
 
-/* libbalsa_messages_copy:
-   makes an assumption that all the messages come from the same mailbox.
-*/
-gboolean
-libbalsa_messages_copy (GList * messages, LibBalsaMailbox * dest)
-{
-    LibBalsaMessage *message;
-    LibBalsaMailbox *mailbox;
-    GList *p;
-    GArray *msgnos;
-    gboolean retval;
-
-    g_return_val_if_fail(messages != NULL, FALSE);
-    g_return_val_if_fail(dest != NULL, FALSE);
-
-    mailbox = LIBBALSA_MESSAGE(messages->data)->mailbox;
-    libbalsa_lock_mailbox(mailbox);
-
-    msgnos = g_array_new(FALSE, FALSE, sizeof(guint));
-    for(p=messages; p; 	p=g_list_next(p)) {
-	message=LIBBALSA_MESSAGE(p->data);
-	if(message->mailbox==NULL) continue;
-	g_array_append_val(msgnos, message->msgno);
-    }
-    libbalsa_mailbox_register_msgnos(mailbox, msgnos);
-    retval = libbalsa_mailbox_messages_copy(mailbox, msgnos, dest) != -1;
-    libbalsa_mailbox_unregister_msgnos(mailbox, msgnos);
-    g_array_free(msgnos, TRUE);
-
-    libbalsa_unlock_mailbox(mailbox);
-
-    return retval;
-}
-
 LibBalsaMessageAttach
 libbalsa_message_get_attach_icon(LibBalsaMessage * message)
 {
@@ -713,6 +635,8 @@ libbalsa_message_body_ref(LibBalsaMessage * message, gboolean read,
                           gboolean fetch_all_headers)
 {
     LibBalsaFetchFlag flags = 0;
+    gboolean retval = TRUE;
+
     g_return_val_if_fail(message, FALSE);
     if (!message->mailbox) return FALSE;
     g_return_val_if_fail(MAILBOX_OPEN(message->mailbox), FALSE);
@@ -726,13 +650,15 @@ libbalsa_message_body_ref(LibBalsaMessage * message, gboolean read,
         /* not fetched yet */
         flags |= LB_FETCH_STRUCTURE;
 
-    if(flags)
-        libbalsa_mailbox_fetch_message_structure(message->mailbox, message,
-                                                 flags);
-    message->body_ref++;
+    if (flags)
+        retval =
+            libbalsa_mailbox_fetch_message_structure(message->mailbox,
+                                                     message, flags);
+    if (retval)
+	message->body_ref++;
     libbalsa_unlock_mailbox(message->mailbox);
     
-    return TRUE;
+    return retval;
 }
 
 
@@ -902,7 +828,7 @@ libbalsa_message_get_subject(LibBalsaMessage* msg)
         ret = g_mime_message_get_subject(msg->mime_msg);
 	if (ret)
 	    LIBBALSA_MESSAGE_SET_SUBJECT(msg, (gchar*)
-				 ret = g_mime_utils_8bit_header_decode(ret));
+				 ret = g_mime_utils_header_decode_text(ret));
     } else
 	ret = msg->subj;
 
@@ -1079,7 +1005,7 @@ libbalsa_message_init_from_gmime(LibBalsaMessage * message,
 #ifdef MESSAGE_COPY_CONTENT
     header = g_mime_message_get_subject(mime_msg);
     if (header) {
-	message->subj = g_mime_utils_8bit_header_decode(header);
+	message->subj = g_mime_utils_header_decode_text(header);
         canonize_header_value(message->subj);
 	libbalsa_utf8_sanitize(&message->subj, TRUE, NULL);
     }
@@ -1204,7 +1130,7 @@ lbmsg_set_header(LibBalsaMessage *message, const gchar *name,
 	    return FALSE;
 #if MESSAGE_COPY_CONTENT
 	g_free(message->subj);
-        message->subj = g_mime_utils_8bit_header_decode(value);
+        message->subj = g_mime_utils_header_decode_text(value);
 	libbalsa_utf8_sanitize(&message->subj, TRUE, NULL);
 #endif
     } else
@@ -1279,22 +1205,23 @@ libbalsa_message_set_headers_from_string(LibBalsaMessage *message,
     return lb_message_set_headers_from_string(message, lines, TRUE);
 }
 
-gboolean
-libbalsa_message_load_envelope_from_file(LibBalsaMessage *message,
-					 const char *filename)
+void
+libbalsa_message_load_envelope(LibBalsaMessage *message)
 {
-    int fd;
     GMimeStream *gmime_stream;
     GMimeStream *gmime_stream_buffer;
     GByteArray *line;
     char lookahead;
     gboolean ret = FALSE;
 
-    fd = open(filename, O_RDONLY);
-    gmime_stream = g_mime_stream_fs_new(fd);
+    gmime_stream =
+	libbalsa_mailbox_get_message_stream(message->mailbox, message);
+    if (!gmime_stream)
+	return;
     gmime_stream_buffer = g_mime_stream_buffer_new(gmime_stream,
 					GMIME_STREAM_BUFFER_BLOCK_READ);
-    g_mime_stream_unref(gmime_stream);
+    g_object_unref(gmime_stream);
+
     line = g_byte_array_new();
     do {
 	g_mime_stream_buffer_readln(gmime_stream_buffer, line);
@@ -1326,7 +1253,6 @@ libbalsa_message_load_envelope_from_file(LibBalsaMessage *message,
     if (ret) {
 	/* calculate size */
     }
-    g_mime_stream_unref(gmime_stream_buffer);
+    g_object_unref(gmime_stream_buffer);
     g_byte_array_free(line, TRUE);
-    return ret;
 }
