@@ -33,7 +33,7 @@
 #include <pthread.h>
 #endif
 
-#include <libgnome/gnome-i18n.h>
+#include <libgnome/libgnome.h>
 #include <string.h>
 
 #include "libbalsa.h"
@@ -54,13 +54,12 @@
 typedef struct _MessageQueueItem MessageQueueItem;
 
 struct _MessageQueueItem {
+    char tempfile[_POSIX_PATH_MAX];
     LibBalsaMessage *orig;
     GMimeStream *stream;
+    LibBalsaFccboxFinder finder;
 #if !ENABLE_ESMTP
     MessageQueueItem *next_message;
-#endif
-    char tempfile[_POSIX_PATH_MAX];
-#if !ENABLE_ESMTP
     enum {MQI_WAITING, MQI_FAILED,MQI_SENT} status;
 #else
     long message_size;
@@ -80,6 +79,8 @@ struct _SendMessageInfo{
        needed to transfer the message to the SMTP server.  This structure
        is opaque to the application. */
     smtp_session_t session;
+#else
+    LibBalsaFccboxFinder finder;
 #endif
     gboolean debug;
 };
@@ -122,11 +123,12 @@ libbalsa_wait_for_sending_thread(gint max_time)
 
 
 static MessageQueueItem *
-msg_queue_item_new(void)
+msg_queue_item_new(LibBalsaFccboxFinder finder)
 {
     MessageQueueItem *mqi;
 
     mqi = g_new(MessageQueueItem, 1);
+    mqi->finder = finder;
 #if !ENABLE_ESMTP
     mqi->next_message = NULL;
     mqi->status = MQI_WAITING;
@@ -142,7 +144,7 @@ msg_queue_item_destroy(MessageQueueItem * mqi)
     if (*mqi->tempfile)
 	unlink(mqi->tempfile);
     if (mqi->stream)
-	g_mime_stream_unref(mqi->stream);
+	g_object_unref(mqi->stream);
     if (mqi->orig)
 	g_object_unref(mqi->orig);
     g_free(mqi);
@@ -164,12 +166,14 @@ send_message_info_new(LibBalsaMailbox* outbox, smtp_session_t session,
 }
 #else
 static SendMessageInfo *
-send_message_info_new(LibBalsaMailbox* outbox, gboolean debug)
+send_message_info_new(LibBalsaMailbox* outbox,
+                      LibBalsaFccboxFinder finder, gboolean debug)
 {
     SendMessageInfo *smi;
 
     smi=g_new(SendMessageInfo,1);
     smi->outbox = outbox;
+    smi->finder = finder;
     smi->debug = debug;
 
     return smi;
@@ -198,7 +202,7 @@ static LibBalsaMsgCreateResult libbalsa_create_msg(LibBalsaMessage * message,
 				    gint encoding, gboolean flow);
 static LibBalsaMsgCreateResult
 libbalsa_fill_msg_queue_item_from_queu(LibBalsaMessage * message,
-				 MessageQueueItem *mqi);
+                                       MessageQueueItem *mqi);
 
 #ifdef BALSA_USE_THREADS
 void balsa_send_thread(MessageQueueItem * first_message);
@@ -268,8 +272,24 @@ ensure_send_progress_dialog()
 #define send_unlock() 
 #endif
 
+static void
+lbs_set_content(GMimePart * mime_part, gchar * content)
+{
+    GMimeStream *stream;
+    GMimeDataWrapper *wrapper;
 
-static GMimePart *
+    stream = g_mime_stream_mem_new();
+    g_mime_stream_write(stream, content, strlen(content));
+
+    wrapper = g_mime_data_wrapper_new();
+    g_mime_data_wrapper_set_stream(wrapper, stream);
+    g_object_unref(stream);
+
+    g_mime_part_set_content_object(mime_part, wrapper);
+    g_object_unref(wrapper);
+}
+
+static GMimeObject *
 add_mime_body_plain(LibBalsaMessageBody *body, gint encoding_style, 
 		    gboolean flow)
 {
@@ -304,16 +324,16 @@ add_mime_body_plain(LibBalsaMessageBody *body, gint encoding_style,
     g_mime_part_set_encoding(mime_part, encoding_style);
     g_mime_object_set_content_type_parameter(GMIME_OBJECT(mime_part),
 					     "charset", charset);
-    if (flow)
+    if (flow) {
 	g_mime_object_set_content_type_parameter(GMIME_OBJECT(mime_part),
 						 "DelSp", "Yes");
         g_mime_object_set_content_type_parameter(GMIME_OBJECT(mime_part),
 						 "Format", "Flowed");
+    }
 
-    if (!charset)
-	charset="us-ascii";
     if (g_ascii_strcasecmp(charset, "UTF-8")!=0 &&
-	g_ascii_strcasecmp(charset, "UTF8")!=0)
+	g_ascii_strcasecmp(charset, "UTF8")!=0 &&
+	g_ascii_strcasecmp(charset, "US-ASCII")!=0)
     {
 	GMimeStream *stream, *filter_stream;
 	GMimeFilter *filter;
@@ -333,12 +353,12 @@ add_mime_body_plain(LibBalsaMessageBody *body, gint encoding_style,
 
 	g_object_unref(G_OBJECT(wrapper));
 	g_object_unref(G_OBJECT(filter));
-	g_mime_stream_unref(filter_stream);
-	g_mime_stream_unref(stream);
+	g_object_unref(filter_stream);
+	g_object_unref(stream);
     } else
-	g_mime_part_set_content(mime_part, body->buffer, strlen(body->buffer));
+	lbs_set_content(mime_part, body->buffer);
 
-    return mime_part;
+    return GMIME_OBJECT(mime_part);
 }
 
 #if 0
@@ -355,8 +375,7 @@ static void dump_queue(const char*msg)
 #endif
 
 /* libbalsa_message_queue:
-   places given message in the outbox. If fcc message field is set, saves
-   it to fcc mailbox as well.
+   places given message in the outbox.
 */
 LibBalsaMsgCreateResult
 libbalsa_message_queue(LibBalsaMessage * message, LibBalsaMailbox * outbox,
@@ -367,27 +386,16 @@ libbalsa_message_queue(LibBalsaMessage * message, LibBalsaMailbox * outbox,
 
     g_return_val_if_fail(message, LIBBALSA_MESSAGE_CREATE_ERROR);
 
-    if ((result = libbalsa_create_msg(message, encoding, flow)) ==
-	LIBBALSA_MESSAGE_CREATE_OK) {
-        if(libbalsa_mailbox_copy_message( message, outbox ) < 0)
-            return LIBBALSA_MESSAGE_QUEUE_ERROR;
-        
-	if (fccbox) {
-            if (LIBBALSA_IS_MAILBOX_LOCAL(fccbox) ||
-                LIBBALSA_IS_MAILBOX_IMAP(fccbox)) {
-                if(libbalsa_mailbox_copy_message(message, fccbox)<0)
-                    return LIBBALSA_MESSAGE_SAVE_ERROR;
-                    /* fcc perhaps does not know about this message yet! 
-                     * later check will discover it and thread. 
-		    libbalsa_mailbox_set_threading(fccbox,
-						   fccbox->view->
-						   threading_type);
-                    */
-            }
-	}
-    } 
+    if ((result = libbalsa_create_msg(message, encoding, flow)) !=
+	LIBBALSA_MESSAGE_CREATE_OK)
+        return result;
 
-    return result;
+    if (fccbox)
+        g_mime_message_set_header(message->mime_msg, "X-Balsa-Fcc",
+                                  fccbox->url);
+    if(libbalsa_mailbox_copy_message( message, outbox, NULL) < 0)
+        return LIBBALSA_MESSAGE_QUEUE_ERROR;
+    return LIBBALSA_MESSAGE_CREATE_OK;
 }
 
 /* libbalsa_message_send:
@@ -397,7 +405,8 @@ libbalsa_message_queue(LibBalsaMessage * message, LibBalsaMailbox * outbox,
 #if ENABLE_ESMTP
 LibBalsaMsgCreateResult
 libbalsa_message_send(LibBalsaMessage * message, LibBalsaMailbox * outbox,
-                      LibBalsaMailbox * fccbox, gint encoding,
+                      LibBalsaMailbox *fccbox, gint encoding,
+                      LibBalsaFccboxFinder finder,
                       gchar * smtp_server, auth_context_t smtp_authctx,
                       gint tls_mode, gboolean flow, gboolean debug)
 {
@@ -407,7 +416,7 @@ libbalsa_message_send(LibBalsaMessage * message, LibBalsaMailbox * outbox,
         result = libbalsa_message_queue(message, outbox, fccbox, encoding,
 					flow);
      if (result == LIBBALSA_MESSAGE_CREATE_OK)
-	 if (!libbalsa_process_queue(outbox, smtp_server,
+	 if (!libbalsa_process_queue(outbox, finder, smtp_server,
 				     smtp_authctx, tls_mode, debug))
 	     return LIBBALSA_MESSAGE_SEND_ERROR;
  
@@ -416,22 +425,33 @@ libbalsa_message_send(LibBalsaMessage * message, LibBalsaMailbox * outbox,
 #else
 LibBalsaMsgCreateResult
 libbalsa_message_send(LibBalsaMessage* message, LibBalsaMailbox* outbox,
-		      LibBalsaMailbox* fccbox, gint encoding,
-		      gboolean flow, gboolean debug)
+		      LibBalsaMailbox* fccbox, LibBalsaFccboxFinder finder,
+                      gint encoding, gboolean flow, gboolean debug)
 {
     LibBalsaMsgCreateResult result = LIBBALSA_MESSAGE_CREATE_OK;
 
     if (message != NULL)
- 	result = libbalsa_message_queue(message, outbox, fccbox, encoding,
- 					flow);
+ 	result = libbalsa_message_queue(message, outbox, fccbox,
+                                        encoding, flow);
     if (result == LIBBALSA_MESSAGE_CREATE_OK)
- 	if (!libbalsa_process_queue(outbox, debug))
+ 	if (!libbalsa_process_queue(outbox, finder, debug))
  	    return LIBBALSA_MESSAGE_SEND_ERROR;
  
     return result;
 }
 #endif
 
+static void
+lbs_change_flags(LibBalsaMessage * msg, LibBalsaMessageFlag set,
+                 LibBalsaMessageFlag clear)
+{
+    GArray *array = g_array_new(FALSE, FALSE, sizeof(guint));
+
+    g_array_append_val(array, msg->msgno);
+    libbalsa_mailbox_messages_change_flags(msg->mailbox, array, set,
+                                           clear);
+    g_array_free(array, TRUE);
+}
 
 #if ENABLE_ESMTP
 /* [BCS] - libESMTP uses a callback function to read the message from the
@@ -496,7 +516,8 @@ add_recipients(smtp_message_t message, smtp_message_t bcc_message,
    sendmail version so don't get fooled by similar variable names.
  */
 gboolean
-libbalsa_process_queue(LibBalsaMailbox * outbox, gchar * smtp_server,
+libbalsa_process_queue(LibBalsaMailbox * outbox, LibBalsaFccboxFinder finder,
+                       gchar * smtp_server,
                        auth_context_t smtp_authctx, gint tls_mode,
                        gboolean debug)
 {
@@ -516,7 +537,6 @@ libbalsa_process_queue(LibBalsaMailbox * outbox, gchar * smtp_server,
 	send_unlock();
 	return FALSE;
     }
-    libbalsa_mailbox_check(outbox);
     if (!libbalsa_mailbox_total_messages(outbox)) {
 	libbalsa_mailbox_close(outbox, TRUE);
 	send_unlock();
@@ -560,19 +580,14 @@ libbalsa_process_queue(LibBalsaMailbox * outbox, gchar * smtp_server,
 
         libbalsa_message_body_ref(msg, TRUE, TRUE); /* FIXME: do we need 
                                                      * all headers? */
-	new_message = msg_queue_item_new();
+	new_message = msg_queue_item_new(finder);
         created = libbalsa_fill_msg_queue_item_from_queu(msg, new_message);
         libbalsa_message_body_unref(msg);
 
 	if (created != LIBBALSA_MESSAGE_CREATE_OK) {
 	    msg_queue_item_destroy(new_message);
 	} else {
-	    GList * messages = g_list_prepend(NULL, msg);
-
-            libbalsa_messages_change_flag(messages,
-                                          LIBBALSA_MESSAGE_FLAG_FLAGGED,
-                                          TRUE);
-	    g_list_free(messages);
+	    lbs_change_flags(msg, LIBBALSA_MESSAGE_FLAG_FLAGGED, 0);
 	    /*
 	       The message needs to be filtered and the newlines converted to
 	       \r\n because internally the lines foolishly terminate with the
@@ -603,9 +618,9 @@ libbalsa_process_queue(LibBalsaMailbox * outbox, gchar * smtp_server,
 		g_object_unref(G_OBJECT(filter));
 		mem_stream = g_mime_stream_mem_new();
 		g_mime_stream_write_to_stream(filter_stream, mem_stream);
-		g_mime_stream_unref(filter_stream);
+		g_object_unref(filter_stream);
 		g_mime_stream_reset(mem_stream);
-		g_mime_stream_unref(new_message->stream);
+		g_object_unref(new_message->stream);
 		new_message->stream = mem_stream;
 	    }
 
@@ -626,7 +641,7 @@ libbalsa_process_queue(LibBalsaMailbox * outbox, gchar * smtp_server,
 	    /* Add this after the Bcc: copy. */
 	    message = smtp_add_message (session);
 	    if (bcc_message)
-		smtp_set_header_option (message, "Bcc", Hdr_PROHIBIT);
+		smtp_set_header_option (message, "Bcc", Hdr_PROHIBIT, 1);
 
 	    smtp_message_set_application_data (message, new_message);
 	    smtp_set_messagecb (message, libbalsa_message_cb, new_message);
@@ -699,8 +714,10 @@ libbalsa_process_queue(LibBalsaMailbox * outbox, gchar * smtp_server,
             add_recipients(bcc_message, NULL, bcc_recip, "Cc");
 
 	    /* Prohibit status headers. */
-	    smtp_set_header_option(message, "Status", Hdr_PROHIBIT);
-	    smtp_set_header_option(message, "X-Status", Hdr_PROHIBIT);
+	    smtp_set_header_option(message, "Status", Hdr_PROHIBIT, 1);
+	    smtp_set_header_option(message, "X-Status", Hdr_PROHIBIT, 1);
+            /* ... and FCC header. */
+            smtp_set_header_option(message, "X-Balsa-Fcc", Hdr_PROHIBIT, 1);
 
 
 	    /* Estimate the size of the message.  This need not be exact
@@ -773,11 +790,10 @@ disp_recipient_status(smtp_recipient_t recipient,
 }
 
 static void
-handle_successful_send (smtp_message_t message, void *be_verbose)
+handle_successful_send(smtp_message_t message, void *be_verbose)
 {
     MessageQueueItem *mqi;
     const smtp_status_t *status;
-    GList * messages;
 
     send_lock();
     /* Get the app data and decrement the reference count.  Only delete
@@ -786,22 +802,43 @@ handle_successful_send (smtp_message_t message, void *be_verbose)
     if (mqi != NULL)
       mqi->refcount--;
 
+    if(mqi != NULL && mqi->orig != NULL && mqi->orig->mailbox)
+	lbs_change_flags(mqi->orig, 0, LIBBALSA_MESSAGE_FLAG_FLAGGED);
+    else printf("mqi: %p mqi->orig: %p mqi->orig->mailbox: %p\n",
+                  mqi, mqi ? mqi->orig : NULL, 
+                  mqi&&mqi->orig ? mqi->orig->mailbox : NULL);
     status = smtp_message_transfer_status (message);
     if (status->code / 100 == 2) {
-	if (mqi != NULL && mqi->orig != NULL && mqi->refcount <= 0) {
-	    if (mqi->orig->mailbox) {
-		messages = g_list_prepend(NULL, mqi->orig);
-		libbalsa_messages_change_flag(messages,
-                                              LIBBALSA_MESSAGE_FLAG_DELETED,
-                                              TRUE);
-		g_list_free(messages);
-	    }
+	if (mqi != NULL && mqi->orig != NULL && mqi->refcount <= 0 &&
+            mqi->orig->mailbox) {
+            gboolean remove = TRUE;
+            GList* fcclist =
+                libbalsa_message_find_user_hdr(mqi->orig, "X-Balsa-Fcc");
+            const gchar **fccurl = fcclist ? fcclist->data : NULL;
+
+	    if (mqi->orig->mailbox && fccurl) {
+                LibBalsaMailbox *fccbox = mqi->finder(fccurl[1]);
+                GError *err = NULL;
+                lbs_change_flags(mqi->orig, 0, LIBBALSA_MESSAGE_FLAG_NEW |
+                                 LIBBALSA_MESSAGE_FLAG_FLAGGED);
+		libbalsa_mailbox_sync_storage(mqi->orig->mailbox, FALSE);
+                remove =
+                    libbalsa_mailbox_copy_message(mqi->orig, fccbox, &err)>=0;
+                if(!remove) 
+                    libbalsa_information
+                        (LIBBALSA_INFORMATION_ERROR, 
+                         _("Saving sent message to %s failed: %s"),
+                         fccbox->url, err ? err->message : "?");
+                g_clear_error(&err);
+            }
+            /* If copy failed, mark the message again as flagged -
+               otherwise it will get resent again. And again, and
+               again... */
+            lbs_change_flags(mqi->orig, remove ?
+                             LIBBALSA_MESSAGE_FLAG_DELETED :
+                             LIBBALSA_MESSAGE_FLAG_FLAGGED, 0);
 	}
     } else {
-	messages = g_list_prepend(NULL, mqi->orig);
-        libbalsa_messages_change_flag(messages, LIBBALSA_MESSAGE_FLAG_FLAGGED,
-                                      FALSE);
-	g_list_free(messages);
 	/* XXX - Show the poor user the status codes and message. */
         if(*(gboolean*)be_verbose) {
             int cnt = 0;
@@ -961,7 +998,8 @@ libbalsa_smtp_event_cb (smtp_session_t session, int event_no, void *arg, ...)
    handler does that.
 */
 gboolean 
-libbalsa_process_queue(LibBalsaMailbox* outbox, gboolean debug)
+libbalsa_process_queue(LibBalsaMailbox* outbox, LibBalsaFccboxFinder finder,
+                       gboolean debug)
 {
     MessageQueueItem *mqi = NULL, *new_message;
     SendMessageInfo *send_message_info;
@@ -986,7 +1024,6 @@ libbalsa_process_queue(LibBalsaMailbox* outbox, gboolean debug)
 	send_unlock();
 	return FALSE;
     }
-    libbalsa_mailbox_check(outbox);
     for (msgno = libbalsa_mailbox_total_messages(outbox);
 	 msgno > 0; msgno--) {
         LibBalsaMsgCreateResult created;
@@ -999,19 +1036,14 @@ libbalsa_process_queue(LibBalsaMailbox* outbox, gboolean debug)
 
         libbalsa_message_body_ref(msg, TRUE, TRUE); /* FIXME: do we need
                                                       * all headers? */
-	new_message = msg_queue_item_new();
+	new_message = msg_queue_item_new(finder);
         created = libbalsa_fill_msg_queue_item_from_queu(msg, new_message);
         libbalsa_message_body_unref(msg);
 	
 	if (created != LIBBALSA_MESSAGE_CREATE_OK) {
 	    msg_queue_item_destroy(new_message);
 	} else {
-	    GList * messages = g_list_prepend(NULL, msg);
-
-            libbalsa_messages_change_flag(messages,
-                                          LIBBALSA_MESSAGE_FLAG_FLAGGED,
-                                          TRUE);
-	    g_list_free(messages);
+	    lbs_change_flags(msg, LIBBALSA_MESSAGE_FLAG_FLAGGED, 0);
 	    if (mqi)
 		mqi->next_message = new_message;
 	    else
@@ -1020,7 +1052,7 @@ libbalsa_process_queue(LibBalsaMailbox* outbox, gboolean debug)
 	}
     }
 
-    send_message_info=send_message_info_new(outbox, debug);
+    send_message_info=send_message_info_new(outbox, finder, debug);
     
 #ifdef BALSA_USE_THREADS
     
@@ -1041,15 +1073,29 @@ libbalsa_process_queue(LibBalsaMailbox* outbox, gboolean debug)
 }
 
 static void
-handle_successful_send(MessageQueueItem *mqi)
+handle_successful_send(MessageQueueItem *mqi, LibBalsaFccboxFinder finder)
 {
     if (mqi->orig->mailbox) {
-	GList * messages = g_list_prepend(NULL, mqi->orig);
+        gboolean remove = TRUE;
+        GList* fcclist =
+            libbalsa_message_find_user_hdr(mqi->orig, "X-Balsa-Fcc");
+        const gchar **fccurl = fcclist ? fcclist->data : NULL;
 
-	libbalsa_messages_change_flag(messages, 
-                                      LIBBALSA_MESSAGE_FLAG_DELETED,
-                                      TRUE);
-	g_list_free(messages);
+        lbs_change_flags(mqi->orig, 0, LIBBALSA_MESSAGE_FLAG_NEW |
+                         LIBBALSA_MESSAGE_FLAG_FLAGGED);
+	libbalsa_mailbox_sync_storage(mqi->orig->mailbox, FALSE);
+
+        if (mqi->orig->mailbox && fccurl) {
+            LibBalsaMailbox *fccbox = mqi->finder(fccurl[1]);
+            remove =
+                libbalsa_mailbox_copy_message(mqi->orig, fccbox, NULL)>=0;
+        }
+        /* If copy failed, mark the message again as flagged -
+           otherwise it will get resent again. And again, and
+           again... */
+        lbs_change_flags(mqi->orig, remove ?
+                         LIBBALSA_MESSAGE_FLAG_DELETED :
+                         LIBBALSA_MESSAGE_FLAG_FLAGGED, 0);
     }
     mqi->status = MQI_SENT;
 }
@@ -1128,13 +1174,21 @@ balsa_send_message_real(SendMessageInfo* info)
             libbalsa_information
                 (LIBBALSA_INFORMATION_ERROR,
                  _("SMTP server refused connection.\n"
-                   "Balsa by default uses submission service (587).\n"
-                   "If you want to submit mail using relay service (25),"
-                   "specify it explicitly via: \"host:smtp\".\n"
-                   "Message is left in outbox."));
+                   "Check your internet connection."));
+        case -EHOSTUNREACH:
+            libbalsa_information
+                (LIBBALSA_INFORMATION_ERROR,
+                 _("SMTP server cannot be reached.\n"
+                   "Check your internet connection."));
             break;
         case SMTP_ERR_NOTHING_TO_DO: /* silence this one? */
             break;
+        case SMTP_ERR_EAI_AGAIN:
+            /* this is really problem with DNS but it is nowadays
+               caused by general connection problems. */
+            libbalsa_information(LIBBALSA_INFORMATION_MESSAGE,
+                                 _("Message left in Outbox (try again later)"));
+            break;
         default:
             libbalsa_information (LIBBALSA_INFORMATION_ERROR,
                                   _("SMTP server problem (%d): %s\n"
@@ -1150,10 +1204,6 @@ balsa_send_message_real(SendMessageInfo* info)
        on the messages with a 2xx status recorded against them.  However
        its possible for individual recipients to fail too.  Need a way to
        report it all.  */
-    /* Do we really need to grab the gdk lock? If so, we'd have to drop
-     * it again before calling libbalsa_messages_change_flag().
-     * gdk_threads_enter();
-     */
     smtp_enumerate_messages (info->session, handle_successful_send, 
                              &session_started);
 
@@ -1250,7 +1300,7 @@ balsa_send_message_real(SendMessageInfo* info)
 	    g_mime_stream_file_set_owner(GMIME_STREAM_FILE(out), FALSE);
 	    if (g_mime_stream_write_to_stream(mqi->stream, out) == -1)
 		mqi->status = MQI_FAILED;
-	    g_mime_stream_unref(out);
+	    g_object_unref(out);
 	    if (pclose(sendmail) != 0)
 		mqi->status = MQI_FAILED;
 	    if (mqi->status != MQI_FAILED)
@@ -1266,7 +1316,7 @@ balsa_send_message_real(SendMessageInfo* info)
     
     while (mqi != NULL) {
 	if (mqi->status == MQI_SENT) 
-	    handle_successful_send(mqi);
+	    handle_successful_send(mqi, info->finder);
 	next_message = mqi->next_message;
 	msg_queue_item_destroy(mqi);
 	mqi = next_message;
@@ -1274,7 +1324,11 @@ balsa_send_message_real(SendMessageInfo* info)
     
     gdk_threads_enter();
     libbalsa_mailbox_close(info->outbox, TRUE);
+#if GTK_CHECK_VERSION(2, 4, 0)
+    gdk_display_flush(gdk_display_get_default());
+#else
     gdk_flush();
+#endif
     gdk_threads_leave();
 
     message_queue = NULL;
@@ -1332,6 +1386,7 @@ get_mailbox_names(GList *list, GList *address_list)
 }
 #endif
 
+#if !NEW_CHARSET_WIDGET
 /* Check whether a file is all ascii or utf-8, and return charset
  * accordingly (NULL if it's neither).
  */
@@ -1387,6 +1442,7 @@ lbs_file_get_charset(const gchar * filename)
     else
 	return NULL;
 }
+#endif                          /* NEW_CHARSET_WIDGET */
 
 /* We could have used g_strsplit_set(s, "/;", 3) but it is not
  * available in older glib. */
@@ -1422,44 +1478,67 @@ libbalsa_message_create_mime_message(LibBalsaMessage* message, gint encoding,
 	mime_root=GMIME_OBJECT(g_mime_multipart_new_with_subtype(message->subtype));
 
     while (body) {
-	GMimePart *mime_part;
+	GMimeObject *mime_part;
 	mime_part=NULL;
 
 	if (body->filename) {
+	    if (body->content_type) {
+		mime_type = parse_content_type(body->content_type);
+	    } else {
+		gchar* mt = libbalsa_lookup_mime_type(body->filename);
+		mime_type = g_strsplit(mt,"/", 2);
+		g_free(mt);
+	    }
+
 	    if (body->attach_as_extbody) {
-		GMimeObject *mime_obj;
-		mime_part=g_mime_part_new_with_type("message", "external-body");
-		g_mime_part_set_encoding(mime_part, GMIME_PART_ENCODING_7BIT);
-		mime_obj = GMIME_OBJECT(mime_part);
+		GMimeContentType *content_type =
+		    g_mime_content_type_new("message", "external-body");
+		mime_part=g_mime_object_new_type("message", "external-body");
+		g_mime_object_set_content_type(mime_part, content_type);
+		g_mime_part_set_encoding(GMIME_PART(mime_part),
+			                 GMIME_PART_ENCODING_7BIT);
 		if(!strncmp( body->filename, "URL", 3 )) {
-		    g_mime_object_set_content_type_parameter(mime_obj,
+		    g_mime_object_set_content_type_parameter(mime_part,
 					     "access-type", "URL");
-		    g_mime_object_set_content_type_parameter(mime_obj,
+		    g_mime_object_set_content_type_parameter(mime_part,
 					     "URL", body->filename + 4);
 		} else {
-		    g_mime_object_set_content_type_parameter(mime_obj,
+		    g_mime_object_set_content_type_parameter(mime_part,
 					     "access-type", "local-file");
-		    g_mime_object_set_content_type_parameter(mime_obj,
+		    g_mime_object_set_content_type_parameter(mime_part,
 					     "name", body->filename);
 		}
-		g_mime_part_set_content(mime_part,
-					"Note: this is _not_ the real body!\n",
-					34);
+		lbs_set_content(GMIME_PART(mime_part),
+                                "Note: this is _not_ the real body!\n");
+	    } else if (g_ascii_strcasecmp(mime_type[0], "message") == 0) {
+		int fd;
+		GMimeStream *stream;
+		GMimeParser *parser;
+		GMimeMessage *mime_message;
+
+		fd = open(body->filename, O_RDONLY);
+		stream = g_mime_stream_fs_new(fd);
+		parser = g_mime_parser_new_with_stream(stream);
+		g_object_unref(stream);
+		mime_message = g_mime_parser_construct_message(parser);
+		g_object_unref(parser);
+                mime_part =
+                    GMIME_OBJECT(g_mime_message_part_new_with_message
+                                 (mime_type[1], mime_message));
+		g_object_unref(mime_message);
 	    } else {
 		const gchar *charset = NULL;
 		GMimeStream *stream;
 		GMimeDataWrapper *content;
 		int fd;
 
-		if (!body->content_type) {
-		    gchar* mt = libbalsa_lookup_mime_type(body->filename);
-		    mime_type = g_strsplit(mt,"/", 2);
-		    g_free(mt);
-		} else
-		    mime_type = parse_content_type(body->content_type);
-
-		if (!strcasecmp(mime_type[0], "text")) {
+		if (!strcasecmp(mime_type[0], "text")
+		    && !(charset = body->charset)) {
+#if NEW_CHARSET_WIDGET
+		    charset = libbalsa_file_get_charset(body->filename);
+#else                           /* NEW_CHARSET_WIDGET */
 		    charset = lbs_file_get_charset(body->filename);
+#endif                          /* NEW_CHARSET_WIDGET */
 		    if (!charset) {
 			static const gchar default_type[] =
 			    "application/octet-stream";
@@ -1476,36 +1555,36 @@ libbalsa_message_create_mime_message(LibBalsaMessage* message, gint encoding,
 
 		/* use BASE64 encoding for non-text mime types 
 		   use 8BIT for message */
-		mime_part=g_mime_part_new_with_type(mime_type[0], mime_type[1]);
-		g_mime_part_set_content_disposition(mime_part, GMIME_DISPOSITION_ATTACHMENT);
-		if(!strcasecmp(mime_type[0],"message") && 
-		   !strcasecmp(mime_type[1],"rfc822")) {
-		    g_mime_part_set_encoding(mime_part, GMIME_PART_ENCODING_8BIT);
-		    g_mime_part_set_content_disposition(mime_part, GMIME_DISPOSITION_INLINE);
-		} else if(strcasecmp(mime_type[0],"text") != 0)
+		mime_part =
+		    GMIME_OBJECT(g_mime_part_new_with_type(mime_type[0],
+				                           mime_type[1]));
+		g_mime_part_set_content_disposition(GMIME_PART(mime_part),
+			GMIME_DISPOSITION_ATTACHMENT);
+		if(strcasecmp(mime_type[0],"text") != 0)
 		{
-		    g_mime_part_set_encoding(mime_part, GMIME_PART_ENCODING_BASE64);
+		    g_mime_part_set_encoding(GMIME_PART(mime_part),
+			    GMIME_PART_ENCODING_BASE64);
 		} else {
 		    /* is text */
-		    g_mime_object_set_content_type_parameter(GMIME_OBJECT
-							     (mime_part),
+		    g_mime_object_set_content_type_parameter(mime_part,
 							     "charset",
 							     charset);
 		}
-		g_strfreev(mime_type);
 
 		tmp = g_path_get_basename(body->filename);
-		g_mime_part_set_filename(mime_part, tmp);
+		g_mime_part_set_filename(GMIME_PART(mime_part), tmp);
 		g_free(tmp);
 
 		fd = open(body->filename, O_RDONLY);
 		stream = g_mime_stream_fs_new(fd);
 		content = g_mime_data_wrapper_new_with_stream(stream,
 			GMIME_PART_ENCODING_DEFAULT);
-		g_mime_stream_unref(stream);
-		g_mime_part_set_content_object(mime_part, content);
+		g_object_unref(stream);
+		g_mime_part_set_content_object(GMIME_PART(mime_part),
+			                       content);
 		g_object_unref(content);
 	    }
+	    g_strfreev(mime_type);
 	} else if (body->buffer) {
 #ifdef HAVE_GPGME
 	    /* force quoted printable encoding if only signing is requested */
@@ -1519,7 +1598,8 @@ libbalsa_message_create_mime_message(LibBalsaMessage* message, gint encoding,
 		message->gpg_mode > 0 &&
 		(message->gpg_mode & LIBBALSA_PROTECT_OPENPGP) != 0) {
 		gint result = 
-		    libbalsa_create_rfc2440_buffer(body, mime_part);
+		    libbalsa_create_rfc2440_buffer(body,
+			                           GMIME_PART(mime_part));
 
 		if (result != LIBBALSA_MESSAGE_CREATE_OK) {
 		    g_object_unref(G_OBJECT(mime_part));
@@ -1537,7 +1617,7 @@ libbalsa_message_create_mime_message(LibBalsaMessage* message, gint encoding,
 				      GMIME_OBJECT(mime_part));
 	    g_object_unref(G_OBJECT(mime_part));
 	} else {
-	    mime_root = GMIME_OBJECT(mime_part);
+	    mime_root = mime_part;
 	}
 
 	body = body->next;
@@ -1654,7 +1734,7 @@ libbalsa_message_postpone(LibBalsaMessage * message,
     if (fcc)
 	g_mime_message_set_header(message->mime_msg, "X-Balsa-Fcc", fcc);
 
-    thereturn = libbalsa_mailbox_copy_message( message, draftbox );
+    thereturn = libbalsa_mailbox_copy_message( message, draftbox, NULL );
 
     return thereturn!=-1;
 }
@@ -1715,12 +1795,24 @@ libbalsa_create_msg(LibBalsaMessage * message,
 
 static LibBalsaMsgCreateResult
 libbalsa_fill_msg_queue_item_from_queu(LibBalsaMessage * message,
-				 MessageQueueItem *mqi)
+                                       MessageQueueItem *mqi)
 {
     mqi->orig = message;
     g_object_ref(mqi->orig);
-    mqi->stream = libbalsa_mailbox_get_message_stream(message->mailbox,
-						      message);
+    if (message->mime_msg) {
+        g_mime_object_remove_header(GMIME_OBJECT(message->mime_msg),
+                                    "Status");
+        g_mime_object_remove_header(GMIME_OBJECT(message->mime_msg),
+                                    "X-Status");
+        g_mime_object_remove_header(GMIME_OBJECT(message->mime_msg),
+                                    "X-Balsa-Fcc");
+	mqi->stream = g_mime_stream_mem_new();
+	g_mime_object_write_to_stream(GMIME_OBJECT(message->mime_msg),
+                                      mqi->stream);
+	g_mime_stream_reset(mqi->stream);
+    } else
+	mqi->stream = libbalsa_mailbox_get_message_stream(message->mailbox,
+                                                          message);
     if (mqi->stream == NULL)
 	return LIBBALSA_MESSAGE_CREATE_ERROR;
   
@@ -1758,7 +1850,7 @@ libbalsa_create_rfc2440_buffer(LibBalsaMessageBody *body, GMimePart *mime_part)
 					    message->headers->from->address_list->data);
 		if (message->headers->bcc_list)
 		    libbalsa_information(LIBBALSA_INFORMATION_WARNING,
-					 _("This message will not be encrpyted for the BCC: recipient(s)."));
+					 _("This message will not be encrypted for the BCC: recipient(s)."));
 
 		if (mode & LIBBALSA_PROTECT_SIGN)
 		    result = 
@@ -1834,7 +1926,7 @@ do_multipart_crypto(LibBalsaMessage * message, GMimeObject ** mime_root)
 					    message->headers->from->address_list->data);
 		if (message->headers->bcc_list)
 		    libbalsa_information(LIBBALSA_INFORMATION_WARNING,
-					 _("This message will not be encrpyted for the BCC: recipient(s)."));
+					 _("This message will not be encrypted for the BCC: recipient(s)."));
 
 		if (message->gpg_mode & LIBBALSA_PROTECT_SIGN)
 		    success = 
