@@ -26,8 +26,7 @@
 #include "libbalsa_private.h"
 #include "libbalsa-marshal.h"
 
-#include <libgnome/gnome-config.h> 
-#include <libgnome/gnome-i18n.h> 
+#include <libgnome/libgnome.h> 
 #include <string.h>
 #include <stdlib.h>
 
