@@ -657,10 +657,6 @@ balsa_find_mailbox(LibBalsaMailbox * mailbox)
     BalsaFind bf;
     gboolean is_sub_thread = libbalsa_am_i_subthread();
 
-#ifdef BALSA_USE_THREADS
-    g_assert(!g_mutex_trylock(gdk_threads_mutex));
-#endif
-
     if (is_sub_thread)
 	gdk_threads_enter();
 
@@ -775,6 +771,13 @@ balsa_find_mailbox_by_url(const gchar * url)
     return mailbox;
 }
 
+LibBalsaMailbox*
+balsa_find_sentbox_by_url(const gchar *url)
+{
+    LibBalsaMailbox *res = balsa_find_mailbox_by_url(url);
+    return res ? res : balsa_app.sentbox;
+}
+
 struct balsa_find_iter_by_data_info {
     GtkTreeIter *iter;
     gpointer data;
