@@ -33,7 +33,9 @@
 #define BALSA_PIXMAP_REPLY_GROUP		"balsa_reply_group"
 #define BALSA_PIXMAP_FORWARD			"balsa_forward"
 #define BALSA_PIXMAP_NEXT			"balsa_next"
+#define BALSA_PIXMAP_NEXT_PART			"balsa_next_part"
 #define BALSA_PIXMAP_PREVIOUS			"balsa_previous"
+#define BALSA_PIXMAP_PREVIOUS_PART			"balsa_previous_part"
 #define BALSA_PIXMAP_POSTPONE			"balsa_postpone"
 #define BALSA_PIXMAP_PRINT			"balsa_print"
 #define BALSA_PIXMAP_SAVE			"balsa_save"
@@ -60,7 +62,9 @@
 #define BALSA_PIXMAP_MENU_POSTPONE		"balsa_menu_postpone"
 #define BALSA_PIXMAP_MENU_PRINT	                "balsa_menu_print"
 #define BALSA_PIXMAP_MENU_NEXT			"balsa_menu_next"
+#define BALSA_PIXMAP_MENU_NEXT_PART			"balsa_menu_next_part"
 #define BALSA_PIXMAP_MENU_PREVIOUS		"balsa_menu_previous"
+#define BALSA_PIXMAP_MENU_PREVIOUS_PART		"balsa_menu_previous_part"
 #define BALSA_PIXMAP_MENU_SAVE			"balsa_menu_save"
 #define BALSA_PIXMAP_MENU_SEND			"balsa_menu_send"
 #define BALSA_PIXMAP_MENU_SEND_RECEIVE		"balsa_menu_send_receive"
@@ -91,6 +95,7 @@
 #ifdef HAVE_GPGME
 #  define BALSA_PIXMAP_GPG_SIGN                 "balsa_gpg_sign"
 #  define BALSA_PIXMAP_GPG_ENCRYPT              "balsa_gpg_encrypt"
+#  define BALSA_PIXMAP_GPG_RECHECK              "balsa_gpg_recheck"
 #  define BALSA_PIXMAP_INFO_SIGN                "balsa_info_sign"
 #  define BALSA_PIXMAP_INFO_SIGN_GOOD           "balsa_info_sign_good"
 #  define BALSA_PIXMAP_INFO_SIGN_NOTRUST        "balsa_info_sign_trust"
