@@ -30,13 +30,6 @@
    key and replace the key on every change of the sort method.  
 */
 
-/* TREE_VIEW_FIXED_HEIGHT enables hight-performance mode of GtkTreeView
- * very useful for large mailboxes (#msg >5000) but: a. is available only
- * in gtk2>=2.3.5 b. may expose some bugs in gtk.
- */
-#define TREE_VIEW_FIXED_HEIGHT 1
-#undef TREE_VIEW_FIXED_HEIGHT
-
 #include "config.h"
 
 #include <stdio.h>
@@ -57,6 +50,16 @@
 #include "filter-funcs.h"
 #include "misc.h"
 
+/* TREE_VIEW_FIXED_HEIGHT enables hight-performance mode of GtkTreeView
+ * very useful for large mailboxes (#msg >5000) but: a. is available only
+ * in gtk2>=2.3.5 b. may expose some bugs in gtk.
+ * gtk-2.4.9 has been tested with a positive result.
+ */
+#if GTK_CHECK_VERSION(2,4,9)
+#define TREE_VIEW_FIXED_HEIGHT 1
+#endif
+
+
 /* gtk widget */
 static void bndx_class_init(BalsaIndexClass * klass);
 static void bndx_instance_init(BalsaIndex * index);
@@ -78,9 +81,20 @@ static void bndx_mailbox_changed_cb(BalsaIndex * index);
 /* GtkTree* callbacks */
 static void bndx_selection_changed(GtkTreeSelection * selection,
                                    gpointer data);
+
+struct BndxSelectionChangedInfo {
+    guint msgno;
+    guint current_msgno;
+    gboolean current_selected;
+    GArray *selected;
+    GArray *new_selected;
+};
+
 static void bndx_selection_changed_func(GtkTreeModel * model,
                                         GtkTreePath * path,
-                                        GtkTreeIter * iter, gpointer data);
+                                        GtkTreeIter * iter,
+                                        struct BndxSelectionChangedInfo
+                                        *sci);
 static gboolean bndx_button_event_press_cb(GtkWidget * tree_view,
                                            GdkEventButton * event,
                                            gpointer data);
@@ -221,6 +235,7 @@ bndx_destroy(GtkObject * obj)
 
 	    libbalsa_mailbox_search_iter_free(index->search_iter);
 	    index->search_iter = NULL;
+	    libbalsa_mailbox_unregister_msgnos(mailbox, index->selected);
 	}
 	g_object_weak_unref(G_OBJECT(index->mailbox_node),
 			    (GWeakNotify) gtk_widget_destroy, index);
@@ -233,14 +248,12 @@ bndx_destroy(GtkObject * obj)
     }
 
     if (index->current_message) {
-	g_object_remove_weak_pointer(G_OBJECT(index->current_message),
-				     (gpointer)&index->current_message);
+	g_object_unref(index->current_message);
 	index->current_message = NULL;
     }
 
     if (index->selected) {
-	g_slist_foreach(index->selected, (GFunc) g_object_unref, NULL);
-	g_slist_free(index->selected);
+	g_array_free(index->selected, TRUE);
 	index->selected = NULL;
     }
 
@@ -413,6 +426,9 @@ bndx_instance_init(BalsaIndex * index)
     g_signal_connect_after(tree_view, "size-allocate",
                            G_CALLBACK(bndx_column_resize),
                            NULL);
+#if GTK_CHECK_VERSION(2,4,9)
+    gtk_tree_view_set_enable_search(tree_view, FALSE);
+#endif
 
     gtk_drag_source_set(GTK_WIDGET (index), 
                         GDK_BUTTON1_MASK | GDK_SHIFT_MASK | GDK_CONTROL_MASK,
@@ -424,6 +440,7 @@ bndx_instance_init(BalsaIndex * index)
 
     balsa_index_set_column_widths(index);
     gtk_widget_show_all (GTK_WIDGET(index));
+    index->selected = g_array_new(FALSE, FALSE, sizeof(guint));
 }
 
 /* Callbacks used by bndx_instance_init. */
@@ -439,11 +456,6 @@ bndx_instance_init(BalsaIndex * index)
 
 /* First some helpers. */
 
-struct BndxSelectionChangedInfo {
-    LibBalsaMessage *message;
-    GSList **selected;
-};
-
 #if !GTK_CHECK_VERSION(2, 2, 0)
 static void
 bndx_selection_any(GtkTreeModel * model, GtkTreePath * path,
@@ -455,81 +467,94 @@ bndx_selection_any(GtkTreeModel * model, GtkTreePath * path,
 #endif				/* !GTK_CHECK_VERSION(2, 2, 0) */
 
 static void
-bndx_filt_check_free(GSList ** check_list)
+bndx_deselected_free(GArray * deselected)
 {
-    g_slist_foreach(*check_list, (GFunc) g_object_unref, NULL);
-    g_slist_free(*check_list);
-    g_free(check_list);
+    g_array_free(deselected, TRUE);
 }
 
-#define BALSA_INDEX_CHECK_LIST "balsa-index-check-list"
+#define BALSA_INDEX_DESELECTED_ARRAY "balsa-index-deselected-array"
 static gboolean
-bndx_filt_check(BalsaIndex * index)
+bndx_deselected_idle(BalsaIndex * index)
 {
-    GSList **check_list;
-    GSList *l;
-    LibBalsaMailboxSearchIter *iter_view;
+    GArray *deselected;
+    LibBalsaMessage *current_message;
 
     gdk_threads_enter();
 
-    check_list =
-        g_object_get_data(G_OBJECT(index), BALSA_INDEX_CHECK_LIST);
-    g_assert(check_list != NULL);
+    deselected =
+        g_object_get_data(G_OBJECT(index), BALSA_INDEX_DESELECTED_ARRAY);
+    g_assert(deselected != NULL);
 
     /* Check whether the index was destroyed. */
-    if (index->mailbox_node) {
-        /* Search results may be cached in iter_view. */
-        iter_view =
-            libbalsa_mailbox_search_iter_view(index->mailbox_node->
-                                              mailbox);
-        if (iter_view) {
-	    if (index->current_message)
-		g_object_ref(index->current_message);
-            for (l = *check_list; l; l = l->next) {
-                LibBalsaMessage *message = l->data;
-                libbalsa_mailbox_msgno_filt_check(message->mailbox,
-                                                  message->msgno,
-                                                  iter_view);
-            }
-	    if (index->current_message) {
-		GtkTreePath *path;
-		if (bndx_find_message(index, &path, NULL,
-				      index->current_message)) {
-		    GtkTreeSelection *selection =
-			gtk_tree_view_get_selection(GTK_TREE_VIEW(index));
-		    if (!gtk_tree_selection_path_is_selected(selection,
-							     path)) {
-			bndx_expand_to_row(index, path);
-			bndx_select_row(index, path);
-		    }
-		    gtk_tree_path_free(path);
-		} /* else ??? */
-		g_object_unref(index->current_message);
-	    }
-            libbalsa_mailbox_search_iter_free(iter_view);
-        }
+    if (!index->mailbox_node) {
+        g_object_set_data(G_OBJECT(index), BALSA_INDEX_DESELECTED_ARRAY,
+                          NULL);
+        g_object_unref(index);
+        gdk_threads_leave();
+        return FALSE;
     }
 
-    g_object_set_data(G_OBJECT(index), BALSA_INDEX_CHECK_LIST, NULL);
+    if ((current_message = index->current_message))
+	/* Do not allow current message to be finalized. */
+	g_object_ref(current_message);
+
+    libbalsa_mailbox_messages_change_flags(index->mailbox_node->mailbox,
+                                           deselected, 0,
+                                           LIBBALSA_MESSAGE_FLAG_SELECTED);
+
+    if (current_message) {
+	/* We had a message... */
+	if (!index->current_message)
+	    /* ...but we don't any more; restore it. */
+	    index->current_message = current_message;
+	else
+	    /* Discard. */
+	    g_object_unref(current_message);
+    }
+
+    if (index->current_message) {
+        GtkTreePath *path;
+        if (bndx_find_message(index, &path, NULL, index->current_message)) {
+            GtkTreeSelection *selection =
+                gtk_tree_view_get_selection(GTK_TREE_VIEW(index));
+            if (!gtk_tree_selection_path_is_selected(selection, path)) {
+                bndx_expand_to_row(index, path);
+                bndx_select_row(index, path);
+            }
+            gtk_tree_path_free(path);
+        } else {
+	    /* Can this happen? */
+	    g_object_unref(index->current_message);
+	    index->current_message = NULL;
+	}
+    }
+
+    g_object_set_data(G_OBJECT(index), BALSA_INDEX_DESELECTED_ARRAY, NULL);
     g_object_unref(index);
 
     gdk_threads_leave();
     return FALSE;
 }
 
+static int
+bndx_compare_guint(gconstpointer a, gconstpointer b)
+{
+    return *(guint *)a - *(guint *)b;
+}
+
 static void
 bndx_selection_changed(GtkTreeSelection * selection, gpointer data)
 {
     BalsaIndex *index = BALSA_INDEX(data);
     GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(index));
+    LibBalsaMailbox *mailbox = LIBBALSA_MAILBOX(model);
     struct BndxSelectionChangedInfo sci;
-    GtkTreeIter iter;
-    GSList *list;
-    GSList *next;
     gboolean have_selected;
-    GSList **check_list;
+    GArray *deselected;
+    gint i;
+    gint current_depth = 0;
 
-    if (LIBBALSA_MAILBOX(model)->state == LB_MAILBOX_STATE_TREECLEANING)
+    if (mailbox->state == LB_MAILBOX_STATE_TREECLEANING)
         return;
 
 #if GTK_CHECK_VERSION(2, 2, 0)
@@ -542,22 +567,20 @@ bndx_selection_changed(GtkTreeSelection * selection, gpointer data)
 					&have_selected);
 #endif /* GTK_CHECK_VERSION(2, 2, 0) */
 
-    check_list = g_object_get_data(G_OBJECT(index), BALSA_INDEX_CHECK_LIST);
-
-    /* Check currently selected messages. */
-    for (list = index->selected; list; list = next) {
+    /* Check previously selected messages. */
+    deselected =
+	g_object_get_data(G_OBJECT(index), BALSA_INDEX_DESELECTED_ARRAY);
+    sci.current_msgno =
+        index->current_message ? index->current_message->msgno : 0;
+    for (i = index->selected->len; --i >= 0; ) {
 	GtkTreePath *path;
-	LibBalsaMessage *message = list->data;
+	guint msgno = g_array_index(index->selected, guint, i);
 
-	next = list->next;
-	if (!bndx_find_message(index, &path, NULL, message)) {
-	    /* The message is no longer in the view--it must have been
-	     * either expunged or filtered out--in either case, we just
-	     * drop it from the list. */
-	    g_object_unref(message);
-	    index->selected = g_slist_delete_link(index->selected, list);
+	if (!libbalsa_mailbox_msgno_find(mailbox, msgno, &path, NULL)) {
+	    g_array_remove_index(index->selected, i);
 	    continue;
 	}
+
 	/* Remove a deselected message from the list, unless it's not
 	 * viewable and no rows are currently selected. */
 	if (!gtk_tree_selection_path_is_selected(selection, path)
@@ -565,58 +588,95 @@ bndx_selection_changed(GtkTreeSelection * selection, gpointer data)
 	    /* The message has been deselected, and not by collapsing a
 	     * thread; we'll notify the mailbox, so it can check whether
 	     * the message still matches the view filter. */
-
-	    index->selected = g_slist_delete_link(index->selected, list);
-
 	    /* Check filtering in an idle handler. */
-	    if (!check_list) {
-		check_list = g_new0(GSList *, 1);
+	    if (!deselected) {
+		deselected = g_array_new(FALSE, FALSE, sizeof(guint));
 		g_object_set_data_full(G_OBJECT(index),
-				       BALSA_INDEX_CHECK_LIST, check_list,
-				       (GDestroyNotify) bndx_filt_check_free);
+                                       BALSA_INDEX_DESELECTED_ARRAY,
+				       deselected,
+				       (GDestroyNotify) bndx_deselected_free);
 		g_object_ref(index);
-		g_idle_add((GSourceFunc) bndx_filt_check, index);
+		g_idle_add((GSourceFunc) bndx_deselected_idle, index);
 	    }
-	    *check_list = g_slist_prepend(*check_list, message);
+	    g_array_append_val(deselected, msgno);
+	    g_array_remove_index(index->selected, i);
 	}
+	if (msgno == sci.current_msgno)
+	    current_depth = gtk_tree_path_get_depth(path);
 	gtk_tree_path_free(path);
     }
 
-    sci.selected = &index->selected;
-    sci.message = NULL;
+    /* Check currently selected messages. */
+    g_array_sort(index->selected, bndx_compare_guint);
+    sci.selected = index->selected;
+    sci.new_selected = g_array_new(FALSE, FALSE, sizeof(guint));
+    sci.msgno = 0;
+    sci.current_selected = FALSE;
     gtk_tree_selection_selected_foreach(selection,
-                                        bndx_selection_changed_func,
+                                        (GtkTreeSelectionForeachFunc)  
+					bndx_selection_changed_func,
                                         &sci);
-    if (g_slist_find(index->selected, index->current_message))
-        return;
+    libbalsa_mailbox_messages_change_flags(mailbox, sci.new_selected,
+                                           LIBBALSA_MESSAGE_FLAG_SELECTED,
+                                           0);
+    g_array_append_vals(index->selected, sci.new_selected->data,
+	                sci.new_selected->len);
+    g_array_free(sci.new_selected, TRUE);
 
-    /* we don't clear the current message if the tree contains any
-     * messages */
-    if (sci.message || !gtk_tree_model_get_iter_first(model, &iter)) {
-	if (index->current_message)
-	    g_object_remove_weak_pointer(G_OBJECT(index->current_message),
-					 (gpointer)&index->current_message);
-        index->current_message = sci.message;
-        if(index->current_message)
-            g_object_add_weak_pointer(G_OBJECT(index->current_message),
-                                      (gpointer)&index->current_message);
+    if (sci.current_selected)
+	return;
+
+    /* sci.msgno is the msgno of the new current message;
+     * if sci.msgno == 0, either:
+     * - no messages remain in the view, in which case we clear
+     *   index->current_message,
+     * or:
+     * - the thread containing the current message was collapsed, in
+     *   which case we leave index->current_message unchanged;
+     * we detect the latter case by checking the depth of the current
+     * message.  */
+    if (sci.msgno || current_depth <= 1) {
+	if (index->current_message) {
+	    g_object_unref(index->current_message);
+	    index->current_message = NULL;
+	}
+        if (sci.msgno
+            && (index->current_message =
+                libbalsa_mailbox_get_message(mailbox, sci.msgno)))
+	    index->current_message_is_deleted =
+		LIBBALSA_MESSAGE_IS_DELETED(index->current_message);
         bndx_changed_find_row(index);
     }
 }
 
+/* Binary search for msgno in array. */
+static gboolean
+bndx_msgno_is_in_array(guint msgno, GArray * array)
+{
+    gint high = array->len, low = -1;
+
+    while (high - low > 1) {
+        gint middle = (high + low) / 2;
+        if (g_array_index(array, guint, middle) > msgno)
+            high = middle;
+        else
+            low = middle;
+    }
+
+    return low >= 0 && g_array_index(array, guint, low) == msgno;
+}
+
 static void
 bndx_selection_changed_func(GtkTreeModel * model, GtkTreePath * path,
-                            GtkTreeIter * iter, gpointer data)
+                            GtkTreeIter * iter,
+			    struct BndxSelectionChangedInfo *sci)
 {
-    struct BndxSelectionChangedInfo *sci = data;
-
-    gtk_tree_model_get(model, iter, LB_MBOX_MESSAGE_COL, &sci->message,
-                       -1);
-    if (!g_slist_find(*sci->selected, sci->message))
-	*sci->selected = g_slist_prepend(*sci->selected, sci->message);
-    else {
-	g_object_unref(sci->message);
-    }
+    gtk_tree_model_get(model, iter, LB_MBOX_MSGNO_COL, &sci->msgno, -1);
+    if (sci->msgno == sci->current_msgno)
+	sci->current_selected = TRUE;
+    if (bndx_msgno_is_in_array(sci->msgno, sci->selected))
+	return;
+    g_array_append_val(sci->new_selected, sci->msgno);
 }
 
 static gboolean
@@ -692,29 +752,33 @@ bndx_tree_expand_cb(GtkTreeView * tree_view, GtkTreeIter * iter,
 {
     BalsaIndex *index = BALSA_INDEX(tree_view);
     GtkTreeSelection *selection = gtk_tree_view_get_selection(tree_view);
+    LibBalsaMailbox *mailbox = index->mailbox_node->mailbox;
     GtkTreePath *current_path;
-    GSList *list;
+    gint i;
 
     g_signal_handler_block(selection, index->selection_changed_id);
     /* If a message in the selected list has become viewable, reselect
      * it. */
-    for (list = index->selected; list; list = list->next) {
-	LibBalsaMessage *message = list->data;
+    for (i = index->selected->len; --i >= 0;) {
+        guint msgno = g_array_index(index->selected, guint, i);
 
-	if (bndx_find_message(index, &current_path, NULL, message)) {
-
-	    if (!gtk_tree_selection_path_is_selected
-		(selection, current_path)
-		&& bndx_row_is_viewable(index, current_path)) {
-		gtk_tree_selection_select_path(selection, current_path);
-		if (message == index->current_message) {
-		    gtk_tree_view_scroll_to_cell(GTK_TREE_VIEW(index),
-						 current_path, NULL, FALSE,
-						 0, 0);
-		}
-	    }
-	    gtk_tree_path_free(current_path);
-	}
+        if (libbalsa_mailbox_msgno_find(mailbox, msgno, &current_path,
+                                        NULL)) {
+            if (!gtk_tree_selection_path_is_selected(selection,
+                                                     current_path)
+                && bndx_row_is_viewable(index, current_path)) {
+                gtk_tree_selection_select_path(selection, current_path);
+                if (index->current_message
+		    && msgno == (guint) index->current_message->msgno) {
+		    gtk_tree_view_set_cursor(tree_view, current_path,
+			                     NULL, FALSE);
+                    gtk_tree_view_scroll_to_cell(GTK_TREE_VIEW(index),
+                                                 current_path, NULL, FALSE,
+                                                 0, 0);
+                }
+            }
+            gtk_tree_path_free(current_path);
+        }
     }
     g_signal_handler_unblock(selection, index->selection_changed_id);
 }
@@ -761,31 +825,15 @@ bndx_drag_cb (GtkWidget* widget, GdkDragContext* drag_context,
                GtkSelectionData* data, guint info, guint time, 
                gpointer user_data)
 { 
-    LibBalsaMessage* message;
-    GPtrArray* message_array = NULL;
-    GList* list, *l;
+    BalsaIndex *index;
     
-
     g_return_if_fail (widget != NULL);
-    
-    message_array = g_ptr_array_new ();
-    l = balsa_index_selected_list(BALSA_INDEX(widget));
-    for (list = l; list; list = g_list_next(list)) {
-        message = list->data;
-        g_ptr_array_add (message_array, message);
-    }
-    g_list_foreach(l, (GFunc)g_object_unref, NULL); /* FIXME: too early? */
-    g_list_free(l);
-    
-    if (message_array) {
-        g_ptr_array_add (message_array, NULL);
-        gtk_selection_data_set (data, data->target, 8, 
-                                (guchar*) message_array->pdata, 
-                                (message_array->len)*sizeof (gpointer));
-        /* the selection data makes a copy of the data, we 
-         * can free it now. */
-        g_ptr_array_free (message_array, TRUE);
-    }
+    index = BALSA_INDEX(widget);
+
+    if (index->selected->len > 0)
+        gtk_selection_data_set(data, data->target, 8,
+                               (const guchar *) &index,
+                               sizeof(BalsaIndex *));
 }
 
 /* Public methods */
@@ -943,6 +991,7 @@ balsa_index_load_mailbox_node (BalsaIndex * index,
     g_free(msg);
     try_cnt = 0;
     do {
+        g_clear_error(err);
         gdk_threads_leave();
         successp = libbalsa_mailbox_open(mailbox, err);
         gdk_threads_enter();
@@ -952,7 +1001,6 @@ balsa_index_load_mailbox_node (BalsaIndex * index,
            !(*err && (*err)->code == LIBBALSA_MAILBOX_TOOMANYOPEN_ERROR))
             break;
         balsa_mblist_close_lru_peer_mbx(balsa_app.mblist, mailbox);
-        g_clear_error(err);
     } while(try_cnt++<3);
     gnome_appbar_pop(balsa_app.appbar);
 
@@ -965,6 +1013,7 @@ balsa_index_load_mailbox_node (BalsaIndex * index,
     index->mailbox_node = mbnode;
     g_object_weak_ref(G_OBJECT(mbnode),
 		      (GWeakNotify) gtk_widget_destroy, index);
+    libbalsa_mailbox_register_msgnos(mailbox, index->selected);
     /*
      * rename "from" column to "to" for outgoing mail
      */
@@ -986,7 +1035,11 @@ balsa_index_load_mailbox_node (BalsaIndex * index,
     /* libbalsa functions must be called with gdk unlocked
      * but balsa_index - locked!
      */
+#if GTK_CHECK_VERSION(2, 4, 0)
+    gdk_display_flush(gdk_display_get_default());
+#else
     gdk_flush();
+#endif
     gdk_threads_leave();
     libbalsa_mailbox_set_view_filter(mailbox,
                                      balsa_window_get_view_filter
@@ -1005,6 +1058,7 @@ balsa_index_load_mailbox_node (BalsaIndex * index,
 
     /* Create a search-iter for SEARCH UNDELETED. */
     index->search_iter = libbalsa_mailbox_search_iter_new(&cond_undeleted);
+    /* Note when this mailbox was opened, for use in auto-closing. */
     time(&index->mailbox_node->last_use);
 
     return FALSE;
@@ -1308,9 +1362,12 @@ bndx_mailbox_changed_func(BalsaIndex * bindex)
     if (!GTK_WIDGET_REALIZED(GTK_WIDGET(bindex)))
 	return;
 
-    if (bindex->current_message
-	&& LIBBALSA_MESSAGE_IS_DELETED(bindex->current_message))
-	bndx_select_next_threaded(bindex);
+    if (bindex->current_message) {
+	if (!LIBBALSA_MESSAGE_IS_DELETED(bindex->current_message))
+	    bindex->current_message_is_deleted = FALSE;
+	else if (!bindex->current_message_is_deleted)
+	    bndx_select_next_threaded(bindex);
+    }
 
     if (mailbox->first_unread
 	&& libbalsa_mailbox_msgno_find(mailbox, mailbox->first_unread,
@@ -1331,7 +1388,6 @@ bndx_mailbox_changed_func(BalsaIndex * bindex)
         gtk_tree_path_free(path);
     }
 
-    time(&bindex->mailbox_node->last_use);
     bndx_changed_find_row(bindex);
 }
 
@@ -1369,17 +1425,19 @@ bndx_mailbox_changed_cb(BalsaIndex * bindex)
 static void
 bndx_view_source(GtkWidget * widget, gpointer data)
 {
-    GList *messages = balsa_index_selected_list(BALSA_INDEX(data));
-    GList *list;
+    BalsaIndex *index = BALSA_INDEX(data);
+    LibBalsaMailbox *mailbox = index->mailbox_node->mailbox;
+    gint i;
 
-    for (list = messages; list; list = list->next) {
-	LibBalsaMessage *message = list->data;
+    for (i = index->selected->len; --i >= 0;) {
+        guint msgno = g_array_index(index->selected, guint, i);
+        LibBalsaMessage *message =
+            libbalsa_mailbox_get_message(mailbox, msgno);
 
 	libbalsa_show_message_source(message, balsa_app.message_font,
 				     &balsa_app.source_escape_specials);
+        g_object_unref(message);
     }
-    g_list_foreach(messages, (GFunc)g_object_unref, NULL);
-    g_list_free(messages);
 }
 
 static void
@@ -1435,17 +1493,19 @@ static void
 bndx_compose_foreach(GtkWidget * w, BalsaIndex * index,
                      SendType send_type)
 {
-    GList *list, *l = balsa_index_selected_list(index);
+    LibBalsaMailbox *mailbox = index->mailbox_node->mailbox;
+    gint i;
 
-    for (list = l; list; list = g_list_next(list)) {
-        LibBalsaMessage *message = list->data;
+    for (i = index->selected->len; --i >= 0;) {
+        guint msgno = g_array_index(index->selected, guint, i);
+        LibBalsaMessage *message =
+            libbalsa_mailbox_get_message(mailbox, msgno);
         BalsaSendmsg *sm = sendmsg_window_new(NULL, message, send_type);
 
         g_signal_connect(G_OBJECT(sm->window), "destroy",
                          G_CALLBACK(sendmsg_window_destroy_cb), NULL);
+        g_object_unref(message);
     }
-    g_list_foreach(l, (GFunc)g_object_unref, NULL);
-    g_list_free(l);
 }
 
 /*
@@ -1528,35 +1588,39 @@ static void
 bndx_do_delete(BalsaIndex* index, gboolean move_to_trash)
 {
     BalsaIndex *trash = balsa_find_index_by_mailbox(balsa_app.trash);
-    GList *messages = balsa_index_selected_list(index);
-    GList *list = messages;
+    GArray *messages;
+    LibBalsaMailbox *mailbox = index->mailbox_node->mailbox;
+    gint i;
 
-    while (list) {
-        GList *next = g_list_next(list);
-        LibBalsaMessage * message = list->data;
-
-        if (LIBBALSA_MESSAGE_IS_DELETED(message)) {
-            messages = g_list_delete_link(messages, list);
-	    g_object_unref(message);
-	}
-
-        list = next;
+    messages = g_array_new(FALSE, FALSE, sizeof(guint));
+    for (i = index->selected->len; --i >= 0;) {
+        guint msgno = g_array_index(index->selected, guint, i);
+	if (libbalsa_mailbox_msgno_has_flags(mailbox, msgno, 0,
+                                             LIBBALSA_MESSAGE_FLAG_DELETED))
+            g_array_append_val(messages, msgno);
     }
 
-    if(messages) {
+    if (messages->len) {
 	if (move_to_trash && (index != trash)) {
-	    libbalsa_messages_move(messages, balsa_app.trash);
+            GError *err = NULL;
+            if(!libbalsa_mailbox_messages_move(mailbox, messages,
+                                               balsa_app.trash, &err)) {
+                balsa_information_parented(GTK_WINDOW(balsa_app.main_window),
+                                           LIBBALSA_INFORMATION_ERROR,
+                                           _("Move to Trash failed: %s"),
+                                           err ? err->message : "?");
+                g_clear_error(&err);
+            }
 	    enable_empty_trash(balsa_app.main_window, TRASH_FULL);
 	} else {
-	    libbalsa_messages_change_flag(messages,
-                                          LIBBALSA_MESSAGE_FLAG_DELETED,
-                                          TRUE);
+            libbalsa_mailbox_messages_change_flags
+		(mailbox, messages, LIBBALSA_MESSAGE_FLAG_DELETED,
+		 (LibBalsaMessageFlag) 0);
 	    if (index == trash)
 		enable_empty_trash(balsa_app.main_window, TRASH_CHECK);
 	}
-	g_list_foreach(messages, (GFunc)g_object_unref, NULL);
-	g_list_free(messages);
     }
+    g_array_free(messages, TRUE);
 }
 
 /*
@@ -1565,15 +1629,21 @@ bndx_do_delete(BalsaIndex* index, gboolean move_to_trash)
 void
 balsa_message_move_to_trash(GtkWidget * widget, gpointer user_data)
 {
+    BalsaIndex *index;
+
     g_return_if_fail(user_data != NULL);
-    bndx_do_delete(BALSA_INDEX(user_data), TRUE);
+    index = BALSA_INDEX(user_data);
+    bndx_do_delete(index, TRUE);
+    /* Note when message was flagged as deleted, for use in
+     * auto-expunge. */
+    time(&index->mailbox_node->last_use);
 }
 
 gint
 balsa_find_notebook_page_num(LibBalsaMailbox * mailbox)
 {
     GtkWidget *page;
-    guint i;
+    gint i;
 
     if (!balsa_app.notebook)
 	return -1;
@@ -1598,41 +1668,50 @@ balsa_find_notebook_page_num(LibBalsaMailbox * mailbox)
 void
 balsa_index_toggle_flag(BalsaIndex* index, LibBalsaMessageFlag flag)
 {
-    GList *list, *l;
+    LibBalsaMailbox *mailbox = index->mailbox_node->mailbox;
     int is_all_flagged = TRUE;
+    gint i;
 
     /* First see if we should set given flag or unset */
-    l = balsa_index_selected_list(index);
-    for (list = l; list; list = g_list_next(list)) {
-	if (!LIBBALSA_MESSAGE_HAS_FLAG(list->data, flag)) {
+    for (i = index->selected->len; --i >= 0;) {
+        guint msgno = g_array_index(index->selected, guint, i);
+        if (libbalsa_mailbox_msgno_has_flags(mailbox, msgno, 0, flag)) {
 	    is_all_flagged = FALSE;
 	    break;
 	}
     }
 
-    libbalsa_messages_change_flag(l, flag, !is_all_flagged);
+    libbalsa_mailbox_messages_change_flags(mailbox, index->selected,
+                                           is_all_flagged ? 0 : flag,
+                                           is_all_flagged ? flag : 0);
 
-    g_list_foreach(l, (GFunc)g_object_unref, NULL);
-    g_list_free(l);
+    if (flag == LIBBALSA_MESSAGE_FLAG_DELETED)
+	/* Note when deleted flag was changed, for use in
+	 * auto-expunge. */
+	time(&index->mailbox_node->last_use);
 }
 
 static void
 bi_toggle_deleted_cb(GtkWidget * widget, gpointer user_data)
 {
     BalsaIndex *index;
+
     g_return_if_fail(user_data != NULL);
 
     index = BALSA_INDEX(user_data);
     balsa_index_toggle_flag(index, LIBBALSA_MESSAGE_FLAG_DELETED);
-    if (widget == index->undelete_item) {
-	GList *l = balsa_index_selected_list(index);
-	if (l && LIBBALSA_MESSAGE_IS_DELETED(l->data))
+
+    if (widget == index->undelete_item && index->selected->len > 0) {
+	LibBalsaMailbox *mailbox = index->mailbox_node->mailbox;
+        guint msgno = g_array_index(index->selected, guint, 0);
+        if (libbalsa_mailbox_msgno_has_flags(mailbox, msgno,
+                                             LIBBALSA_MESSAGE_FLAG_DELETED,
+					     0))
 	    /* Oops! */
 	    balsa_index_toggle_flag(index, LIBBALSA_MESSAGE_FLAG_DELETED);
-	g_list_foreach(l, (GFunc) g_object_unref, NULL);
-	g_list_free(l);
     }
 }
+
 /* This function toggles the FLAGGED attribute of a list of messages
  */
 static void
@@ -1661,12 +1740,8 @@ mru_menu_cb(gchar * url, BalsaIndex * index)
 
     g_return_if_fail(mailbox != NULL);
 
-    if (index->mailbox_node->mailbox != mailbox) {
-        GList *messages = balsa_index_selected_list(index);
-        balsa_index_transfer(index, messages, mailbox, FALSE);
-	g_list_foreach(messages, (GFunc)g_object_unref, NULL);
-        g_list_free(messages);
-    }
+    if (index->mailbox_node->mailbox != mailbox)
+        balsa_index_transfer(index, index->selected, mailbox, FALSE);
 }
 
 /*
@@ -1765,24 +1840,24 @@ bndx_do_popup(BalsaIndex * index, GdkEventButton * event)
     gboolean any_not_deleted = FALSE;
     gint event_button;
     guint event_time;
- 
+    gint i;
+
     BALSA_DEBUG();
 
-    l = balsa_index_selected_list(index);
-    for (list = l; list; list = g_list_next(list)) {
-        LibBalsaMessage *message = list->data;
-
-        if (LIBBALSA_MESSAGE_IS_DELETED(message))
+    mailbox = index->mailbox_node->mailbox;
+    for (i = index->selected->len; --i >= 0;) {
+        guint msgno = g_array_index(index->selected, guint, i);
+        if (libbalsa_mailbox_msgno_has_flags(mailbox, msgno,
+                                             LIBBALSA_MESSAGE_FLAG_DELETED,
+					     0))
             any_deleted = TRUE;
         else
             any_not_deleted = TRUE;
     }
-    g_list_foreach(l, (GFunc)g_object_unref, NULL);
-    g_list_free(l);
-    any = (l != NULL);
+    any = index->selected->len > 0;
 
     l = gtk_container_get_children(GTK_CONTAINER(menu));
-    for (list = l; list; list = g_list_next(list))
+    for (list = l; list; list = list->next)
         gtk_widget_set_sensitive(GTK_WIDGET(list->data), any);
     g_list_free(l);
 
@@ -1901,13 +1976,15 @@ balsa_index_set_sos_filter(BalsaIndex *bindex, const gchar *sos_filter,
     bindex->sos_filter = g_strdup(sos_filter);
 
     if(sos_filter && sos_filter[0] != '\0') {
+	/* Check the To: field for sentbox and From for others */
         LibBalsaCondition *name = 
             libbalsa_condition_new_bool_ptr
             (FALSE, CONDITION_OR,
              libbalsa_condition_new_string
              (FALSE, CONDITION_MATCH_SUBJECT, g_strdup(sos_filter), NULL),
              libbalsa_condition_new_string
-             (FALSE, CONDITION_MATCH_FROM, g_strdup(sos_filter), NULL));
+             (FALSE, (libbalsa_mailbox_get_show(mailbox) == LB_MAILBOX_SHOW_TO) ? CONDITION_MATCH_TO : CONDITION_MATCH_FROM,
+	      g_strdup(sos_filter), NULL));
         
         if(flag_filter)
             flag_filter = libbalsa_condition_new_bool_ptr
@@ -1940,7 +2017,7 @@ balsa_index_remove_duplicates(BalsaIndex * index)
 
     mailbox = BALSA_INDEX(index)->mailbox_node->mailbox;
     table = g_hash_table_new(g_str_hash, g_str_equal);
-    for (list = mailbox->message_list; list; list = g_list_next(list)) {
+    for (list = mailbox->message_list; list; list = list->next) {
         LibBalsaMessage *message = list->data;
         LibBalsaMessage *master;
 
@@ -1957,8 +2034,7 @@ balsa_index_remove_duplicates(BalsaIndex * index)
             GtkTreePath *path;
 
             messages = g_list_prepend(messages, message);
-            if (!balsa_app.hide_deleted
-                && bndx_find_message(index, &path, NULL, message)) {
+            if (bndx_find_message(index, &path, NULL, message)) {
                 bndx_expand_to_row(index, path);
                 gtk_tree_path_free(path);
             }
@@ -1967,7 +2043,7 @@ balsa_index_remove_duplicates(BalsaIndex * index)
 
     if (messages) {
 	if (mailbox != balsa_app.trash) {
-	    libbalsa_messages_move(messages, balsa_app.trash);
+	    libbalsa_messages_move(messages, balsa_app.trash, NULL);
 	    enable_empty_trash(TRASH_FULL);
 	} else {
 	    libbalsa_messages_change_flag(messages, 
@@ -1997,27 +2073,30 @@ balsa_index_refresh_date(BalsaIndex * index)
 
 /* Transfer messages. */
 void
-balsa_index_transfer(BalsaIndex *index, GList * messages,
+balsa_index_transfer(BalsaIndex *index, GArray * msgnos,
                      LibBalsaMailbox * to_mailbox, gboolean copy)
 {
     gboolean success;
     LibBalsaMailbox *from_mailbox;
+    GError *e = NULL;
 
-    if (messages == NULL)
+    if (msgnos->len == 0)
         return;
 
-    success = copy ? libbalsa_messages_copy(messages, to_mailbox)
-		   : libbalsa_messages_move(messages, to_mailbox);
+    from_mailbox = index->mailbox_node->mailbox;
+    success = copy ?
+        libbalsa_mailbox_messages_copy(from_mailbox, msgnos, to_mailbox, &e) :
+        libbalsa_mailbox_messages_move(from_mailbox, msgnos, to_mailbox, &e);
+
     if (!success) {
 	balsa_information(LIBBALSA_INFORMATION_WARNING,
-			  messages->next 
-			  ? _("Failed to copy messages to mailbox \"%s\".")
-			  : _("Failed to copy message to mailbox \"%s\"."),
-			  to_mailbox->name);
+			  msgnos->len > 1 
+			  ? _("Failed to copy messages to mailbox \"%s\": %s")
+			  : _("Failed to copy message to mailbox \"%s\": %s"),
+			  to_mailbox->name, e ? e->message : "?");
 	return;
     }
 
-    from_mailbox = index->mailbox_node->mailbox;
     balsa_mblist_set_status_bar(from_mailbox);
 
     if (from_mailbox == balsa_app.trash && !copy)
@@ -2027,6 +2106,10 @@ balsa_index_transfer(BalsaIndex *index, GList * messages,
     balsa_information(LIBBALSA_INFORMATION_MESSAGE,
                       copy ? _("Copied to \"%s\".") 
                       : _("Moved to \"%s\"."), to_mailbox->name);
+    if (!copy)
+	/* Note when message was flagged as deleted, for use in
+	 * auto-expunge. */
+	time(&index->mailbox_node->last_use);
 }
 
 /* General helpers. */
@@ -2147,3 +2230,84 @@ balsa_index_expunge(BalsaIndex * index)
     g_signal_handler_unblock(selection, index->selection_changed_id);
     g_signal_emit_by_name(G_OBJECT(selection), "changed");
 }
+
+/* Message window */
+static guint
+bndx_next_msgno(BalsaIndex * index, guint current_msgno,
+	        LibBalsaMailboxSearchIter * search_iter,
+                BndxSearchDirection direction, BndxSearchViewable viewable,
+                BndxSearchWrap wrap)
+{
+    LibBalsaMailbox *mailbox = index->mailbox_node->mailbox;
+    GtkTreeModel *model = GTK_TREE_MODEL(mailbox);
+    GtkTreeIter iter;
+    guint msgno = 0;
+    guint stop_msgno;
+
+    if (!libbalsa_mailbox_msgno_find(mailbox, current_msgno, NULL, &iter))
+        return msgno;
+
+    stop_msgno = 0;
+    if (wrap == BNDX_SEARCH_WRAP_YES)
+        stop_msgno = current_msgno;
+    if (bndx_search_iter(index, search_iter, &iter, direction, viewable,
+                         stop_msgno))
+        gtk_tree_model_get(model, &iter, LB_MBOX_MSGNO_COL, &msgno, -1);
+
+    return msgno;
+}
+
+guint
+balsa_index_next_msgno(BalsaIndex * index, guint current_msgno)
+{
+    return bndx_next_msgno(index, current_msgno, index->search_iter,
+                           BNDX_SEARCH_DIRECTION_NEXT,
+                           BNDX_SEARCH_VIEWABLE_ONLY, BNDX_SEARCH_WRAP_NO);
+}
+
+guint
+balsa_index_previous_msgno(BalsaIndex * index, guint current_msgno)
+{
+    return bndx_next_msgno(index, current_msgno, index->search_iter,
+                           BNDX_SEARCH_DIRECTION_PREV,
+                           BNDX_SEARCH_VIEWABLE_ONLY, BNDX_SEARCH_WRAP_NO);
+}
+
+static guint
+bndx_next_msgno_with_flag(BalsaIndex * index, guint current_msgno,
+                          LibBalsaMessageFlag flag)
+{
+    LibBalsaCondition *condition;
+    LibBalsaMailboxSearchIter *search_iter;
+    guint msgno;
+
+    condition = libbalsa_condition_new_flag_enum(FALSE, flag);
+    condition =
+        libbalsa_condition_new_bool_ptr(FALSE, CONDITION_AND, condition,
+                                        libbalsa_condition_clone
+                                        (&cond_undeleted));
+    search_iter = libbalsa_mailbox_search_iter_new(condition);
+    libbalsa_condition_free(condition);
+
+    msgno =
+        bndx_next_msgno(index, current_msgno, search_iter,
+		        BNDX_SEARCH_DIRECTION_NEXT,
+                        BNDX_SEARCH_VIEWABLE_ANY, BNDX_SEARCH_WRAP_YES);
+    libbalsa_mailbox_search_iter_free(search_iter);
+
+    return msgno;
+}
+
+guint
+balsa_index_next_unread_msgno(BalsaIndex * index, guint current_msgno)
+{
+    return bndx_next_msgno_with_flag(index, current_msgno,
+                                     LIBBALSA_MESSAGE_FLAG_NEW);
+}
+
+guint
+balsa_index_next_flagged_msgno(BalsaIndex * index, guint current_msgno)
+{
+    return bndx_next_msgno_with_flag(index, current_msgno,
+                                     LIBBALSA_MESSAGE_FLAG_FLAGGED);
+}
