@@ -59,9 +59,10 @@ extern "C" {
         BalsaMailboxNode* mailbox_node;
         LibBalsaMessage* first_new_message;
         LibBalsaMessage* current_message;
+	gboolean current_message_is_deleted;
         gboolean prev_message;
         gboolean next_message;
-	GSList *selected;
+	GArray *selected;
         gchar *sos_filter; /* SenderOrSubject filter string, if any */
 
         /* signal handler ids */
@@ -110,7 +111,7 @@ extern "C" {
                                     LibBalsaCondition *flag_filter);
 
 /* move or copy a list of messages */
-    void balsa_index_transfer(BalsaIndex * index, GList * messages,
+    void balsa_index_transfer(BalsaIndex * index, GArray * msgnos,
                               LibBalsaMailbox * to_mailbox, gboolean copy);
 
 /* select up/down the index */
@@ -157,6 +158,15 @@ extern "C" {
     /* Expunge deleted messages. */
     void balsa_index_expunge(BalsaIndex * index);
 
+    /* Message window */
+    guint balsa_index_next_msgno(BalsaIndex * index, guint current_msgno);
+    guint balsa_index_previous_msgno(BalsaIndex * index,
+                                     guint current_msgno);
+    guint balsa_index_next_unread_msgno(BalsaIndex * index,
+                                        guint current_msgno);
+    guint balsa_index_next_flagged_msgno(BalsaIndex * index,
+                                         guint current_msgno);
+
 #ifdef __cplusplus
 }
 #endif				/* __cplusplus */
