@@ -111,7 +111,7 @@ static void bmbl_row_activated_cb(GtkTreeView * tree_view,
                                   GtkTreeViewColumn * column,
                                   gpointer data);
 static void bmbl_mailbox_changed_cb(LibBalsaMailbox * mailbox,
-				    GtkTreeStore * store);
+				    gpointer data);
 /* helpers */
 static gboolean bmbl_find_all_unread_mboxes_func(GtkTreeModel * model,
                                                  GtkTreePath * path,
@@ -400,7 +400,7 @@ balsa_mblist_get_store(void)
 {
     if (balsa_app.mblist_tree_store)
 	g_object_ref(balsa_app.mblist_tree_store);
-    else
+    else {
         balsa_app.mblist_tree_store =
             gtk_tree_store_new(N_COLUMNS,
                                G_TYPE_OBJECT,     /* MBNODE_COLUMN */
@@ -411,6 +411,10 @@ balsa_mblist_get_store(void)
                                G_TYPE_STRING,     /* UNREAD_COLUMN */
                                G_TYPE_STRING      /* TOTAL_COLUMN  */
             );
+        g_object_add_weak_pointer(G_OBJECT(balsa_app.mblist_tree_store),
+                                  (gpointer) & balsa_app.
+                                  mblist_tree_store);
+    }
 
     return balsa_app.mblist_tree_store;
 }
@@ -724,8 +728,8 @@ bmbl_column_resize(GtkWidget * widget,
 /* bmbl_drag_cb
  * 
  * Description: This is the drag_data_recieved signal handler for the
- * BalsaMBList.  It retrieves a list of messages terminated by a NULL
- * pointer, converts it to a GList, then transfers it to the selected
+ * BalsaMBList.  It retrieves the source BalsaIndex and transfers the
+ * index's selected messages to the target
  * mailbox.  Depending on what key is held down when the message(s)
  * are dropped they are either copied or moved.  The default action is
  * to copy.
@@ -742,22 +746,19 @@ bmbl_drag_cb(GtkWidget * widget, GdkDragContext * context,
     GtkTreeIter iter;
     LibBalsaMailbox *mailbox;
     LibBalsaMailbox *orig_mailbox;
-    GList *messages = NULL;
-    LibBalsaMessage **message;
+    BalsaIndex *orig_index;
 
+    if (!selection_data)
+	/* Drag'n'drop is weird... */
+	return;
 
-
-    /* convert pointer array to GList */
-    for (message = (LibBalsaMessage **) selection_data->data; *message;
-         message++) {
-        messages = g_list_prepend(messages, *message);
-    }
-    if(!messages) /* it is actually possible to drag from GtkTreeView
-                   * * when no rows are selected: Disable preview for
-                   * that. */
+    orig_index = *(BalsaIndex **) selection_data->data;
+    if (orig_index->selected->len == 0)
+	/* it is actually possible to drag from GtkTreeView when no rows
+	 * are selected: Disable preview for that. */
        return; 
 
-    orig_mailbox = ((LibBalsaMessage *) messages->data)->mailbox;
+    orig_mailbox = orig_index->mailbox_node->mailbox;
 
     /* find the node and mailbox */
 
@@ -776,16 +777,13 @@ bmbl_drag_cb(GtkWidget * widget, GdkDragContext * context,
 
         /* cannot transfer to the originating mailbox */
         if (mailbox != NULL && mailbox != orig_mailbox)
-            balsa_index_transfer(balsa_find_index_by_mailbox(orig_mailbox),
-                                 messages, mailbox,
+            balsa_index_transfer(orig_index, orig_index->selected, mailbox,
                                  context->action != GDK_ACTION_MOVE);
         gtk_tree_path_free(path);
     }
 
     if (balsa_find_iter_by_data(&iter, orig_mailbox))
         gtk_tree_selection_select_iter(selection, &iter);
-
-    g_list_free(messages);
 }
 
 /* bmbl_select_mailbox
@@ -893,7 +891,6 @@ bmbl_row_activated_cb(GtkTreeView * tree_view, GtkTreePath * path,
 
 struct update_mbox_data {
     LibBalsaMailbox *mailbox;
-    GtkTreeStore *store;
     gint total_messages; /* to be compatible with update_mailbox() arg. */
     gboolean notify;
 };
@@ -901,19 +898,18 @@ static void bmbl_update_mailbox(GtkTreeStore * store,
 				LibBalsaMailbox * mailbox,
 				gint total_messages);
 static gboolean
-update_mailbox_idle(struct update_mbox_data*umd)
+update_mailbox_idle(struct update_mbox_data *umd)
 {
     gdk_threads_enter();
-    if (umd->mailbox)
+    if (umd->mailbox) {
         g_object_remove_weak_pointer(G_OBJECT(umd->mailbox),
-                                     (gpointer) &umd->mailbox);
-    if (umd->store)
-        g_object_remove_weak_pointer(G_OBJECT(umd->store),
-                                     (gpointer) &umd->store);
-    if (umd->mailbox && umd->store) {
-        bmbl_update_mailbox(umd->store, umd->mailbox, umd->total_messages);
-	check_new_messages_count(umd->mailbox, umd->notify);
-	g_object_set_data(G_OBJECT(umd->mailbox), "mblist-update", NULL);
+                                     (gpointer) & umd->mailbox);
+        if (balsa_app.mblist_tree_store) {
+            bmbl_update_mailbox(balsa_app.mblist_tree_store, umd->mailbox,
+                                umd->total_messages);
+            check_new_messages_count(umd->mailbox, umd->notify);
+        }
+        g_object_set_data(G_OBJECT(umd->mailbox), "mblist-update", NULL);
     }
     gdk_threads_leave();
     g_free(umd);
@@ -921,11 +917,10 @@ update_mailbox_idle(struct update_mbox_data*umd)
 }
 
 static void
-bmbl_mailbox_changed_cb(LibBalsaMailbox * mailbox, GtkTreeStore * store)
+bmbl_mailbox_changed_cb(LibBalsaMailbox * mailbox, gpointer data)
 {
     struct update_mbox_data *umd;
     g_return_if_fail(mailbox);
-    g_return_if_fail(store);
     if( (umd = g_object_get_data(G_OBJECT(mailbox), "mblist-update")) ) {
         if(!MAILBOX_CLOSED(mailbox))
             umd->total_messages = libbalsa_mailbox_total_messages(mailbox);
@@ -935,8 +930,6 @@ bmbl_mailbox_changed_cb(LibBalsaMailbox * mailbox, GtkTreeStore * store)
     g_object_set_data(G_OBJECT(mailbox), "mblist-update", umd);
     umd->mailbox = mailbox;
     g_object_add_weak_pointer(G_OBJECT(mailbox), (gpointer) &umd->mailbox);
-    umd->store = store;
-    g_object_add_weak_pointer(G_OBJECT(store), (gpointer) &umd->store);
     umd->total_messages = 
         !MAILBOX_CLOSED(mailbox) 
         ? (gint)libbalsa_mailbox_total_messages(mailbox) : -1;
@@ -1028,6 +1021,8 @@ balsa_mblist_open_mailbox(LibBalsaMailbox * mailbox)
     } else { /* page with mailbox not found, open it */
 	balsa_window_open_mbnode(balsa_app.main_window, mbnode);
 
+	balsa_window_set_filter_label(balsa_app.main_window,
+				  libbalsa_mailbox_get_show(mailbox) == LB_MAILBOX_SHOW_TO);
 	if (balsa_app.mblist->display_info)
 	    balsa_mblist_update_mailbox(balsa_app.mblist_tree_store,
                                         mailbox);
@@ -1145,7 +1140,7 @@ bmbl_real_disconnect_mbnode_signals(BalsaMailboxNode * mbnode,
         g_signal_handlers_disconnect_by_func(mbnode->mailbox,
                                              G_CALLBACK
                                              (bmbl_mailbox_changed_cb),
-                                             model);
+                                             NULL);
 }
 
 /* bmbl_store_redraw_mbnode
@@ -1207,7 +1202,7 @@ bmbl_store_redraw_mbnode(GtkTreeIter * iter, BalsaMailboxNode * mbnode)
 	    /* Now we have a mailbox: */
 	    g_signal_connect(mbnode->mailbox, "changed",
 			     G_CALLBACK(bmbl_mailbox_changed_cb),
-			     balsa_app.mblist_tree_store);
+			     NULL);
 	    /* If necessary, expand rows to expose this mailbox after
 	     * setting its mbnode in the tree-store. */
 	    expose = libbalsa_mailbox_get_exposed(mbnode->mailbox);
@@ -1722,9 +1717,10 @@ bmbl_mru_show_tree(GtkWidget * widget, gpointer data)
     if (req.height > balsa_app.mw_height)
         req.height = balsa_app.mw_height;
     /* For the mailbox list width, we use the one used on the main
-     * window. This is the user choice and required because the mblist
-     * widget saves the size in balsa_app.mblist_width */
-    req.width = balsa_app.mblist_width;
+     * window enhanced somewhat. This is the user choice and required
+     * because the mblist widget saves the size in
+     * balsa_app.mblist_width */
+    req.width = balsa_app.mblist_width> 200 ? balsa_app.mblist_width : 200;
     gtk_widget_set_size_request(GTK_WIDGET(mblist), req.width, req.height);
 
     gtk_container_add(GTK_CONTAINER(scroll), mblist);
@@ -1752,8 +1748,6 @@ bmbl_mru_selected_cb(GtkTreeSelection * selection, gpointer data)
     GtkTreeView *tree_view;
     GtkTreePath *path;
     GtkTreeViewColumn *column;
-    GdkRectangle rect;
-    GValue expander_size;
 
     if (!data)
         return;
@@ -1772,13 +1766,7 @@ bmbl_mru_selected_cb(GtkTreeSelection * selection, gpointer data)
         return;
     }
 
-    gtk_tree_view_get_cell_area(tree_view, path, column, &rect);
-    g_value_init(&expander_size, G_TYPE_INT);
-    gtk_widget_style_get_property((GtkWidget *) tree_view, "expander-size",
-                                  &expander_size);
-    if (event->button.x < rect.x + g_value_get_int(&expander_size))
-        gtk_tree_selection_unselect_all(selection);
-    else if (gtk_tree_selection_path_is_selected(selection, path)) {
+    if (gtk_tree_selection_path_is_selected(selection, path)) {
         GtkTreeModel *model;
         GtkTreeIter iter;
         BalsaMailboxNode *mbnode;
@@ -1863,6 +1851,172 @@ balsa_mblist_mru_drop(GList ** list, const gchar * url)
     }
 }
 
+#if GTK_CHECK_VERSION(2, 4, 0)
+/* balsa_mblist_mru_option_menu: create a GtkComboBox to manage an MRU
+ * list */
+
+/* The info that needs to be passed around */
+struct _BalsaMBListMRUOption {
+    GtkWindow *window;
+    GList **url_list;
+    GSList *real_urls;
+};
+typedef struct _BalsaMBListMRUOption BalsaMBListMRUOption;
+
+/* Helper. */
+static void bmbl_mru_combo_box_changed(GtkComboBox * combo_box,
+                                       BalsaMBListMRUOption * mro);
+static void
+bmbl_mru_combo_box_setup(GtkComboBox * combo_box)
+{
+    BalsaMBListMRUOption *mro =
+        g_object_get_data(G_OBJECT(combo_box), "mro");
+    GList *list;
+    gint i;
+
+    gtk_combo_box_set_active(combo_box, -1);
+    for (i = g_slist_length(mro->real_urls) + 1; --i >= 0;)
+        gtk_combo_box_remove_text(combo_box, i);
+    g_slist_foreach(mro->real_urls, (GFunc) g_free, NULL);
+    g_slist_free(mro->real_urls);
+    mro->real_urls = NULL;
+
+    for (list = *mro->url_list; list; list = list->next) {
+        const gchar *url = list->data;
+        LibBalsaMailbox *mailbox;
+
+        if ((mailbox = balsa_find_mailbox_by_url(url)) || !*url) {
+            gtk_combo_box_append_text(GTK_COMBO_BOX(combo_box),
+                                      mailbox ? mailbox->name : "");
+            mro->real_urls = g_slist_append(mro->real_urls, g_strdup(url));
+        }
+    }
+
+    gtk_combo_box_append_text(GTK_COMBO_BOX(combo_box), _("Other..."));
+    gtk_combo_box_set_active(combo_box, 0);
+}
+
+/* Callbacks */
+static void
+bmbl_mru_combo_box_changed(GtkComboBox * combo_box,
+                           BalsaMBListMRUOption * mro)
+{
+    BalsaMBListMRUEntry mru;
+    const gchar *url;
+    gint active = gtk_combo_box_get_active(combo_box);
+
+    if (active < 0)
+	return;
+    if ((url = g_slist_nth_data(mro->real_urls, active))) {
+	/* Move this url to the top. */
+	balsa_mblist_mru_add(mro->url_list, url);
+        return;
+    }
+
+    /* User clicked on "Other..." */
+    mru.window = mro->window;
+    mru.url_list = mro->url_list;
+    mru.user_func = NULL;
+    mru.setup_cb = NULL;
+    mru.user_data = combo_box;
+    mru.url = NULL;
+    bmbl_mru_show_tree(NULL, &mru);
+    g_free(mru.url);
+    bmbl_mru_combo_box_setup(combo_box);
+}
+
+static void
+bmbl_mru_combo_box_destroy_cb(BalsaMBListMRUOption * mro)
+{
+    g_slist_foreach(mro->real_urls, (GFunc) g_free, NULL);
+    g_slist_free(mro->real_urls);
+    g_free(mro);
+}
+
+/*
+ * balsa_mblist_mru_option_menu:
+ *
+ * window:      parent window for the dialog;
+ * url_list:    pointer to a list of urls;
+ *
+ * Returns a GtkComboBox.
+ *
+ * Takes a list of urls and creates a combo-box with an entry for
+ * each one that resolves to a mailbox, labeled with the mailbox name,
+ * including the special case of an empty url and a NULL mailbox.
+ * Adds a last entry that pops up the whole mailbox tree. When an item
+ * is clicked, the url_list is updated.
+ */
+
+GtkWidget *
+balsa_mblist_mru_option_menu(GtkWindow * window, GList ** url_list)
+{
+    GtkWidget *combo_box;
+    BalsaMBListMRUOption *mro;
+
+    g_return_val_if_fail(url_list != NULL, NULL);
+
+    combo_box = gtk_combo_box_new_text();
+    mro = g_new(BalsaMBListMRUOption, 1);
+
+    mro->window = window;
+    mro->url_list = url_list;
+    mro->real_urls = NULL;
+
+    g_object_set_data_full(G_OBJECT(combo_box), "mro", mro,
+                           (GDestroyNotify) bmbl_mru_combo_box_destroy_cb);
+    bmbl_mru_combo_box_setup(GTK_COMBO_BOX(combo_box));
+    g_signal_connect(G_OBJECT(combo_box), "changed",
+                     G_CALLBACK(bmbl_mru_combo_box_changed), mro);
+
+    return combo_box;
+}
+
+/*
+ * balsa_mblist_mru_option_menu_set
+ *
+ * combo_box: a GtkComboBox created by balsa_mblist_mru_option_menu;
+ * url:       URL of a mailbox
+ *
+ * Adds url to the front of the url_list managed by combo_box, resets
+ * combo_box to show the new url, and stores a copy in the mro
+ * structure.
+ */
+void
+balsa_mblist_mru_option_menu_set(GtkWidget * combo_box, const gchar * url)
+{
+    BalsaMBListMRUOption *mro =
+        g_object_get_data(G_OBJECT(combo_box), "mro");
+
+    balsa_mblist_mru_add(mro->url_list, url);
+    bmbl_mru_combo_box_setup(GTK_COMBO_BOX(combo_box));
+}
+
+/*
+ * balsa_mblist_mru_option_menu_get
+ *
+ * combo_box: a GtkComboBox created by balsa_mblist_mru_option_menu.
+ *
+ * Returns the address of the current URL.
+ *
+ * Note that the url is held in the mro structure, and is freed when the
+ * widget is destroyed. The calling code must make its own copy of the
+ * string if it is needed more than temporarily.
+ */
+const gchar *
+balsa_mblist_mru_option_menu_get(GtkWidget * combo_box)
+{
+    gint active;
+    BalsaMBListMRUOption *mro =
+        g_object_get_data(G_OBJECT(combo_box), "mro");
+
+    active = gtk_combo_box_get_active(GTK_COMBO_BOX(combo_box));
+
+    return g_slist_nth_data(mro->real_urls, active);
+}
+
+#else /* GTK_CHECK_VERSION(2, 4, 0) */
+
 /* balsa_mblist_mru_option_menu: create a GtkOptionMenu to manage an MRU
  * list */
 
@@ -2022,6 +2176,7 @@ bmbl_mru_option_menu_init(BalsaMBListMRUOption * mro)
         gtk_menu_item_activate(item);
     }
 }
+#endif /* GTK_CHECK_VERSION(2, 4, 0) */
 
 void
 balsa_mblist_set_status_bar(LibBalsaMailbox * mailbox)
