@@ -64,7 +64,7 @@ balsa_information_real(GtkWindow *parent, LibBalsaInformationType type,
                        const char *msg)
 {
     BalsaInformationShow show;
-    gchar * show_msg = strdup(msg);
+    gchar * show_msg = g_strdup(msg);
     
     libbalsa_utf8_sanitize(&show_msg, balsa_app.convert_unknown_8bit, NULL);
     switch (type) {
