@@ -60,6 +60,9 @@ struct _BalsaMailboxConfView {
     GtkWidget *identity_label;
     GtkWidget *show_from;
     GtkWidget *show_to;
+#ifdef HAVE_GPGME
+    GtkWidget *chk_crypt;
+#endif
 };
 typedef struct _MailboxConfWindow MailboxConfWindow;
 struct _MailboxConfWindow {
@@ -85,7 +88,7 @@ struct _MailboxConfWindow {
             GtkWidget *remember;
 	    GtkWidget *password;
 	    GtkWidget *folderpath;
-            GtkWidget *enable_persistent;
+            GtkWidget *enable_persistent, *has_bugs;
             BalsaServerConf bsc;
 	} imap;
 
@@ -130,6 +133,40 @@ struct menu_data {
     const char *label;
     int tag;
 };
+#if GTK_CHECK_VERSION(2, 4, 0)
+struct mailbox_conf_combo_box_info {
+    GSList *tags;
+};
+#define BALSA_MC_COMBO_BOX_INFO "balsa-mailbox-conf-combo-box-info"
+
+static void
+mailbox_conf_combo_box_info_free(struct mailbox_conf_combo_box_info *info)
+{
+    g_slist_free(info->tags);
+    g_free(info);
+}
+
+static void
+mailbox_conf_combo_box_make(GtkComboBox * combo_box, unsigned cnt,
+                            const struct menu_data *data)
+{
+    struct mailbox_conf_combo_box_info *info =
+        g_new(struct mailbox_conf_combo_box_info, 1);
+    gint i;
+
+    info->tags = NULL;
+    for (i = cnt; --i >= 0;) {
+        gtk_combo_box_prepend_text(combo_box, _(data[i].label));
+        info->tags =
+            g_slist_prepend(info->tags, GINT_TO_POINTER(data[i].tag));
+    }
+    g_object_set_data_full(G_OBJECT(combo_box), BALSA_MC_COMBO_BOX_INFO,
+                           info,
+                           (GDestroyNotify)
+                           mailbox_conf_combo_box_info_free);
+}
+
+#else /* GTK_CHECK_VERSION(2, 4, 0) */
 static GtkWidget*
 get_combo_menu(unsigned cnt, const struct menu_data *data)
 {
@@ -145,6 +182,7 @@ get_combo_menu(unsigned cnt, const struct menu_data *data)
     gtk_widget_show_all(menu);
     return menu;
 }
+#endif /* GTK_CHECK_VERSION(2, 4, 0) */
 
 static void
 bsc_ssl_toggled_cb(GtkWidget * widget, BalsaServerConf * bsc)
@@ -191,12 +229,20 @@ balsa_server_conf_get_advanced_widget(BalsaServerConf *bsc, LibBalsaServer *s,
     gtk_table_attach(bsc->table, label, 0, 1, 1, 2,
 		     GTK_FILL, GTK_FILL, 0, 0);
 
+#if GTK_CHECK_VERSION(2, 4, 0)
+    bsc->tls_option = gtk_combo_box_new_text();
+    mailbox_conf_combo_box_make(GTK_COMBO_BOX(bsc->tls_option),
+                                ELEMENTS(tls_menu), tls_menu);
+    gtk_combo_box_set_active(GTK_COMBO_BOX(bsc->tls_option),
+                             s ? s->tls_mode : LIBBALSA_TLS_ENABLED);
+#else /* GTK_CHECK_VERSION(2, 4, 0) */
     bsc->tls_option = gtk_option_menu_new ();
     bsc->tls_mode = get_combo_menu(ELEMENTS(tls_menu), tls_menu);
     gtk_option_menu_set_menu (GTK_OPTION_MENU(bsc->tls_option),
 			      bsc->tls_mode);
     gtk_option_menu_set_history (GTK_OPTION_MENU (bsc->tls_option),
 				 s ? s->tls_mode : LIBBALSA_TLS_ENABLED);
+#endif /* GTK_CHECK_VERSION(2, 4, 0) */
     gtk_table_attach(bsc->table, bsc->tls_option, 1, 2, 1, 2,
 		     GTK_FILL, GTK_FILL, 5, 5);
     gtk_label_set_mnemonic_widget(GTK_LABEL(label), bsc->tls_option);
@@ -258,8 +304,13 @@ balsa_server_conf_set_values(BalsaServerConf *bsc, LibBalsaServer *server)
 
     gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(bsc->use_ssl), 
                                  server->use_ssl);
+#if GTK_CHECK_VERSION(2, 4, 0)
+    gtk_combo_box_set_active(GTK_COMBO_BOX(bsc->tls_option),
+                             server->tls_mode);
+#else /* GTK_CHECK_VERSION(2, 4, 0) */
     gtk_option_menu_set_history (GTK_OPTION_MENU (bsc->tls_option),
 				 server->tls_mode);
+#endif /* GTK_CHECK_VERSION(2, 4, 0) */
     gtk_widget_set_sensitive(bsc->tls_option, !server->use_ssl);
 }
 
@@ -273,9 +324,19 @@ balsa_server_conf_get_use_ssl(BalsaServerConf *bsc)
 LibBalsaTlsMode
 balsa_server_conf_get_tls_mode(BalsaServerConf *bsc)
 {
+#if GTK_CHECK_VERSION(2, 4, 0)
+    struct mailbox_conf_combo_box_info *info =
+        g_object_get_data(G_OBJECT(bsc->tls_option),
+                          BALSA_MC_COMBO_BOX_INFO);
+    gint active = gtk_combo_box_get_active(GTK_COMBO_BOX(bsc->tls_option));
+
+    return (LibBalsaTlsMode)
+        GPOINTER_TO_INT(g_slist_nth_data(info->tags, active));
+#else /* GTK_CHECK_VERSION(2, 4, 0) */
     GtkWidget *menu_item = gtk_menu_get_active(GTK_MENU(bsc->tls_mode));
     return GPOINTER_TO_INT(g_object_get_data(G_OBJECT(menu_item),
                                              "balsa-data"));
+#endif /* GTK_CHECK_VERSION(2, 4, 0) */
 }
 
 /* END BalsaServerConf ===================================== */
@@ -660,6 +721,11 @@ mailbox_conf_set_values(MailboxConfWindow *mcw)
                 (GTK_TOGGLE_BUTTON(mcw->mb_data.imap.enable_persistent),
                  TRUE);
 #endif
+        if(libbalsa_imap_server_has_bug(LIBBALSA_IMAP_SERVER(server),
+                                        ISBUG_FETCH))
+            gtk_toggle_button_set_active
+                (GTK_TOGGLE_BUTTON(mcw->mb_data.imap.has_bugs),
+                 TRUE);
         if(!server->remember_passwd)
             gtk_widget_set_sensitive(GTK_WIDGET(mcw->mb_data.imap.password),
                                      FALSE);
@@ -801,6 +867,10 @@ update_imap_mailbox(MailboxConfWindow *mcw)
          gtk_toggle_button_get_active
          GTK_TOGGLE_BUTTON(mcw->mb_data.imap.enable_persistent));
 #endif
+    libbalsa_imap_server_set_bug
+        (LIBBALSA_IMAP_SERVER(server), ISBUG_FETCH,
+         gtk_toggle_button_get_active
+         GTK_TOGGLE_BUTTON(mcw->mb_data.imap.has_bugs));
     /* Set host after all other server changes, as it triggers
      * save-to-config for any folder or mailbox using this server. */
     libbalsa_server_set_host(server,
@@ -1110,7 +1180,7 @@ create_pop_mailbox_page(MailboxConfWindow *mcw)
 
     advanced =
         balsa_server_conf_get_advanced_widget(&mcw->mb_data.pop3.bsc,
-                                              NULL, 1);
+                                              NULL, 2);
     gtk_notebook_append_page(GTK_NOTEBOOK(notebook), advanced,
                              gtk_label_new_with_mnemonic(_("_Advanced")));
     /* toggle for apop */
@@ -1187,16 +1257,16 @@ create_imap_mailbox_page(MailboxConfWindow *mcw)
 
     label = create_label(_("F_older Path:"), table, ++row);
 
-#ifdef HAVE_GTK24
+#if GTK_CHECK_VERSION(2, 4, 0)
     mcw->mb_data.imap.folderpath = entry = gtk_entry_new();
-#else
+#else /* GTK_CHECK_VERSION(2, 4, 0) */
     entry = gnome_entry_new("IMAPFolderHistory");
     mcw->mb_data.imap.folderpath = gnome_entry_gtk_entry(GNOME_ENTRY(entry));
     gnome_entry_append_history(GNOME_ENTRY(entry), 1, "INBOX");
     gnome_entry_append_history(GNOME_ENTRY(entry), 1, "INBOX.Sent");
     gnome_entry_append_history(GNOME_ENTRY(entry), 1, "INBOX.Draft");
     gnome_entry_append_history(GNOME_ENTRY(entry), 1, "INBOX.outbox");
-#endif
+#endif /* GTK_CHECK_VERSION(2, 4, 0) */
     gtk_entry_set_text(GTK_ENTRY(mcw->mb_data.imap.folderpath), "INBOX");
 
     gtk_label_set_mnemonic_widget(GTK_LABEL(label), 
@@ -1221,6 +1291,9 @@ create_imap_mailbox_page(MailboxConfWindow *mcw)
         balsa_server_conf_add_checkbox(&mcw->mb_data.imap.bsc,
                                        _("Enable _persistent cache"));
 #endif
+    mcw->mb_data.imap.has_bugs = 
+        balsa_server_conf_add_checkbox(&mcw->mb_data.imap.bsc,
+                                       _("Enable _bug workarounds"));
     gtk_notebook_append_page(GTK_NOTEBOOK(notebook), advanced,
                              gtk_label_new_with_mnemonic(_("_Advanced")));
 
@@ -1305,9 +1378,69 @@ mailbox_conf_view_new(LibBalsaMailbox * mailbox,
 	      view_info->show_to : view_info->show_from);
     gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);
 
+#ifdef HAVE_GPGME
+    {
+	/* scope */
+	static const struct menu_data chk_crypt_menu[] = {
+	    { N_("Never"),       LB_MAILBOX_CHK_CRYPT_NEVER  },
+	    { N_("If Possible"), LB_MAILBOX_CHK_CRYPT_MAYBE  },
+	    { N_("Always"),      LB_MAILBOX_CHK_CRYPT_ALWAYS }
+	};
+#if !GTK_CHECK_VERSION(2, 4, 0)
+	GtkWidget * chk_crypt_w;
+#endif /* GTK_CHECK_VERSION(2, 4, 0) */
+	
+	create_label(_("decrypt and check\nsignatures automatically"), table, ++row);
+	
+	box = gtk_hbox_new(FALSE, 12);
+#if GTK_CHECK_VERSION(2, 4, 0)
+        view_info->chk_crypt = gtk_combo_box_new_text();
+        mailbox_conf_combo_box_make(GTK_COMBO_BOX(view_info->chk_crypt),
+                                    ELEMENTS(chk_crypt_menu),
+                                    chk_crypt_menu);
+        gtk_combo_box_set_active(GTK_COMBO_BOX(view_info->chk_crypt),
+                                 libbalsa_mailbox_get_crypto_mode
+                                 (mailbox));
+        gtk_box_pack_start(GTK_BOX(box), view_info->chk_crypt, FALSE,
+                           FALSE, 0);
+#else /* GTK_CHECK_VERSION(2, 4, 0) */
+	chk_crypt_w = gtk_option_menu_new ();
+	view_info->chk_crypt = get_combo_menu(ELEMENTS(chk_crypt_menu), chk_crypt_menu);
+	gtk_option_menu_set_menu (GTK_OPTION_MENU(chk_crypt_w),
+				  view_info->chk_crypt);
+	gtk_option_menu_set_history (GTK_OPTION_MENU (chk_crypt_w),
+				     libbalsa_mailbox_get_crypto_mode(mailbox));
+	gtk_box_pack_start(GTK_BOX(box), chk_crypt_w, FALSE, FALSE, 0);
+#endif /* GTK_CHECK_VERSION(2, 4, 0) */
+	gtk_table_attach(GTK_TABLE(table), box, 1, 2, row, row + 1,
+			 GTK_FILL, 0, 0, 0);
+    }
+#endif
+
     return view_info;
 }
 
+#ifdef HAVE_GPGME
+static LibBalsaChkCryptoMode
+balsa_mailbox_conf_get_crypto_mode(BalsaMailboxConfView *view_info)
+{
+#if GTK_CHECK_VERSION(2, 4, 0)
+    struct mailbox_conf_combo_box_info *info =
+        g_object_get_data(G_OBJECT(view_info->chk_crypt),
+                          BALSA_MC_COMBO_BOX_INFO);
+    gint active =
+        gtk_combo_box_get_active(GTK_COMBO_BOX(view_info->chk_crypt));
+
+    return (LibBalsaChkCryptoMode)
+        GPOINTER_TO_INT(g_slist_nth_data(info->tags, active));
+#else /* GTK_CHECK_VERSION(2, 4, 0) */
+    GtkWidget *menu_item = gtk_menu_get_active(GTK_MENU(view_info->chk_crypt));
+    return GPOINTER_TO_INT(g_object_get_data(G_OBJECT(menu_item),
+                                             "balsa-data"));
+#endif /* GTK_CHECK_VERSION(2, 4, 0) */
+}
+#endif
+
 /* When closing the dialog, check whether any view items were changed,
  * and carry out the changes if necessary.
  *
@@ -1345,6 +1478,12 @@ mailbox_conf_view_check(BalsaMailboxConfView * view_info,
 				  : LB_MAILBOX_SHOW_TO))
 	changed = TRUE;
 
+#ifdef HAVE_GPGME
+    if (libbalsa_mailbox_set_crypto_mode(mailbox,
+					 balsa_mailbox_conf_get_crypto_mode(view_info)))
+	changed = TRUE;
+#endif
+
     if (!changed || !libbalsa_mailbox_get_open(mailbox))
 	return;
 
