@@ -354,8 +354,13 @@ imap_scan_attach_mailbox(BalsaMailboxNode * mbnode, imap_scan_item * isi)
 {
     LibBalsaMailboxImap *m;
 
-    g_signal_connect(G_OBJECT(mbnode), "show-prop-dialog",
-                     G_CALLBACK(folder_conf_imap_sub_node), NULL);
+    /* If the mailbox was added from the config file, it is already
+     * connected to "show-prop-dialog". */
+    if (!g_signal_has_handler_pending(G_OBJECT(mbnode),
+                                      balsa_mailbox_node_signals
+                                      [SHOW_PROP_DIALOG], 0, FALSE))
+	g_signal_connect(G_OBJECT(mbnode), "show-prop-dialog",
+                         G_CALLBACK(folder_conf_imap_sub_node), NULL);
     if (LIBBALSA_IS_MAILBOX_IMAP(mbnode->mailbox))
         /* it already has a mailbox */
         return FALSE;
@@ -1073,7 +1078,7 @@ add_local_mailbox(BalsaMailboxNode *root, const gchar * name,
 		LIBBALSA_MAILBOX(libbalsa_mailbox_maildir_new(path, FALSE));
 	} else {
 	    /* type is not a valid local mailbox type. */
-	    balsa_information(LIBBALSA_INFORMATION_WARNING,
+	    balsa_information(LIBBALSA_INFORMATION_DEBUG,
 			      _("The path \"%s\" does not lead to a mailbox."),
 			      path);
 	    mailbox = NULL;
@@ -1208,7 +1213,8 @@ add_imap_entry(const char *fn, char delim, gboolean noscan,
     item->marked = marked;
 
     tree->list = g_slist_prepend(tree->list, item);
-    tree->delim = delim;
+    if(delim)  /* some servers may set delim to NIL for some mailboxes */
+        tree->delim = delim;
 }
 
 static void
