@@ -39,6 +39,8 @@
 #include <locale.h>
 #endif
 
+#include "address-entry.h"
+
 #include "balsa-app.h"
 #include "balsa-icons.h"
 #include "balsa-index.h"
@@ -323,18 +325,41 @@ mailboxes_init(gboolean check_only)
 
 
 #ifdef BALSA_USE_THREADS
+#if GTK_CHECK_VERSION(2, 4, 0)
+/* Recursive mutex for gdk_threads_{enter,leave}. */
+static pthread_mutex_t balsa_threads_mutex;
+
+static void
+balsa_threads_enter(void)
+{
+    pthread_mutex_lock(&balsa_threads_mutex);
+}
+
+static void
+balsa_threads_leave(void)
+{
+    pthread_mutex_unlock(&balsa_threads_mutex);
+}
+#endif /* GTK_CHECK_VERSION(2, 4, 0) */
+
 static void
 threads_init(void)
 {
     int status;
     pthread_mutexattr_t attr;
     g_thread_init(NULL);
-    gdk_threads_init();
     
     if( (status=pthread_mutexattr_init(&attr)) )
         g_warning("pthread_mutexattr_init failed with %d\n", status);
     pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);
     pthread_mutex_init(&mailbox_lock, &attr);
+#if GTK_CHECK_VERSION(2, 4, 0)
+    pthread_mutex_init(&balsa_threads_mutex, &attr);
+    gdk_threads_set_lock_functions(G_CALLBACK(balsa_threads_enter),
+                                   G_CALLBACK(balsa_threads_leave));
+#endif /* GTK_CHECK_VERSION(2, 4, 0) */
+    gdk_threads_init();
+
     pthread_mutexattr_destroy(&attr);
     pthread_mutex_init(&send_messages_lock, NULL);
     checking_mail = 0;
@@ -365,6 +390,9 @@ threads_destroy(void)
 {
     pthread_mutex_destroy(&mailbox_lock);
     pthread_mutex_destroy(&send_messages_lock);
+#if GTK_CHECK_VERSION(2, 4, 0)
+    pthread_mutex_destroy(&balsa_threads_mutex);
+#endif /* GTK_CHECK_VERSION(2, 4, 0) */
 }
 
 #endif				/* BALSA_USE_THREADS */
@@ -576,6 +604,10 @@ main(int argc, char *argv[])
     
     /* checking for valid config files */
     config_init(cmd_get_stats);
+#if NEW_ADDRESS_ENTRY_WIDGET
+    libbalsa_address_entry_set_address_book_list(balsa_app.
+                                                 address_book_list);
+#endif /* NEW_ADDRESS_ENTRY_WIDGET */
 
     libbalsa_mailbox_view_table =
 	g_hash_table_new_full(g_str_hash, g_str_equal,
