@@ -64,6 +64,8 @@ static void mw_set_buttons_sensitive(MessageWindow * mw);
 static void copy_cb(GtkWidget * widget, MessageWindow * mw);
 static void select_all_cb(GtkWidget * widget, gpointer);
 
+static void next_message_cb(GtkWidget * widget, gpointer data);
+static void previous_message_cb(GtkWidget * widget, gpointer data);
 static void next_unread_cb(GtkWidget * widget, gpointer);
 static void next_flagged_cb(GtkWidget * widget, gpointer);
 static void print_cb(GtkWidget * widget, gpointer);
@@ -167,15 +169,17 @@ static GnomeUIInfo message_menu[] = {
      forward_message_inline_cb, NULL, NULL, GNOME_APP_PIXMAP_STOCK,
      BALSA_PIXMAP_MENU_FORWARD, 'F', GDK_CONTROL_MASK, NULL},
     GNOMEUIINFO_SEPARATOR,
+#define MENU_MESSAGE_NEXT_PART_POS 6
     {
      GNOME_APP_UI_ITEM, N_("Next Part"), N_("Next part in message"),
      next_part_cb, NULL, NULL, GNOME_APP_PIXMAP_STOCK,
-     BALSA_PIXMAP_MENU_NEXT, '.', GDK_CONTROL_MASK, NULL},
+     BALSA_PIXMAP_MENU_NEXT_PART, '.', GDK_CONTROL_MASK, NULL},
+#define MENU_MESSAGE_PREVIOUS_PART_POS 7
     {
      GNOME_APP_UI_ITEM, N_("Previous Part"),
      N_("Previous part in message"),
      previous_part_cb, NULL, NULL, GNOME_APP_PIXMAP_STOCK,
-     BALSA_PIXMAP_MENU_PREVIOUS, ',', GDK_CONTROL_MASK, NULL},
+     BALSA_PIXMAP_MENU_PREVIOUS_PART, ',', GDK_CONTROL_MASK, NULL},
     {
      GNOME_APP_UI_ITEM, N_("Save Current Part..."),
      N_("Save current part in message"),
@@ -187,12 +191,41 @@ static GnomeUIInfo message_menu[] = {
      view_msg_source_cb, NULL, NULL, GNOME_APP_PIXMAP_STOCK,
      BALSA_PIXMAP_MENU_SAVE, 'v', GDK_CONTROL_MASK, NULL},
      GNOMEUIINFO_SEPARATOR,
-#define MENU_MESSAGE_NEXT_UNREAD_POS 11
+    /* N */
+#define MENU_MESSAGE_NEXT_POS 11
+    {
+     GNOME_APP_UI_ITEM, N_("_Next Message"), N_("Next message"),
+     next_message_cb, NULL, NULL, GNOME_APP_PIXMAP_STOCK,
+     BALSA_PIXMAP_MENU_NEXT, 'N', 0, NULL},
+    /* P */
+#define MENU_MESSAGE_PREVIOUS_POS 12
+    {
+     GNOME_APP_UI_ITEM, N_("_Previous Message"),
+     N_("Previous Message"),
+     previous_message_cb, NULL, NULL, GNOME_APP_PIXMAP_STOCK,
+     BALSA_PIXMAP_MENU_PREVIOUS, 'P', 0, NULL},
+    /* Ctrl + N */
+#define MENU_MESSAGE_NEXT_UNREAD_POS 13
     {
      GNOME_APP_UI_ITEM, N_("Next Unread Message"),
      N_("Next Unread Message"), next_unread_cb, NULL, NULL,
      GNOME_APP_PIXMAP_STOCK, BALSA_PIXMAP_MENU_NEXT_UNREAD, 'N',
      GDK_CONTROL_MASK, NULL},
+    /* Ctrl + Alt + F */
+#define MENU_MESSAGE_NEXT_FLAGGED_POS 14
+    {
+     GNOME_APP_UI_ITEM, N_("Next Flagged Message"),
+     N_("Next Flagged Message"), next_flagged_cb, NULL, NULL,
+     GNOME_APP_PIXMAP_STOCK, BALSA_PIXMAP_MENU_NEXT_FLAGGED, 'F',
+     GDK_MOD1_MASK|GDK_CONTROL_MASK, NULL},
+     GNOMEUIINFO_SEPARATOR,
+     /* D */
+#define MENU_MESSAGE_TRASH_POS 16
+    {
+     GNOME_APP_UI_ITEM, N_("_Move to Trash"),
+     N_("Move the current message to Trash mailbox"),
+     trash_cb, NULL, NULL, GNOME_APP_PIXMAP_STOCK,
+     GNOME_STOCK_TRASH, 'D', 0, NULL},
     GNOMEUIINFO_END
 };
 
@@ -223,6 +256,7 @@ struct _MessageWindow {
 
     /* Pointers to our copies of widgets. */
     GtkWidget *next_unread;
+    GtkWidget *next_flagged;
 #ifdef HAVE_GTKHTML
     GtkWidget *zoom_in;
     GtkWidget *zoom_out;
@@ -242,7 +276,7 @@ mru_menu_cb(gchar * url, gpointer data)
 static void
 message_window_move_message(MessageWindow * mw, LibBalsaMailbox * mailbox)
 {
-    GList *list;
+    GArray *messages;
 
     g_return_if_fail(mailbox != NULL);
 
@@ -250,13 +284,39 @@ message_window_move_message(MessageWindow * mw, LibBalsaMailbox * mailbox)
     if (mw->message->mailbox == mailbox)
 	return;
 
-    list = g_list_append(NULL, mw->message);
-    balsa_index_transfer(mw->bindex, list, mailbox, FALSE);
-    g_list_free_1(list);
+    messages = g_array_new(FALSE, FALSE, sizeof(guint));
+    g_array_append_val(messages, mw->message->msgno);
+    balsa_index_transfer(mw->bindex, messages, mailbox, FALSE);
+    g_array_free(messages, TRUE);
     
     close_message_window(NULL, (gpointer) mw);
 }
 
+static void
+mw_set_part_buttons_sensitive(MessageWindow * mw)
+{
+    BalsaMessage *msg = BALSA_MESSAGE(mw->bmessage);
+    GtkWidget *toolbar;
+    gboolean enable;
+
+    if (!msg || !msg->treeview)
+	return;
+
+    toolbar = balsa_toolbar_get_from_gnome_app(GNOME_APP(mw->window));
+
+    enable = balsa_message_has_next_part(msg);
+    balsa_toolbar_set_button_sensitive(toolbar, BALSA_PIXMAP_NEXT_PART,
+                                       enable);
+    gtk_widget_set_sensitive(message_menu[MENU_MESSAGE_NEXT_PART_POS].
+                             widget, enable);
+
+    enable = balsa_message_has_previous_part(msg);
+    balsa_toolbar_set_button_sensitive(toolbar, BALSA_PIXMAP_PREVIOUS_PART,
+                                       enable);
+    gtk_widget_set_sensitive(message_menu[MENU_MESSAGE_PREVIOUS_PART_POS].
+                             widget, enable);
+}
+
 static gboolean
 message_window_idle_handler(MessageWindow* mw)
 {
@@ -269,6 +329,12 @@ message_window_idle_handler(MessageWindow* mw)
 
     mw->idle_handler_id = 0;
 
+    if (!message->mailbox) {
+	gtk_widget_destroy(mw->window);
+	gdk_threads_leave();
+	return FALSE;
+    }
+
     /* set window title */
     from = libbalsa_address_to_gchar(message->headers->from, 0);
     title = g_strdup_printf(_("Message from %s: %s"), from,
@@ -278,19 +344,18 @@ message_window_idle_handler(MessageWindow* mw)
     g_free(title);
 
     balsa_message_set(msg, message);
-    balsa_message_set_close(msg, TRUE);
-
-    if(msg && msg->treeview) {
-	if (msg->info_count > 1) {
-            GtkWidget *toolbar =
-                balsa_toolbar_get_from_gnome_app(GNOME_APP(mw->window));
-	    balsa_toolbar_set_button_sensitive(toolbar,
-                                               BALSA_PIXMAP_PREVIOUS, TRUE);
-	    balsa_toolbar_set_button_sensitive(toolbar,
-                                               BALSA_PIXMAP_NEXT, TRUE);
-	}
+    balsa_message_grab_focus(BALSA_MESSAGE(mw->bmessage));
+    /* We may have triggered a mailbox-check, so we need to verify that
+     * the message is still good. */
+    if (!message->mailbox) {
+	gtk_widget_destroy(mw->window);
+	gdk_threads_leave();
+	return FALSE;
     }
 
+    balsa_message_set_close(msg, TRUE);
+    mw_set_part_buttons_sensitive(mw);
+
     /* Update the style and message counts in the mailbox list */
     balsa_mblist_update_mailbox(balsa_app.mblist_tree_store,
                                 message->mailbox);
@@ -308,8 +373,10 @@ static const struct callback_item {
     { BALSA_PIXMAP_REPLY_ALL,    replytoall_message_cb },
     { BALSA_PIXMAP_REPLY_GROUP,  replytogroup_message_cb },
     { BALSA_PIXMAP_FORWARD,      forward_message_default_cb },
-    { BALSA_PIXMAP_PREVIOUS,     previous_part_cb },
-    { BALSA_PIXMAP_NEXT,         next_part_cb },
+    { BALSA_PIXMAP_PREVIOUS,     previous_message_cb },
+    { BALSA_PIXMAP_PREVIOUS_PART,previous_part_cb },
+    { BALSA_PIXMAP_NEXT,         next_message_cb },
+    { BALSA_PIXMAP_NEXT_PART,    next_part_cb },
     { BALSA_PIXMAP_NEXT_UNREAD,  next_unread_cb },
     { BALSA_PIXMAP_NEXT_FLAGGED, next_flagged_cb },
     { BALSA_PIXMAP_TRASH,        trash_cb },
@@ -321,6 +388,19 @@ static const struct callback_item {
 
 /* Standard buttons; "" means a separator. */
 static const gchar* message_toolbar[] = {
+#if defined(ENABLE_TOUCH_UI)
+    BALSA_PIXMAP_NEXT_UNREAD,
+    "",
+    BALSA_PIXMAP_REPLY,
+    BALSA_PIXMAP_REPLY_ALL,
+    BALSA_PIXMAP_FORWARD,
+    "",
+    BALSA_PIXMAP_PRINT,
+    "",
+    BALSA_PIXMAP_TRASH,
+    "",
+    GTK_STOCK_CLOSE
+#else /* ENABLE_TOUCH_UI */
     BALSA_PIXMAP_NEXT_UNREAD,
     "",
     BALSA_PIXMAP_REPLY,
@@ -328,13 +408,14 @@ static const gchar* message_toolbar[] = {
     BALSA_PIXMAP_REPLY_GROUP,
     BALSA_PIXMAP_FORWARD,
     "",
-    BALSA_PIXMAP_PREVIOUS,
-    BALSA_PIXMAP_NEXT,
+    BALSA_PIXMAP_PREVIOUS_PART,
+    BALSA_PIXMAP_NEXT_PART,
     BALSA_PIXMAP_SAVE,
     "",
     BALSA_PIXMAP_PRINT,
     "",
-    BALSA_PIXMAP_TRASH,
+    BALSA_PIXMAP_TRASH
+#endif /* ENEBLE_TOUCH_UI */
 };
 
 /* Create the toolbar model for the message window's toolbar.
@@ -372,12 +453,13 @@ static void
 mw_set_message(MessageWindow *mw, LibBalsaMessage * message)
 {
     mw->message = message;
-    mw_set_buttons_sensitive(mw);
     g_object_set_data(G_OBJECT(message), BALSA_MESSAGE_WINDOW_KEY, mw);
 
     g_object_ref(message); /* protect from destroying */
     mw->idle_handler_id =
         g_idle_add((GSourceFunc) message_window_idle_handler, mw);
+
+    mw_set_buttons_sensitive(mw);
 }
 
 static void
@@ -388,6 +470,14 @@ bindex_closed_cb(gpointer data, GObject *bindex)
     gtk_widget_destroy(mw->window);
 }
 
+static void
+mw_disable_trash(GtkWidget * toolbar)
+{
+    balsa_toolbar_set_button_sensitive(toolbar, BALSA_PIXMAP_TRASH, FALSE);
+    gtk_widget_set_sensitive(message_menu[MENU_MESSAGE_TRASH_POS].widget,
+                             FALSE);
+}
+
 void
 message_window_new(LibBalsaMessage * message)
 {
@@ -428,8 +518,8 @@ message_window_new(LibBalsaMessage * message)
                                    mw);
     gnome_app_set_toolbar(GNOME_APP(mw->window), GTK_TOOLBAR(toolbar));
 
-    balsa_toolbar_set_button_sensitive(toolbar, BALSA_PIXMAP_PREVIOUS, FALSE);
-    balsa_toolbar_set_button_sensitive(toolbar, BALSA_PIXMAP_NEXT, FALSE);
+    balsa_toolbar_set_button_sensitive(toolbar, BALSA_PIXMAP_PREVIOUS_PART, FALSE);
+    balsa_toolbar_set_button_sensitive(toolbar, BALSA_PIXMAP_NEXT_PART, FALSE);
 
     gtk_window_set_wmclass(GTK_WINDOW(mw->window), "message", "Balsa");
 
@@ -447,7 +537,8 @@ message_window_new(LibBalsaMessage * message)
 
     /* Save the widgets that we need to change--they'll be overwritten
      * if another message window is opened. */
-    mw->next_unread = message_menu[MENU_MESSAGE_NEXT_UNREAD_POS].widget;
+    mw->next_unread  = message_menu[MENU_MESSAGE_NEXT_UNREAD_POS].widget;
+    mw->next_flagged = message_menu[MENU_MESSAGE_NEXT_FLAGGED_POS].widget;
 #ifdef HAVE_GTKHTML
     mw->zoom_in  = view_menu[MENU_VIEW_ZOOM_IN].widget;
     mw->zoom_out = view_menu[MENU_VIEW_ZOOM_OUT].widget;
@@ -465,8 +556,12 @@ message_window_new(LibBalsaMessage * message)
                                     G_CALLBACK(mru_menu_cb), mw);
     move_menu = main_menu[MAIN_MENU_MOVE_POS].widget;
     gtk_menu_item_set_submenu(GTK_MENU_ITEM(move_menu), submenu);
-    if(message->mailbox->readonly)
+    if(message->mailbox->readonly) {
 	gtk_widget_set_sensitive(move_menu, FALSE);
+	mw_disable_trash(toolbar);
+    }
+    if (message->mailbox == balsa_app.trash)
+	mw_disable_trash(toolbar);
     mw->bmessage = balsa_message_new();
     
     gnome_app_set_contents(GNOME_APP(mw->window), mw->bmessage);
@@ -498,6 +593,7 @@ message_window_new(LibBalsaMessage * message)
 			       GNOME_APP(mw->window)->accel_group,
 			       GDK_Escape, (GdkModifierType)0, 
                                (GtkAccelFlags) 0);
+    GTK_WIDGET_UNSET_FLAGS(toolbar, GTK_CAN_FOCUS);
 }
 
 static void
@@ -604,6 +700,7 @@ next_part_cb(GtkWidget * widget, gpointer data)
     MessageWindow *mw = (MessageWindow *) data;
 
     balsa_message_next_part(BALSA_MESSAGE(mw->bmessage));
+    mw_set_part_buttons_sensitive(mw);
 }
 
 static void
@@ -612,6 +709,7 @@ previous_part_cb(GtkWidget * widget, gpointer data)
     MessageWindow *mw = (MessageWindow *) data;
 
     balsa_message_previous_part(BALSA_MESSAGE(mw->bmessage));
+    mw_set_part_buttons_sensitive(mw);
 }
 
 static void
@@ -693,18 +791,39 @@ static void
 mw_set_buttons_sensitive(MessageWindow * mw)
 {
     GtkWidget *toolbar =
-	balsa_toolbar_get_from_gnome_app(GNOME_APP(mw->window));
+        balsa_toolbar_get_from_gnome_app(GNOME_APP(mw->window));
     LibBalsaMailbox *mailbox = mw->message->mailbox;
-    gint other_unread =
-	mailbox->unread_messages - LIBBALSA_MESSAGE_IS_UNREAD(mw->message);
+    guint current_msgno = mw->message->msgno;
+    BalsaIndex *index = mw->bindex;
+    gboolean enable;
 
-    gtk_widget_set_sensitive(mw->next_unread, other_unread > 0);
+    if (!mailbox) {
+        gtk_widget_destroy(mw->window);
+        return;
+    }
+
+    enable = index && balsa_index_next_msgno(index, current_msgno) > 0;
+    balsa_toolbar_set_button_sensitive(toolbar, BALSA_PIXMAP_NEXT, enable);
+    gtk_widget_set_sensitive(message_menu[MENU_MESSAGE_NEXT_POS].widget,
+                             enable);
+
+    enable = index && balsa_index_previous_msgno(index, current_msgno) > 0;
+    balsa_toolbar_set_button_sensitive(toolbar, BALSA_PIXMAP_PREVIOUS,
+                                       enable);
+    gtk_widget_set_sensitive(message_menu[MENU_MESSAGE_PREVIOUS_POS].
+                             widget, enable);
+
+    enable = index
+        && balsa_index_next_unread_msgno(index, current_msgno) > 0;
     balsa_toolbar_set_button_sensitive(toolbar, BALSA_PIXMAP_NEXT_UNREAD,
-				       other_unread > 0);
+                                       enable);
+    gtk_widget_set_sensitive(mw->next_unread, enable);
+
+    enable = index
+        && balsa_index_next_flagged_msgno(index, current_msgno) > 0;
     balsa_toolbar_set_button_sensitive(toolbar, BALSA_PIXMAP_NEXT_FLAGGED,
-				       mailbox
-				       && libbalsa_mailbox_total_messages
-				       (mailbox) > 0);
+                                       enable);
+    gtk_widget_set_sensitive(mw->next_flagged, enable);
 }
 
 static void
@@ -728,9 +847,24 @@ select_all_cb(GtkWidget * widget, gpointer data)
 }
 
 static void
-mw_set_selected(MessageWindow * mw)
+mw_set_msgno(MessageWindow * mw, guint msgno)
 {
-    GList *list;
+    LibBalsaMessage *message;
+    MessageWindow *tmp;
+
+    message =
+        libbalsa_mailbox_get_message(mw->bindex->mailbox_node->mailbox,
+                                     msgno);
+    tmp = g_object_get_data(G_OBJECT(message), BALSA_MESSAGE_WINDOW_KEY);
+    if (tmp) {
+        /*
+         * The message is already displayed in a window, so just use
+         * that one.
+         */
+        g_object_unref(message);
+        gdk_window_raise(tmp->window->window);
+        return;
+    }
 
     /* Temporarily tell the BalsaMessage not to close when its message
      * is finalized, so we can safely unref it in mw_clear_message.
@@ -738,12 +872,27 @@ mw_set_selected(MessageWindow * mw)
      * the new message. */
     balsa_message_set_close(BALSA_MESSAGE(mw->bmessage), FALSE);
     mw_clear_message(mw);
+    mw_set_message(mw, message);
+    g_object_unref(message);
+}
 
-    list = balsa_index_selected_list(mw->bindex);
-    if (g_list_length(list) == 1)
-	mw_set_message(mw, LIBBALSA_MESSAGE(list->data));
-    g_list_foreach(list, (GFunc) g_object_unref, NULL);
-    g_list_free(list);
+static void
+next_message_cb(GtkWidget * widget, gpointer data)
+{
+    MessageWindow *mw = (MessageWindow *) (data);
+
+    mw_set_msgno(mw,
+                 balsa_index_next_msgno(mw->bindex, mw->message->msgno));
+}
+
+static void
+previous_message_cb(GtkWidget * widget, gpointer data)
+{
+    MessageWindow *mw = (MessageWindow *) (data);
+
+    mw_set_msgno(mw,
+                 balsa_index_previous_msgno(mw->bindex,
+                                            mw->message->msgno));
 }
 
 static void
@@ -751,16 +900,18 @@ next_unread_cb(GtkWidget * widget, gpointer data)
 {
     MessageWindow *mw = (MessageWindow *) (data);
 
-    balsa_index_select_next_unread(mw->bindex);
-    mw_set_selected(mw);
+    mw_set_msgno(mw,
+                 balsa_index_next_unread_msgno(mw->bindex,
+                                               mw->message->msgno));
 }
 
 static void next_flagged_cb(GtkWidget * widget, gpointer data)
 {
     MessageWindow *mw = (MessageWindow *) (data);
 
-    balsa_index_select_next_flagged(mw->bindex);
-    mw_set_selected(mw);
+    mw_set_msgno(mw,
+                 balsa_index_next_flagged_msgno(mw->bindex,
+                                                mw->message->msgno));
 }
 
 
