@@ -589,6 +589,7 @@ static gint
 config_global_load(void)
 {
     gboolean def_used;
+    guint tmp;
 
     config_address_books_load();
     config_identities_load();
@@ -677,6 +678,7 @@ config_global_load(void)
     g_free(balsa_app.date_string);
     balsa_app.date_string =
 	gnome_config_get_string("DateFormat=" DEFAULT_DATE_FORMAT);
+    libbalsa_mailbox_date_format = balsa_app.date_string;
 
     /* ... Headers to show */
     balsa_app.shown_headers = d_get_gint("ShownHeaders", HEADERS_SELECTED);
@@ -693,6 +695,10 @@ config_global_load(void)
 
     {                             /* scope */
         guint type =
+            gnome_config_get_int_with_default("SortField", &def_used);
+        if (!def_used)
+            libbalsa_mailbox_set_sort_field(NULL, type);
+        type =
             gnome_config_get_int_with_default("ThreadingType", &def_used);
         if (!def_used)
             libbalsa_mailbox_set_threading_type(NULL, type);
@@ -780,15 +786,22 @@ config_global_load(void)
     balsa_app.pwindow_option = d_get_gint("ProgressWindow", WHILERETR);
 
     /* ... deleting messages: defaults enshrined here */
-    balsa_app.hide_deleted =
-        gnome_config_get_bool("HideDeleted=true");
-    libbalsa_mailbox_set_filter(NULL, balsa_app.hide_deleted ? 1 : 0);
+    tmp = libbalsa_mailbox_get_filter(NULL);
+    if (gnome_config_get_bool("HideDeleted=true"))
+	tmp |= (1 << 0);
+    else
+	tmp &= ~(1 << 0);
+    libbalsa_mailbox_set_filter(NULL, tmp);
+
     balsa_app.expunge_on_close =
         gnome_config_get_bool("ExpungeOnClose=true");
     balsa_app.expunge_auto = gnome_config_get_bool("AutoExpunge=true");
-    /* timeout in hours */
+    /* timeout in munutes (was hours, so fall back if needed) */
     balsa_app.expunge_timeout =
-        gnome_config_get_int("AutoExpungeHours=2") * 3600;
+        gnome_config_get_int("AutoExpungeMinutes") * 60;
+    if (!balsa_app.expunge_timeout)
+	balsa_app.expunge_timeout = 
+	    gnome_config_get_int("AutoExpungeHours=2") * 3600;
 
     gnome_config_pop_prefix();
 
@@ -1120,6 +1133,8 @@ config_save(void)
     gnome_config_set_int("ShownHeaders", balsa_app.shown_headers);
     gnome_config_set_string("SelectedHeaders", balsa_app.selected_headers);
     gnome_config_set_bool("ExpandTree", balsa_app.expand_tree);
+    gnome_config_set_int("SortField",
+			 libbalsa_mailbox_get_sort_field(NULL));
     gnome_config_set_int("ThreadingType",
 			 libbalsa_mailbox_get_threading_type(NULL));
     gnome_config_set_string("QuoteRegex", balsa_app.quote_regex);
@@ -1163,11 +1178,11 @@ config_save(void)
     gnome_config_set_bool("FileFormatCheck", balsa_app.do_file_format_check);
     gnome_config_set_bool("ViewFilter",      balsa_app.enable_view_filter);
 #endif /* ENABLE_TOUCH_UI */
-    gnome_config_set_bool("HideDeleted", balsa_app.hide_deleted);
+    gnome_config_set_bool("HideDeleted", libbalsa_mailbox_get_filter(NULL) & 1);
     gnome_config_set_bool("ExpungeOnClose", balsa_app.expunge_on_close);
     gnome_config_set_bool("AutoExpunge", balsa_app.expunge_auto);
-    gnome_config_set_int("AutoExpungeHours",
-                         balsa_app.expunge_timeout / 3600);
+    gnome_config_set_int("AutoExpungeMinutes",
+                         balsa_app.expunge_timeout / 60);
 
     gnome_config_pop_prefix();
 
@@ -1607,6 +1622,11 @@ config_views_load_with_prefix(const gchar * prefix, gboolean compat)
 		tmp = gnome_config_get_bool_with_default("Open", &def);
                 if (!def)
 		    view->open = tmp;
+#ifdef HAVE_GPGME
+		tmp = gnome_config_get_int_with_default("CryptoMode", &def);
+                if (!def)
+		    view->gpg_chk_mode = tmp;
+#endif
             }
             gnome_config_pop_prefix();
             g_free(url);
@@ -1670,6 +1690,10 @@ save_view(const gchar * url, LibBalsaMailboxView * view)
 	gnome_config_set_bool("Exposed",    view->exposed);
     if (view->open           != libbalsa_mailbox_get_open(NULL))
 	gnome_config_set_bool("Open",	    view->open);
+#ifdef HAVE_GPGME
+    if (view->gpg_chk_mode   != libbalsa_mailbox_get_crypto_mode(NULL))
+	gnome_config_set_int("CryptoMode",  view->gpg_chk_mode);
+#endif
 
     gnome_config_pop_prefix();
 }
@@ -1830,8 +1854,10 @@ load_mru(GList **mru)
     
     count=d_get_gint("MRUCount", 0);
     for(i=0;i<count;i++) {
+        gchar *val;
 	sprintf(tmpkey, "MRU%d", i+1);
-	(*mru)=g_list_append((*mru), gnome_config_get_string(tmpkey));
+        if( (val = gnome_config_get_string(tmpkey)) != NULL )
+            (*mru)=g_list_append((*mru), val);
     }
 }
 
